Add LoadPath "metatheory-8.2".
Require Export Coq.Program.Equality.
Require Export Metatheory.
Require Export LNgenFacts.
Require Import Wf_nat.

(** NOTE: Auxiliary theorems are hidden in generated documentation.
    In general, there is a [_rec] version of every lemma involving
    [open] and [close]. *)

(** For generated HTML documentation. *)

(** remove printing <= *)

(** For proving the uniqueness of [lc] proofs. *)

Axiom axiom_fun_ext_eq :
  forall (A : Type) (B : A -> Type) (f g : forall x : A, B x),
    (forall x, f x = g x) ->
    f = g.


(* *********************************************************************** *)
(** * Metavariable declarations *)

(** The rest of this file assumes that [EqDec] and [Atom] instances
    are already declared for the implementation types below. *)

Notation typvar := (string).

Notation expvar := (string).


(* *********************************************************************** *)
(** * Nonterminal declarations *)

Inductive typ : Set :=
  | typ_var : typvar -> typ
  | typ_var_ind : nat -> typ
  | typ_top : typ
  | typ_arrow : typ -> typ -> typ
  | typ_all : typ -> typ -> typ.

Scheme typ_ind' := Induction for typ Sort Prop.

Definition typ_mutind :=
  fun H H1 H2 H3 H4 H5 =>
  typ_ind' H H1 H2 H3 H4 H5.

Scheme typ_rec' := Induction for typ Sort Set.

Definition typ_mutrec :=
  fun H H1 H2 H3 H4 H5 =>
  typ_rec' H H1 H2 H3 H4 H5.

Inductive exp : Set :=
  | exp_var : expvar -> exp
  | exp_var_ind : nat -> exp
  | exp_abs : typ -> exp -> exp
  | exp_app : exp -> exp -> exp
  | exp_tabs : typ -> exp -> exp
  | exp_tapp : exp -> typ -> exp.

Scheme exp_ind' := Induction for exp Sort Prop.

Definition exp_mutind :=
  fun H H1 H2 H3 H4 H5 H6 =>
  exp_ind' H H1 H2 H3 H4 H5 H6.

Scheme exp_rec' := Induction for exp Sort Set.

Definition exp_mutrec :=
  fun H H1 H2 H3 H4 H5 H6 =>
  exp_rec' H H1 H2 H3 H4 H5 H6.


(* *********************************************************************** *)
(** * Close *)

Fixpoint close_tX_rec (n : nat) (X : typvar) (t : typ) {struct t} : typ :=
  match t with
    | typ_var X1 => if (X == X1) then (typ_var_ind n) else (typ_var X1)
    | typ_var_ind n1 => if (lt_ge_dec n1 n) then (typ_var_ind n1) else (typ_var_ind (S n1))
    | typ_top => typ_top
    | typ_arrow t1 t2 => typ_arrow (close_tX_rec n X t1) (close_tX_rec n X t2)
    | typ_all t1 t2 => typ_all (close_tX_rec n X t1) (close_tX_rec (S n) X t2)
  end.

Definition close_tX X t := close_tX_rec 0 X t.

Fixpoint close_ex_rec (n : nat) (x : expvar) (e : exp) {struct e} : exp :=
  match e with
    | exp_var x1 => if (x == x1) then (exp_var_ind n) else (exp_var x1)
    | exp_var_ind n1 => if (lt_ge_dec n1 n) then (exp_var_ind n1) else (exp_var_ind (S n1))
    | exp_abs t e1 => exp_abs t (close_ex_rec (S n) x e1)
    | exp_app e1 e2 => exp_app (close_ex_rec n x e1) (close_ex_rec n x e2)
    | exp_tabs t e1 => exp_tabs t (close_ex_rec n x e1)
    | exp_tapp e1 t => exp_tapp (close_ex_rec n x e1) t
  end.

Fixpoint close_eX_rec (n : nat) (X : typvar) (e : exp) {struct e} : exp :=
  match e with
    | exp_var x => exp_var x
    | exp_var_ind n1 => exp_var_ind n1
    | exp_abs t e1 => exp_abs (close_tX_rec n X t) (close_eX_rec n X e1)
    | exp_app e1 e2 => exp_app (close_eX_rec n X e1) (close_eX_rec n X e2)
    | exp_tabs t e1 => exp_tabs (close_tX_rec n X t) (close_eX_rec (S n) X e1)
    | exp_tapp e1 t => exp_tapp (close_eX_rec n X e1) (close_tX_rec n X t)
  end.

Definition close_ex x e := close_ex_rec 0 x e.

Definition close_eX X e := close_eX_rec 0 X e.


(* *********************************************************************** *)
(** * Free variables *)

Fixpoint fv_tX (t : typ) {struct t} : fset typvar :=
  match t with
    | typ_var X => (singleton X)
    | typ_var_ind n => empty
    | typ_top => empty
    | typ_arrow t1 t2 => (fv_tX t1) `union` (fv_tX t2)
    | typ_all t1 t2 => (fv_tX t1) `union` (fv_tX t2)
  end.

Fixpoint fv_ex (e : exp) {struct e} : fset expvar :=
  match e with
    | exp_var x => (singleton x)
    | exp_var_ind n => empty
    | exp_abs t e1 => (fv_ex e1)
    | exp_app e1 e2 => (fv_ex e1) `union` (fv_ex e2)
    | exp_tabs t e1 => (fv_ex e1)
    | exp_tapp e1 t => (fv_ex e1)
  end.

Fixpoint fv_eX (e : exp) {struct e} : fset typvar :=
  match e with
    | exp_var x => empty
    | exp_var_ind n => empty
    | exp_abs t e1 => (fv_tX t) `union` (fv_eX e1)
    | exp_app e1 e2 => (fv_eX e1) `union` (fv_eX e2)
    | exp_tabs t e1 => (fv_tX t) `union` (fv_eX e1)
    | exp_tapp e1 t => (fv_eX e1) `union` (fv_tX t)
  end.


(* *********************************************************************** *)
(** * Open *)

(** The definitions below are only correct under the following
    condition: when substituting for index [n], the [degree] of
    the term being substituted in should also be [n]. *)

Fixpoint open_tX_rec (n : nat) (t : typ) (t1 : typ) {struct t1} : typ :=
  match t1 with
    | typ_var X => typ_var X
    | typ_var_ind n1 =>
      match lt_eq_lt_dec n1 n with
        | inleft (left _)  => typ_var_ind n1
        | inleft (right _) => t
        | inright _        => typ_var_ind (n1 - 1)
      end
    | typ_top => typ_top
    | typ_arrow t2 t3 => typ_arrow (open_tX_rec n t t2) (open_tX_rec n t t3)
    | typ_all t2 t3 => typ_all (open_tX_rec n t t2) (open_tX_rec (S n) t t3)
  end.

Definition open_tX t1 t := open_tX_rec 0 t1 t.

Fixpoint open_ex_rec (n : nat) (e : exp) (e1 : exp) {struct e1} : exp :=
  match e1 with
    | exp_var x => exp_var x
    | exp_var_ind n1 =>
      match lt_eq_lt_dec n1 n with
        | inleft (left _)  => exp_var_ind n1
        | inleft (right _) => e
        | inright _        => exp_var_ind (n1 - 1)
      end
    | exp_abs t e2 => exp_abs t (open_ex_rec (S n) e e2)
    | exp_app e2 e3 => exp_app (open_ex_rec n e e2) (open_ex_rec n e e3)
    | exp_tabs t e2 => exp_tabs t (open_ex_rec n e e2)
    | exp_tapp e2 t => exp_tapp (open_ex_rec n e e2) t
  end.

Fixpoint open_eX_rec (n : nat) (t : typ) (e : exp) {struct e} : exp :=
  match e with
    | exp_var x => exp_var x
    | exp_var_ind n1 => exp_var_ind n1
    | exp_abs t1 e1 => exp_abs (open_tX_rec n t t1) (open_eX_rec n t e1)
    | exp_app e1 e2 => exp_app (open_eX_rec n t e1) (open_eX_rec n t e2)
    | exp_tabs t1 e1 => exp_tabs (open_tX_rec n t t1) (open_eX_rec (S n) t e1)
    | exp_tapp e1 t1 => exp_tapp (open_eX_rec n t e1) (open_tX_rec n t t1)
  end.

Definition open_ex e1 e := open_ex_rec 0 e1 e.

Definition open_eX t e := open_eX_rec 0 t e.


(* *********************************************************************** *)
(** * Size *)

Fixpoint size_t (t : typ) {struct t} : nat :=
  match t with
    | typ_var X => 1
    | typ_var_ind n => 1
    | typ_top => 1
    | typ_arrow t1 t2 => 1 + (size_t t1) + (size_t t2)
    | typ_all t1 t2 => 1 + (size_t t1) + (size_t t2)
  end.

Fixpoint size_e (e : exp) {struct e} : nat :=
  match e with
    | exp_var x => 1
    | exp_var_ind n => 1
    | exp_abs t e1 => 1 + (size_t t) + (size_e e1)
    | exp_app e1 e2 => 1 + (size_e e1) + (size_e e2)
    | exp_tabs t e1 => 1 + (size_t t) + (size_e e1)
    | exp_tapp e1 t => 1 + (size_e e1) + (size_t t)
  end.


(* *********************************************************************** *)
(** * Substitution *)

Fixpoint subst_tX (t : typ) (X : typvar) (t1 : typ) {struct t1} : typ :=
  match t1 with
    | typ_var X1 => if (X == X1) then (t) else (typ_var X1)
    | typ_var_ind n => typ_var_ind n
    | typ_top => typ_top
    | typ_arrow t2 t3 => typ_arrow (subst_tX t X t2) (subst_tX t X t3)
    | typ_all t2 t3 => typ_all (subst_tX t X t2) (subst_tX t X t3)
  end.

Fixpoint subst_ex (e : exp) (x : expvar) (e1 : exp) {struct e1} : exp :=
  match e1 with
    | exp_var x1 => if (x == x1) then (e) else (exp_var x1)
    | exp_var_ind n => exp_var_ind n
    | exp_abs t e2 => exp_abs t (subst_ex e x e2)
    | exp_app e2 e3 => exp_app (subst_ex e x e2) (subst_ex e x e3)
    | exp_tabs t e2 => exp_tabs t (subst_ex e x e2)
    | exp_tapp e2 t => exp_tapp (subst_ex e x e2) t
  end.

Fixpoint subst_eX (t : typ) (X : typvar) (e : exp) {struct e} : exp :=
  match e with
    | exp_var x => exp_var x
    | exp_var_ind n => exp_var_ind n
    | exp_abs t1 e1 => exp_abs (subst_tX t X t1) (subst_eX t X e1)
    | exp_app e1 e2 => exp_app (subst_eX t X e1) (subst_eX t X e2)
    | exp_tabs t1 e1 => exp_tabs (subst_tX t X t1) (subst_eX t X e1)
    | exp_tapp e1 t1 => exp_tapp (subst_eX t X e1) (subst_tX t X t1)
  end.


(* *********************************************************************** *)
(** * Degree *)

(** These define only an upper bound, not a strict upper bound. *)

Inductive degree_tX : nat -> typ -> Set :=
  | degree_tX_var : forall n X,
    degree_tX n (typ_var X)
  | degree_tX_var_ind : forall n n1,
    lt n1 n ->
    degree_tX n (typ_var_ind n1)
  | degree_tX_top : forall n,
    degree_tX n (typ_top)
  | degree_tX_arrow : forall n t t1,
    degree_tX n t ->
    degree_tX n t1 ->
    degree_tX n (typ_arrow t t1)
  | degree_tX_all : forall n t t1,
    degree_tX n t ->
    degree_tX (S n) t1 ->
    degree_tX n (typ_all t t1).

Scheme degree_tX_ind' := Induction for degree_tX Sort Prop.

Definition degree_tX_mutind :=
  fun H H1 H2 H3 H4 H5 =>
  degree_tX_ind' H H1 H2 H3 H4 H5.

Scheme degree_tX_rec' := Induction for degree_tX Sort Set.

Definition degree_tX_mutrec :=
  fun H H1 H2 H3 H4 H5 =>
  degree_tX_rec' H H1 H2 H3 H4 H5.

Hint Constructors degree_tX : core lngen.

Inductive degree_ex : nat -> exp -> Set :=
  | degree_ex_var : forall n x,
    degree_ex n (exp_var x)
  | degree_ex_var_ind : forall n n1,
    lt n1 n ->
    degree_ex n (exp_var_ind n1)
  | degree_ex_abs : forall n t e,
    degree_ex (S n) e ->
    degree_ex n (exp_abs t e)
  | degree_ex_app : forall n e e1,
    degree_ex n e ->
    degree_ex n e1 ->
    degree_ex n (exp_app e e1)
  | degree_ex_tabs : forall n t e,
    degree_ex n e ->
    degree_ex n (exp_tabs t e)
  | degree_ex_tapp : forall n e t,
    degree_ex n e ->
    degree_ex n (exp_tapp e t).

Inductive degree_eX : nat -> exp -> Set :=
  | degree_eX_var : forall n x,
    degree_eX n (exp_var x)
  | degree_eX_var_ind : forall n n1,
    degree_eX n (exp_var_ind n1)
  | degree_eX_abs : forall n t e,
    degree_tX n t ->
    degree_eX n e ->
    degree_eX n (exp_abs t e)
  | degree_eX_app : forall n e e1,
    degree_eX n e ->
    degree_eX n e1 ->
    degree_eX n (exp_app e e1)
  | degree_eX_tabs : forall n t e,
    degree_tX n t ->
    degree_eX (S n) e ->
    degree_eX n (exp_tabs t e)
  | degree_eX_tapp : forall n e t,
    degree_eX n e ->
    degree_tX n t ->
    degree_eX n (exp_tapp e t).

Scheme degree_ex_ind' := Induction for degree_ex Sort Prop.

Definition degree_ex_mutind :=
  fun H H1 H2 H3 H4 H5 H6 =>
  degree_ex_ind' H H1 H2 H3 H4 H5 H6.

Scheme degree_eX_ind' := Induction for degree_eX Sort Prop.

Definition degree_eX_mutind :=
  fun H H1 H2 H3 H4 H5 H6 =>
  degree_eX_ind' H H1 H2 H3 H4 H5 H6.

Scheme degree_ex_rec' := Induction for degree_ex Sort Set.

Definition degree_ex_mutrec :=
  fun H H1 H2 H3 H4 H5 H6 =>
  degree_ex_rec' H H1 H2 H3 H4 H5 H6.

Scheme degree_eX_rec' := Induction for degree_eX Sort Set.

Definition degree_eX_mutrec :=
  fun H H1 H2 H3 H4 H5 H6 =>
  degree_eX_rec' H H1 H2 H3 H4 H5 H6.

Hint Constructors degree_ex : core lngen.

Hint Constructors degree_eX : core lngen.


(* *********************************************************************** *)
(** * Local closure *)

Inductive lc_t : typ -> Set :=
  | lc_t_var : forall X,
    lc_t (typ_var X)
  | lc_t_top :
    lc_t (typ_top)
  | lc_t_arrow : forall t t1,
    lc_t t ->
    lc_t t1 ->
    lc_t (typ_arrow t t1)
  | lc_t_all : forall t t1,
    lc_t t ->
    (forall X : typvar, lc_t (open_tX (typ_var X) t1)) ->
    lc_t (typ_all t t1).

Scheme lc_t_ind' := Induction for lc_t Sort Prop.

Definition lc_t_mutind :=
  fun H H1 H2 H3 H4 =>
  lc_t_ind' H H1 H2 H3 H4.

Scheme lc_t_rec' := Induction for lc_t Sort Set.

Definition lc_t_mutrec :=
  fun H H1 H2 H3 H4 =>
  lc_t_rec' H H1 H2 H3 H4.

Hint Constructors lc_t : core lngen.

Inductive lc_e : exp -> Set :=
  | lc_e_var : forall x,
    lc_e (exp_var x)
  | lc_e_abs : forall t e,
    lc_t t ->
    (forall x : expvar, lc_e (open_ex (exp_var x) e)) ->
    lc_e (exp_abs t e)
  | lc_e_app : forall e e1,
    lc_e e ->
    lc_e e1 ->
    lc_e (exp_app e e1)
  | lc_e_tabs : forall t e,
    lc_t t ->
    (forall X : typvar, lc_e (open_eX (typ_var X) e)) ->
    lc_e (exp_tabs t e)
  | lc_e_tapp : forall e t,
    lc_e e ->
    lc_t t ->
    lc_e (exp_tapp e t).

Scheme lc_e_ind' := Induction for lc_e Sort Prop.

Definition lc_e_mutind :=
  fun H H1 H2 H3 H4 H5 =>
  lc_e_ind' H H1 H2 H3 H4 H5.

Scheme lc_e_rec' := Induction for lc_e Sort Set.

Definition lc_e_mutrec :=
  fun H H1 H2 H3 H4 H5 =>
  lc_e_rec' H H1 H2 H3 H4 H5.

Hint Constructors lc_e : core lngen.


(* *********************************************************************** *)
(** * Tactic support *)

(** Tactic definitions. *)

Ltac gather_strings :=
  let S := ltac_map (fun x : string => singleton x) in
  let S1 := ltac_map (fun x : fset string => x) in
  let S2 := ltac_map (fun x : exp => fv_ex x) in
  let S3 := ltac_map (fun x : exp => fv_eX x) in
  let S4 := ltac_map (fun x : typ => fv_tX x) in
  simplify_list_of_atom_sets (S ++ S1 ++ S2 ++ S3 ++ S4).

Ltac pick_fresh_typvar x :=
  let L := gather_strings in
  (pick fresh x for L).

Tactic Notation "pick" "fresh" "typvar" ident(x) :=
  let L := gather_strings in
  (pick fresh x for L).

Ltac pick_fresh_expvar x :=
  let L := gather_strings in
  (pick fresh x for L).

Tactic Notation "pick" "fresh" "expvar" ident(x) :=
  let L := gather_strings in
  (pick fresh x for L).

(** Additional hint declarations. *)

Hint Resolve @plus_le_compat : lngen. (* from Coq.Arith.Plus *)

(** Redefine [default_auto] and [default_autorewrite]. *)

Ltac default_auto ::= auto with arith set brute_force lngen.
Ltac default_autorewrite ::= autorewrite with lngen in *.


(* *********************************************************************** *)
(** * Theorems about [size] *)

(* begin hide *)

Lemma size_t_min_mutual :
(forall t, 1 <= size_t t).
Proof.
apply_mutual_ind typ_mutind;
default_simp.
Qed.

(* end hide *)

Lemma size_t_min :
forall t, 1 <= size_t t.
Proof.
pose proof size_t_min_mutual as H; intuition eauto.
Qed.

Hint Resolve size_t_min : lngen.

(* begin hide *)

Lemma size_e_min_mutual :
(forall e, 1 <= size_e e).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma size_e_min :
forall e, 1 <= size_e e.
Proof.
pose proof size_e_min_mutual as H; intuition eauto.
Qed.

Hint Resolve size_e_min : lngen.

(* begin hide *)

Lemma size_t_close_tX_rec_mutual :
(forall t X n,
  size_t (close_tX_rec n X t) = size_t t).
Proof.
apply_mutual_ind typ_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_t_close_tX_rec :
forall t X n,
  size_t (close_tX_rec n X t) = size_t t.
Proof.
pose proof size_t_close_tX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_t_close_tX_rec : lngen.
Hint Rewrite size_t_close_tX_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma size_e_close_ex_rec_mutual :
(forall e x n,
  size_e (close_ex_rec n x e) = size_e e).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_e_close_ex_rec :
forall e x n,
  size_e (close_ex_rec n x e) = size_e e.
Proof.
pose proof size_e_close_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_e_close_ex_rec : lngen.
Hint Rewrite size_e_close_ex_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma size_e_close_eX_rec_mutual :
(forall e X n,
  size_e (close_eX_rec n X e) = size_e e).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_e_close_eX_rec :
forall e X n,
  size_e (close_eX_rec n X e) = size_e e.
Proof.
pose proof size_e_close_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_e_close_eX_rec : lngen.
Hint Rewrite size_e_close_eX_rec using solve [auto] : lngen.

(* end hide *)

Lemma size_t_close_tX :
forall t X,
  size_t (close_tX X t) = size_t t.
Proof.
unfold close_tX; default_simp.
Qed.

Hint Resolve size_t_close_tX : lngen.
Hint Rewrite size_t_close_tX using solve [auto] : lngen.

Lemma size_e_close_ex :
forall e x,
  size_e (close_ex x e) = size_e e.
Proof.
unfold close_ex; default_simp.
Qed.

Hint Resolve size_e_close_ex : lngen.
Hint Rewrite size_e_close_ex using solve [auto] : lngen.

Lemma size_e_close_eX :
forall e X,
  size_e (close_eX X e) = size_e e.
Proof.
unfold close_eX; default_simp.
Qed.

Hint Resolve size_e_close_eX : lngen.
Hint Rewrite size_e_close_eX using solve [auto] : lngen.

(* begin hide *)

Lemma size_t_open_tX_rec_mutual :
(forall t t1 n,
  size_t t <= size_t (open_tX_rec n t1 t)).
Proof.
apply_mutual_ind typ_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_t_open_tX_rec :
forall t t1 n,
  size_t t <= size_t (open_tX_rec n t1 t).
Proof.
pose proof size_t_open_tX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_t_open_tX_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma size_e_open_ex_rec_mutual :
(forall e e1 n,
  size_e e <= size_e (open_ex_rec n e1 e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_e_open_ex_rec :
forall e e1 n,
  size_e e <= size_e (open_ex_rec n e1 e).
Proof.
pose proof size_e_open_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_e_open_ex_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma size_e_open_eX_rec_mutual :
(forall e t n,
  size_e e <= size_e (open_eX_rec n t e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_e_open_eX_rec :
forall e t n,
  size_e e <= size_e (open_eX_rec n t e).
Proof.
pose proof size_e_open_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_e_open_eX_rec : lngen.

(* end hide *)

Lemma size_t_open_tX :
forall t t1,
  size_t t <= size_t (open_tX t1 t).
Proof.
unfold open_tX; default_simp.
Qed.

Hint Resolve size_t_open_tX : lngen.

Lemma size_e_open_ex :
forall e e1,
  size_e e <= size_e (open_ex e1 e).
Proof.
unfold open_ex; default_simp.
Qed.

Hint Resolve size_e_open_ex : lngen.

Lemma size_e_open_eX :
forall e t,
  size_e e <= size_e (open_eX t e).
Proof.
unfold open_eX; default_simp.
Qed.

Hint Resolve size_e_open_eX : lngen.

(* begin hide *)

Lemma size_t_open_tX_rec_var_mutual :
(forall t X n,
  size_t (open_tX_rec n (typ_var X) t) = size_t t).
Proof.
apply_mutual_ind typ_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_t_open_tX_rec_var :
forall t X n,
  size_t (open_tX_rec n (typ_var X) t) = size_t t.
Proof.
pose proof size_t_open_tX_rec_var_mutual as H; intuition eauto.
Qed.

Hint Resolve size_t_open_tX_rec_var : lngen.
Hint Rewrite size_t_open_tX_rec_var using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma size_e_open_ex_rec_var_mutual :
(forall e x n,
  size_e (open_ex_rec n (exp_var x) e) = size_e e).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_e_open_ex_rec_var :
forall e x n,
  size_e (open_ex_rec n (exp_var x) e) = size_e e.
Proof.
pose proof size_e_open_ex_rec_var_mutual as H; intuition eauto.
Qed.

Hint Resolve size_e_open_ex_rec_var : lngen.
Hint Rewrite size_e_open_ex_rec_var using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma size_e_open_eX_rec_var_mutual :
(forall e X n,
  size_e (open_eX_rec n (typ_var X) e) = size_e e).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_e_open_eX_rec_var :
forall e X n,
  size_e (open_eX_rec n (typ_var X) e) = size_e e.
Proof.
pose proof size_e_open_eX_rec_var_mutual as H; intuition eauto.
Qed.

Hint Resolve size_e_open_eX_rec_var : lngen.
Hint Rewrite size_e_open_eX_rec_var using solve [auto] : lngen.

(* end hide *)

Lemma size_t_open_tX_var :
forall t X,
  size_t (open_tX (typ_var X) t) = size_t t.
Proof.
unfold open_tX; default_simp.
Qed.

Hint Resolve size_t_open_tX_var : lngen.
Hint Rewrite size_t_open_tX_var using solve [auto] : lngen.

Lemma size_e_open_ex_var :
forall e x,
  size_e (open_ex (exp_var x) e) = size_e e.
Proof.
unfold open_ex; default_simp.
Qed.

Hint Resolve size_e_open_ex_var : lngen.
Hint Rewrite size_e_open_ex_var using solve [auto] : lngen.

Lemma size_e_open_eX_var :
forall e X,
  size_e (open_eX (typ_var X) e) = size_e e.
Proof.
unfold open_eX; default_simp.
Qed.

Hint Resolve size_e_open_eX_var : lngen.
Hint Rewrite size_e_open_eX_var using solve [auto] : lngen.


(* *********************************************************************** *)
(** * Theorems about [degree] *)

(* begin hide *)

Lemma degree_tX_S_mutual :
(forall n t,
  degree_tX n t ->
  degree_tX (S n) t).
Proof.
apply_mutual_ind degree_tX_mutrec;
default_simp.
Qed.

(* end hide *)

Lemma degree_tX_S :
forall n t,
  degree_tX n t ->
  degree_tX (S n) t.
Proof.
pose proof degree_tX_S_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_tX_S : lngen.

(* begin hide *)

Lemma degree_ex_S_mutual :
(forall n e,
  degree_ex n e ->
  degree_ex (S n) e).
Proof.
apply_mutual_ind degree_ex_mutrec;
default_simp.
Qed.

(* end hide *)

Lemma degree_ex_S :
forall n e,
  degree_ex n e ->
  degree_ex (S n) e.
Proof.
pose proof degree_ex_S_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_ex_S : lngen.

(* begin hide *)

Lemma degree_eX_S_mutual :
(forall n e,
  degree_eX n e ->
  degree_eX (S n) e).
Proof.
apply_mutual_ind degree_eX_mutrec;
default_simp.
Qed.

(* end hide *)

Lemma degree_eX_S :
forall n e,
  degree_eX n e ->
  degree_eX (S n) e.
Proof.
pose proof degree_eX_S_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_eX_S : lngen.

Lemma degree_tX_O :
forall n t,
  degree_tX O t ->
  degree_tX n t.
Proof.
induction n; default_simp.
Qed.

Hint Resolve degree_tX_O : lngen.

Lemma degree_ex_O :
forall n e,
  degree_ex O e ->
  degree_ex n e.
Proof.
induction n; default_simp.
Qed.

Hint Resolve degree_ex_O : lngen.

Lemma degree_eX_O :
forall n e,
  degree_eX O e ->
  degree_eX n e.
Proof.
induction n; default_simp.
Qed.

Hint Resolve degree_eX_O : lngen.

(* begin hide *)

Lemma degree_tX_close_tX_rec_mutual :
(forall t X n,
  degree_tX n t ->
  degree_tX (S n) (close_tX_rec n X t)).
Proof.
apply_mutual_ind typ_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_tX_close_tX_rec :
forall t X n,
  degree_tX n t ->
  degree_tX (S n) (close_tX_rec n X t).
Proof.
pose proof degree_tX_close_tX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_tX_close_tX_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_ex_close_ex_rec_mutual :
(forall e x n,
  degree_ex n e ->
  degree_ex (S n) (close_ex_rec n x e)).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_ex_close_ex_rec :
forall e x n,
  degree_ex n e ->
  degree_ex (S n) (close_ex_rec n x e).
Proof.
pose proof degree_ex_close_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_ex_close_ex_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_ex_close_eX_rec_mutual :
(forall e X n n1,
  degree_ex n e ->
  degree_ex n (close_eX_rec n1 X e)).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_ex_close_eX_rec :
forall e X n n1,
  degree_ex n e ->
  degree_ex n (close_eX_rec n1 X e).
Proof.
pose proof degree_ex_close_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_ex_close_eX_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_eX_close_ex_rec_mutual :
(forall e x n n1,
  degree_eX n e ->
  degree_eX n (close_ex_rec n1 x e)).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_eX_close_ex_rec :
forall e x n n1,
  degree_eX n e ->
  degree_eX n (close_ex_rec n1 x e).
Proof.
pose proof degree_eX_close_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_eX_close_ex_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_eX_close_eX_rec_mutual :
(forall e X n,
  degree_eX n e ->
  degree_eX (S n) (close_eX_rec n X e)).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_eX_close_eX_rec :
forall e X n,
  degree_eX n e ->
  degree_eX (S n) (close_eX_rec n X e).
Proof.
pose proof degree_eX_close_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_eX_close_eX_rec : lngen.

(* end hide *)

Lemma degree_tX_close_tX :
forall t X,
  degree_tX 0 t ->
  degree_tX 1 (close_tX X t).
Proof.
unfold close_tX; default_simp.
Qed.

Hint Resolve degree_tX_close_tX : lngen.

Lemma degree_ex_close_ex :
forall e x,
  degree_ex 0 e ->
  degree_ex 1 (close_ex x e).
Proof.
unfold close_ex; default_simp.
Qed.

Hint Resolve degree_ex_close_ex : lngen.

Lemma degree_ex_close_eX :
forall e X n,
  degree_ex n e ->
  degree_ex n (close_eX X e).
Proof.
unfold close_eX; default_simp.
Qed.

Hint Resolve degree_ex_close_eX : lngen.

Lemma degree_eX_close_ex :
forall e x n,
  degree_eX n e ->
  degree_eX n (close_ex x e).
Proof.
unfold close_ex; default_simp.
Qed.

Hint Resolve degree_eX_close_ex : lngen.

Lemma degree_eX_close_eX :
forall e X,
  degree_eX 0 e ->
  degree_eX 1 (close_eX X e).
Proof.
unfold close_eX; default_simp.
Qed.

Hint Resolve degree_eX_close_eX : lngen.

(* begin hide *)

Lemma degree_tX_close_tX_rec_inv_mutual :
(forall t X n,
  degree_tX (S n) (close_tX_rec n X t) ->
  degree_tX n t).
Proof.
apply_mutual_ind typ_mutrec;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_tX_close_tX_rec_inv :
forall t X n,
  degree_tX (S n) (close_tX_rec n X t) ->
  degree_tX n t.
Proof.
pose proof degree_tX_close_tX_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_tX_close_tX_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_ex_close_ex_rec_inv_mutual :
(forall e x n,
  degree_ex (S n) (close_ex_rec n x e) ->
  degree_ex n e).
Proof.
apply_mutual_ind exp_mutrec;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_ex_close_ex_rec_inv :
forall e x n,
  degree_ex (S n) (close_ex_rec n x e) ->
  degree_ex n e.
Proof.
pose proof degree_ex_close_ex_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_ex_close_ex_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_ex_close_eX_rec_inv_mutual :
(forall e X n n1,
  degree_ex n (close_eX_rec n1 X e) ->
  degree_ex n e).
Proof.
apply_mutual_ind exp_mutrec;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_ex_close_eX_rec_inv :
forall e X n n1,
  degree_ex n (close_eX_rec n1 X e) ->
  degree_ex n e.
Proof.
pose proof degree_ex_close_eX_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_ex_close_eX_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_eX_close_ex_rec_inv_mutual :
(forall e x n n1,
  degree_eX n (close_ex_rec n1 x e) ->
  degree_eX n e).
Proof.
apply_mutual_ind exp_mutrec;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_eX_close_ex_rec_inv :
forall e x n n1,
  degree_eX n (close_ex_rec n1 x e) ->
  degree_eX n e.
Proof.
pose proof degree_eX_close_ex_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_eX_close_ex_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_eX_close_eX_rec_inv_mutual :
(forall e X n,
  degree_eX (S n) (close_eX_rec n X e) ->
  degree_eX n e).
Proof.
apply_mutual_ind exp_mutrec;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_eX_close_eX_rec_inv :
forall e X n,
  degree_eX (S n) (close_eX_rec n X e) ->
  degree_eX n e.
Proof.
pose proof degree_eX_close_eX_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_eX_close_eX_rec_inv : lngen.

(* end hide *)

Lemma degree_tX_close_tX_inv :
forall t X,
  degree_tX 1 (close_tX X t) ->
  degree_tX 0 t.
Proof.
unfold close_tX; eauto 2 with lngen.
Qed.

Hint Immediate degree_tX_close_tX_inv : lngen.

Lemma degree_ex_close_ex_inv :
forall e x,
  degree_ex 1 (close_ex x e) ->
  degree_ex 0 e.
Proof.
unfold close_ex; eauto 2 with lngen.
Qed.

Hint Immediate degree_ex_close_ex_inv : lngen.

Lemma degree_ex_close_eX_inv :
forall e X n,
  degree_ex n (close_eX X e) ->
  degree_ex n e.
Proof.
unfold close_eX; eauto 2 with lngen.
Qed.

Hint Immediate degree_ex_close_eX_inv : lngen.

Lemma degree_eX_close_ex_inv :
forall e x n,
  degree_eX n (close_ex x e) ->
  degree_eX n e.
Proof.
unfold close_ex; eauto 2 with lngen.
Qed.

Hint Immediate degree_eX_close_ex_inv : lngen.

Lemma degree_eX_close_eX_inv :
forall e X,
  degree_eX 1 (close_eX X e) ->
  degree_eX 0 e.
Proof.
unfold close_eX; eauto 2 with lngen.
Qed.

Hint Immediate degree_eX_close_eX_inv : lngen.

(* begin hide *)

Lemma degree_tX_open_tX_rec_mutual :
(forall t t1 n,
  degree_tX (S n) t ->
  degree_tX n t1 ->
  degree_tX n (open_tX_rec n t1 t)).
Proof.
apply_mutual_ind typ_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_tX_open_tX_rec :
forall t t1 n,
  degree_tX (S n) t ->
  degree_tX n t1 ->
  degree_tX n (open_tX_rec n t1 t).
Proof.
pose proof degree_tX_open_tX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_tX_open_tX_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_ex_open_ex_rec_mutual :
(forall e e1 n,
  degree_ex (S n) e ->
  degree_ex n e1 ->
  degree_ex n (open_ex_rec n e1 e)).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_ex_open_ex_rec :
forall e e1 n,
  degree_ex (S n) e ->
  degree_ex n e1 ->
  degree_ex n (open_ex_rec n e1 e).
Proof.
pose proof degree_ex_open_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_ex_open_ex_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_ex_open_eX_rec_mutual :
(forall e t n n1,
  degree_ex n e ->
  degree_ex n (open_eX_rec n1 t e)).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_ex_open_eX_rec :
forall e t n n1,
  degree_ex n e ->
  degree_ex n (open_eX_rec n1 t e).
Proof.
pose proof degree_ex_open_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_ex_open_eX_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_eX_open_ex_rec_mutual :
(forall e e1 n n1,
  degree_eX n e ->
  degree_eX n e1 ->
  degree_eX n (open_ex_rec n1 e1 e)).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_eX_open_ex_rec :
forall e e1 n n1,
  degree_eX n e ->
  degree_eX n e1 ->
  degree_eX n (open_ex_rec n1 e1 e).
Proof.
pose proof degree_eX_open_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_eX_open_ex_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_eX_open_eX_rec_mutual :
(forall e t n,
  degree_eX (S n) e ->
  degree_tX n t ->
  degree_eX n (open_eX_rec n t e)).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_eX_open_eX_rec :
forall e t n,
  degree_eX (S n) e ->
  degree_tX n t ->
  degree_eX n (open_eX_rec n t e).
Proof.
pose proof degree_eX_open_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_eX_open_eX_rec : lngen.

(* end hide *)

Lemma degree_tX_open_tX :
forall t t1,
  degree_tX 1 t ->
  degree_tX 0 t1 ->
  degree_tX 0 (open_tX t1 t).
Proof.
unfold open_tX; default_simp.
Qed.

Hint Resolve degree_tX_open_tX : lngen.

Lemma degree_ex_open_ex :
forall e e1,
  degree_ex 1 e ->
  degree_ex 0 e1 ->
  degree_ex 0 (open_ex e1 e).
Proof.
unfold open_ex; default_simp.
Qed.

Hint Resolve degree_ex_open_ex : lngen.

Lemma degree_ex_open_eX :
forall e t n,
  degree_ex n e ->
  degree_ex n (open_eX t e).
Proof.
unfold open_eX; default_simp.
Qed.

Hint Resolve degree_ex_open_eX : lngen.

Lemma degree_eX_open_ex :
forall e e1 n,
  degree_eX n e ->
  degree_eX n e1 ->
  degree_eX n (open_ex e1 e).
Proof.
unfold open_ex; default_simp.
Qed.

Hint Resolve degree_eX_open_ex : lngen.

Lemma degree_eX_open_eX :
forall e t,
  degree_eX 1 e ->
  degree_tX 0 t ->
  degree_eX 0 (open_eX t e).
Proof.
unfold open_eX; default_simp.
Qed.

Hint Resolve degree_eX_open_eX : lngen.

(* begin hide *)

Lemma degree_tX_open_tX_rec_inv_mutual :
(forall t t1 n,
  degree_tX n (open_tX_rec n t1 t) ->
  degree_tX (S n) t).
Proof.
apply_mutual_ind typ_mutrec;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_tX_open_tX_rec_inv :
forall t t1 n,
  degree_tX n (open_tX_rec n t1 t) ->
  degree_tX (S n) t.
Proof.
pose proof degree_tX_open_tX_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_tX_open_tX_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_ex_open_ex_rec_inv_mutual :
(forall e e1 n,
  degree_ex n (open_ex_rec n e1 e) ->
  degree_ex (S n) e).
Proof.
apply_mutual_ind exp_mutrec;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_ex_open_ex_rec_inv :
forall e e1 n,
  degree_ex n (open_ex_rec n e1 e) ->
  degree_ex (S n) e.
Proof.
pose proof degree_ex_open_ex_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_ex_open_ex_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_ex_open_eX_rec_inv_mutual :
(forall e t n n1,
  degree_ex n (open_eX_rec n1 t e) ->
  degree_ex n e).
Proof.
apply_mutual_ind exp_mutrec;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_ex_open_eX_rec_inv :
forall e t n n1,
  degree_ex n (open_eX_rec n1 t e) ->
  degree_ex n e.
Proof.
pose proof degree_ex_open_eX_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_ex_open_eX_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_eX_open_ex_rec_inv_mutual :
(forall e e1 n n1,
  degree_eX n (open_ex_rec n1 e1 e) ->
  degree_eX n e).
Proof.
apply_mutual_ind exp_mutrec;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_eX_open_ex_rec_inv :
forall e e1 n n1,
  degree_eX n (open_ex_rec n1 e1 e) ->
  degree_eX n e.
Proof.
pose proof degree_eX_open_ex_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_eX_open_ex_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_eX_open_eX_rec_inv_mutual :
(forall e t n,
  degree_eX n (open_eX_rec n t e) ->
  degree_eX (S n) e).
Proof.
apply_mutual_ind exp_mutrec;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_eX_open_eX_rec_inv :
forall e t n,
  degree_eX n (open_eX_rec n t e) ->
  degree_eX (S n) e.
Proof.
pose proof degree_eX_open_eX_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_eX_open_eX_rec_inv : lngen.

(* end hide *)

Lemma degree_tX_open_tX_inv :
forall t t1,
  degree_tX 0 (open_tX t1 t) ->
  degree_tX 1 t.
Proof.
unfold open_tX; eauto 2 with lngen.
Qed.

Hint Resolve degree_tX_open_tX_inv : lngen.

Lemma degree_ex_open_ex_inv :
forall e e1,
  degree_ex 0 (open_ex e1 e) ->
  degree_ex 1 e.
Proof.
unfold open_ex; eauto 2 with lngen.
Qed.

Hint Resolve degree_ex_open_ex_inv : lngen.

Lemma degree_ex_open_eX_inv :
forall e t n,
  degree_ex n (open_eX t e) ->
  degree_ex n e.
Proof.
unfold open_eX; eauto 2 with lngen.
Qed.

Hint Resolve degree_ex_open_eX_inv : lngen.

Lemma degree_eX_open_ex_inv :
forall e e1 n,
  degree_eX n (open_ex e1 e) ->
  degree_eX n e.
Proof.
unfold open_ex; eauto 2 with lngen.
Qed.

Hint Resolve degree_eX_open_ex_inv : lngen.

Lemma degree_eX_open_eX_inv :
forall e t,
  degree_eX 0 (open_eX t e) ->
  degree_eX 1 e.
Proof.
unfold open_eX; eauto 2 with lngen.
Qed.

Hint Resolve degree_eX_open_eX_inv : lngen.


(* *********************************************************************** *)
(** * Theorems about [open] and [close] *)

(* begin hide *)

Lemma close_tX_rec_inj_mutual :
(forall t1 t X n,
  close_tX_rec n X t1 = close_tX_rec n X t ->
  t1 = t).
Proof.
apply_mutual_ind typ_mutind;
intros; (destruct t);
default_simp; eauto 2 with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_tX_rec_inj :
forall t1 t X n,
  close_tX_rec n X t1 = close_tX_rec n X t ->
  t1 = t.
Proof.
pose proof close_tX_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate close_tX_rec_inj : lngen.

(* end hide *)

(* begin hide *)

Lemma close_ex_rec_inj_mutual :
(forall e1 e x n,
  close_ex_rec n x e1 = close_ex_rec n x e ->
  e1 = e).
Proof.
apply_mutual_ind exp_mutind;
intros; (destruct e);
default_simp; eauto 2 with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_ex_rec_inj :
forall e1 e x n,
  close_ex_rec n x e1 = close_ex_rec n x e ->
  e1 = e.
Proof.
pose proof close_ex_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate close_ex_rec_inj : lngen.

(* end hide *)

(* begin hide *)

Lemma close_eX_rec_inj_mutual :
(forall e1 e X n,
  close_eX_rec n X e1 = close_eX_rec n X e ->
  e1 = e).
Proof.
apply_mutual_ind exp_mutind;
intros; (destruct e);
default_simp; eauto 2 with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_eX_rec_inj :
forall e1 e X n,
  close_eX_rec n X e1 = close_eX_rec n X e ->
  e1 = e.
Proof.
pose proof close_eX_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate close_eX_rec_inj : lngen.

(* end hide *)

Lemma close_tX_inj :
forall t1 t X,
  close_tX X t1 = close_tX X t ->
  t1 = t.
Proof.
unfold close_tX; eauto 2 with lngen.
Qed.

Hint Immediate close_tX_inj : lngen.

Lemma close_ex_inj :
forall e1 e x,
  close_ex x e1 = close_ex x e ->
  e1 = e.
Proof.
unfold close_ex; eauto 2 with lngen.
Qed.

Hint Immediate close_ex_inj : lngen.

Lemma close_eX_inj :
forall e1 e X,
  close_eX X e1 = close_eX X e ->
  e1 = e.
Proof.
unfold close_eX; eauto 2 with lngen.
Qed.

Hint Immediate close_eX_inj : lngen.

(* begin hide *)

Lemma close_tX_rec_open_tX_rec_mutual :
(forall t X n,
  X `notin` fv_tX t ->
  close_tX_rec n X (open_tX_rec n (typ_var X) t) = t).
Proof.
apply_mutual_ind typ_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_tX_rec_open_tX_rec :
forall t X n,
  X `notin` fv_tX t ->
  close_tX_rec n X (open_tX_rec n (typ_var X) t) = t.
Proof.
pose proof close_tX_rec_open_tX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve close_tX_rec_open_tX_rec : lngen.
Hint Rewrite close_tX_rec_open_tX_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma close_ex_rec_open_ex_rec_mutual :
(forall e x n,
  x `notin` fv_ex e ->
  close_ex_rec n x (open_ex_rec n (exp_var x) e) = e).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_ex_rec_open_ex_rec :
forall e x n,
  x `notin` fv_ex e ->
  close_ex_rec n x (open_ex_rec n (exp_var x) e) = e.
Proof.
pose proof close_ex_rec_open_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve close_ex_rec_open_ex_rec : lngen.
Hint Rewrite close_ex_rec_open_ex_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma close_eX_rec_open_eX_rec_mutual :
(forall e X n,
  X `notin` fv_eX e ->
  close_eX_rec n X (open_eX_rec n (typ_var X) e) = e).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_eX_rec_open_eX_rec :
forall e X n,
  X `notin` fv_eX e ->
  close_eX_rec n X (open_eX_rec n (typ_var X) e) = e.
Proof.
pose proof close_eX_rec_open_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve close_eX_rec_open_eX_rec : lngen.
Hint Rewrite close_eX_rec_open_eX_rec using solve [auto] : lngen.

(* end hide *)

Lemma close_tX_open_tX :
forall t X,
  X `notin` fv_tX t ->
  close_tX X (open_tX (typ_var X) t) = t.
Proof.
unfold close_tX; unfold open_tX; default_simp.
Qed.

Hint Resolve close_tX_open_tX : lngen.
Hint Rewrite close_tX_open_tX using solve [auto] : lngen.

Lemma close_ex_open_ex :
forall e x,
  x `notin` fv_ex e ->
  close_ex x (open_ex (exp_var x) e) = e.
Proof.
unfold close_ex; unfold open_ex; default_simp.
Qed.

Hint Resolve close_ex_open_ex : lngen.
Hint Rewrite close_ex_open_ex using solve [auto] : lngen.

Lemma close_eX_open_eX :
forall e X,
  X `notin` fv_eX e ->
  close_eX X (open_eX (typ_var X) e) = e.
Proof.
unfold close_eX; unfold open_eX; default_simp.
Qed.

Hint Resolve close_eX_open_eX : lngen.
Hint Rewrite close_eX_open_eX using solve [auto] : lngen.

(* begin hide *)

Lemma open_tX_rec_close_tX_rec_mutual :
(forall t X n,
  open_tX_rec n (typ_var X) (close_tX_rec n X t) = t).
Proof.
apply_mutual_ind typ_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_tX_rec_close_tX_rec :
forall t X n,
  open_tX_rec n (typ_var X) (close_tX_rec n X t) = t.
Proof.
pose proof open_tX_rec_close_tX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve open_tX_rec_close_tX_rec : lngen.
Hint Rewrite open_tX_rec_close_tX_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma open_ex_rec_close_ex_rec_mutual :
(forall e x n,
  open_ex_rec n (exp_var x) (close_ex_rec n x e) = e).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_ex_rec_close_ex_rec :
forall e x n,
  open_ex_rec n (exp_var x) (close_ex_rec n x e) = e.
Proof.
pose proof open_ex_rec_close_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve open_ex_rec_close_ex_rec : lngen.
Hint Rewrite open_ex_rec_close_ex_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma open_eX_rec_close_eX_rec_mutual :
(forall e X n,
  open_eX_rec n (typ_var X) (close_eX_rec n X e) = e).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_eX_rec_close_eX_rec :
forall e X n,
  open_eX_rec n (typ_var X) (close_eX_rec n X e) = e.
Proof.
pose proof open_eX_rec_close_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve open_eX_rec_close_eX_rec : lngen.
Hint Rewrite open_eX_rec_close_eX_rec using solve [auto] : lngen.

(* end hide *)

Lemma open_tX_close_tX :
forall t X,
  open_tX (typ_var X) (close_tX X t) = t.
Proof.
unfold close_tX; unfold open_tX; default_simp.
Qed.

Hint Resolve open_tX_close_tX : lngen.
Hint Rewrite open_tX_close_tX using solve [auto] : lngen.

Lemma open_ex_close_ex :
forall e x,
  open_ex (exp_var x) (close_ex x e) = e.
Proof.
unfold close_ex; unfold open_ex; default_simp.
Qed.

Hint Resolve open_ex_close_ex : lngen.
Hint Rewrite open_ex_close_ex using solve [auto] : lngen.

Lemma open_eX_close_eX :
forall e X,
  open_eX (typ_var X) (close_eX X e) = e.
Proof.
unfold close_eX; unfold open_eX; default_simp.
Qed.

Hint Resolve open_eX_close_eX : lngen.
Hint Rewrite open_eX_close_eX using solve [auto] : lngen.

(* begin hide *)

Lemma open_tX_rec_inj_mutual :
(forall t1 t X n,
  X `notin` fv_tX t1 ->
  X `notin` fv_tX t ->
  open_tX_rec n (typ_var X) t1 = open_tX_rec n (typ_var X) t ->
  t1 = t).
Proof.
apply_mutual_ind typ_mutind;
intros; (destruct t);
default_simp; eauto 2 with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_tX_rec_inj :
forall t1 t X n,
  X `notin` fv_tX t1 ->
  X `notin` fv_tX t ->
  open_tX_rec n (typ_var X) t1 = open_tX_rec n (typ_var X) t ->
  t1 = t.
Proof.
pose proof open_tX_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate open_tX_rec_inj : lngen.

(* end hide *)

(* begin hide *)

Lemma open_ex_rec_inj_mutual :
(forall e1 e x n,
  x `notin` fv_ex e1 ->
  x `notin` fv_ex e ->
  open_ex_rec n (exp_var x) e1 = open_ex_rec n (exp_var x) e ->
  e1 = e).
Proof.
apply_mutual_ind exp_mutind;
intros; (destruct e);
default_simp; eauto 2 with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_ex_rec_inj :
forall e1 e x n,
  x `notin` fv_ex e1 ->
  x `notin` fv_ex e ->
  open_ex_rec n (exp_var x) e1 = open_ex_rec n (exp_var x) e ->
  e1 = e.
Proof.
pose proof open_ex_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate open_ex_rec_inj : lngen.

(* end hide *)

(* begin hide *)

Lemma open_eX_rec_inj_mutual :
(forall e1 e X n,
  X `notin` fv_eX e1 ->
  X `notin` fv_eX e ->
  open_eX_rec n (typ_var X) e1 = open_eX_rec n (typ_var X) e ->
  e1 = e).
Proof.
apply_mutual_ind exp_mutind;
intros; (destruct e);
default_simp; eauto 2 with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_eX_rec_inj :
forall e1 e X n,
  X `notin` fv_eX e1 ->
  X `notin` fv_eX e ->
  open_eX_rec n (typ_var X) e1 = open_eX_rec n (typ_var X) e ->
  e1 = e.
Proof.
pose proof open_eX_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate open_eX_rec_inj : lngen.

(* end hide *)

Lemma open_tX_inj :
forall t1 t X,
  X `notin` fv_tX t1 ->
  X `notin` fv_tX t ->
  open_tX (typ_var X) t1 = open_tX (typ_var X) t ->
  t1 = t.
Proof.
unfold open_tX; eauto 2 with lngen.
Qed.

Hint Immediate open_tX_inj : lngen.

Lemma open_ex_inj :
forall e1 e x,
  x `notin` fv_ex e1 ->
  x `notin` fv_ex e ->
  open_ex (exp_var x) e1 = open_ex (exp_var x) e ->
  e1 = e.
Proof.
unfold open_ex; eauto 2 with lngen.
Qed.

Hint Immediate open_ex_inj : lngen.

Lemma open_eX_inj :
forall e1 e X,
  X `notin` fv_eX e1 ->
  X `notin` fv_eX e ->
  open_eX (typ_var X) e1 = open_eX (typ_var X) e ->
  e1 = e.
Proof.
unfold open_eX; eauto 2 with lngen.
Qed.

Hint Immediate open_eX_inj : lngen.


(* *********************************************************************** *)
(** * Theorems about [lc] *)

(* begin hide *)

Lemma degree_tX_of_lc_t_mutual :
(forall t,
  lc_t t ->
  degree_tX 0 t).
Proof.
apply_mutual_ind lc_t_mutrec;
intros;
let X := fresh "X" in pick_fresh_typvar X;
repeat (match goal with
          | H : _, H1 : _ |- _ => specialize H with H1
        end);
default_simp; eauto with lngen.
Qed.

(* end hide *)

Lemma degree_tX_of_lc_t :
forall t,
  lc_t t ->
  degree_tX 0 t.
Proof.
pose proof degree_tX_of_lc_t_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_tX_of_lc_t : lngen.

(* begin hide *)

Lemma degree_ex_of_lc_e_mutual :
(forall e,
  lc_e e ->
  degree_ex 0 e).
Proof.
apply_mutual_ind lc_e_mutrec;
intros;
let x := fresh "x" in pick_fresh_expvar x;
let X := fresh "X" in pick_fresh_typvar X;
repeat (match goal with
          | H : _, H1 : _ |- _ => specialize H with H1
        end);
default_simp; eauto with lngen.
Qed.

(* end hide *)

Lemma degree_ex_of_lc_e :
forall e,
  lc_e e ->
  degree_ex 0 e.
Proof.
pose proof degree_ex_of_lc_e_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_ex_of_lc_e : lngen.

(* begin hide *)

Lemma degree_eX_of_lc_e_mutual :
(forall e,
  lc_e e ->
  degree_eX 0 e).
Proof.
apply_mutual_ind lc_e_mutrec;
intros;
let x := fresh "x" in pick_fresh_expvar x;
let X := fresh "X" in pick_fresh_typvar X;
repeat (match goal with
          | H : _, H1 : _ |- _ => specialize H with H1
        end);
default_simp; eauto with lngen.
Qed.

(* end hide *)

Lemma degree_eX_of_lc_e :
forall e,
  lc_e e ->
  degree_eX 0 e.
Proof.
pose proof degree_eX_of_lc_e_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_eX_of_lc_e : lngen.

(* begin hide *)

Lemma lc_t_of_degree_size_mutual :
forall i,
(forall t,
  size_t t = i ->
  degree_tX 0 t ->
  lc_t t).
Proof.
intros i; pattern i; apply lt_wf_rec;
clear i; intros i H;
apply_mutual_ind typ_mutrec;
default_simp;
(* non-trivial cases *)
constructor; default_simp; eapply_first_hyp;
(* instantiate the size *)
match goal with
  | |- _ = _ => reflexivity
  | _ => idtac
end;
(* everything should be easy now *)
default_simp.
Qed.

(* end hide *)

Lemma lc_t_of_degree :
forall t,
  degree_tX 0 t ->
  lc_t t.
Proof.
intros t; intros;
pose proof (lc_t_of_degree_size_mutual (size_t t));
intuition eauto.
Qed.

Hint Resolve lc_t_of_degree : lngen.

(* begin hide *)

Lemma lc_e_of_degree_size_mutual :
forall i,
(forall e,
  size_e e = i ->
  degree_ex 0 e ->
  degree_eX 0 e ->
  lc_e e).
Proof.
intros i; pattern i; apply lt_wf_rec;
clear i; intros i H;
apply_mutual_ind exp_mutrec;
default_simp;
(* non-trivial cases *)
constructor; default_simp; eapply_first_hyp;
(* instantiate the size *)
match goal with
  | |- _ = _ => reflexivity
  | _ => idtac
end;
(* everything should be easy now *)
default_simp.
Qed.

(* end hide *)

Lemma lc_e_of_degree :
forall e,
  degree_ex 0 e ->
  degree_eX 0 e ->
  lc_e e.
Proof.
intros e; intros;
pose proof (lc_e_of_degree_size_mutual (size_e e));
intuition eauto.
Qed.

Hint Resolve lc_e_of_degree : lngen.

Lemma lc_t_all_exists :
forall X t t1,
  lc_t t ->
  lc_t (open_tX (typ_var X) t1) ->
  lc_t (typ_all t t1).
Proof.
default_simp; eauto 10 with lngen.
Qed.

Hint Resolve lc_t_all_exists : lngen.

Lemma lc_e_abs_exists :
forall x t e,
  lc_t t ->
  lc_e (open_ex (exp_var x) e) ->
  lc_e (exp_abs t e).
Proof.
default_simp; eauto 10 with lngen.
Qed.

Hint Resolve lc_e_abs_exists : lngen.

Lemma lc_e_tabs_exists :
forall X t e,
  lc_t t ->
  lc_e (open_eX (typ_var X) e) ->
  lc_e (exp_tabs t e).
Proof.
default_simp; eauto 10 with lngen.
Qed.

Hint Resolve lc_e_tabs_exists : lngen.

(* begin hide *)

Lemma lc_t_unique_mutual :
(forall t (proof proof1 : lc_t t), proof = proof1).
Proof.
apply_mutual_ind lc_t_mutind;
intros; dependent destruction proof1; default_simp; auto using axiom_fun_ext_eq.
Qed.

(* end hide *)

Lemma lc_t_unique :
forall t (proof proof1 : lc_t t), proof = proof1.
Proof.
pose proof lc_t_unique_mutual as H; intuition eauto.
Qed.

Hint Resolve lc_t_unique : lngen.

(* begin hide *)

Lemma lc_e_unique_mutual :
(forall e (proof proof1 : lc_e e), proof = proof1).
Proof.
apply_mutual_ind lc_e_mutind;
intros; dependent destruction proof1; default_simp; auto using axiom_fun_ext_eq.
Qed.

(* end hide *)

Lemma lc_e_unique :
forall e (proof proof1 : lc_e e), proof = proof1.
Proof.
pose proof lc_e_unique_mutual as H; intuition eauto.
Qed.

Hint Resolve lc_e_unique : lngen.


(* *********************************************************************** *)
(** * More theorems about [open] and [close] *)

(* begin hide *)

Lemma close_tX_rec_degree_tX_mutual :
(forall t X n,
  degree_tX n t ->
  X `notin` fv_tX t ->
  close_tX_rec n X t = t).
Proof.
apply_mutual_ind typ_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_tX_rec_degree_tX :
forall t X n,
  degree_tX n t ->
  X `notin` fv_tX t ->
  close_tX_rec n X t = t.
Proof.
pose proof close_tX_rec_degree_tX_mutual as H; intuition eauto.
Qed.

Hint Resolve close_tX_rec_degree_tX : lngen.
Hint Rewrite close_tX_rec_degree_tX using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma close_ex_rec_degree_ex_mutual :
(forall e x n,
  degree_ex n e ->
  x `notin` fv_ex e ->
  close_ex_rec n x e = e).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_ex_rec_degree_ex :
forall e x n,
  degree_ex n e ->
  x `notin` fv_ex e ->
  close_ex_rec n x e = e.
Proof.
pose proof close_ex_rec_degree_ex_mutual as H; intuition eauto.
Qed.

Hint Resolve close_ex_rec_degree_ex : lngen.
Hint Rewrite close_ex_rec_degree_ex using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma close_eX_rec_degree_eX_mutual :
(forall e X n,
  degree_eX n e ->
  X `notin` fv_eX e ->
  close_eX_rec n X e = e).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_eX_rec_degree_eX :
forall e X n,
  degree_eX n e ->
  X `notin` fv_eX e ->
  close_eX_rec n X e = e.
Proof.
pose proof close_eX_rec_degree_eX_mutual as H; intuition eauto.
Qed.

Hint Resolve close_eX_rec_degree_eX : lngen.
Hint Rewrite close_eX_rec_degree_eX using solve [auto] : lngen.

(* end hide *)

Lemma close_tX_lc_t :
forall t X,
  lc_t t ->
  X `notin` fv_tX t ->
  close_tX X t = t.
Proof.
unfold close_tX; default_simp.
Qed.

Hint Resolve close_tX_lc_t : lngen.
Hint Rewrite close_tX_lc_t using solve [auto] : lngen.

Lemma close_ex_lc_e :
forall e x,
  lc_e e ->
  x `notin` fv_ex e ->
  close_ex x e = e.
Proof.
unfold close_ex; default_simp.
Qed.

Hint Resolve close_ex_lc_e : lngen.
Hint Rewrite close_ex_lc_e using solve [auto] : lngen.

Lemma close_eX_lc_e :
forall e X,
  lc_e e ->
  X `notin` fv_eX e ->
  close_eX X e = e.
Proof.
unfold close_eX; default_simp.
Qed.

Hint Resolve close_eX_lc_e : lngen.
Hint Rewrite close_eX_lc_e using solve [auto] : lngen.

(* begin hide *)

Lemma open_tX_rec_degree_tX_mutual :
(forall t1 t n,
  degree_tX n t1 ->
  open_tX_rec n t t1 = t1).
Proof.
apply_mutual_ind typ_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_tX_rec_degree_tX :
forall t1 t n,
  degree_tX n t1 ->
  open_tX_rec n t t1 = t1.
Proof.
pose proof open_tX_rec_degree_tX_mutual as H; intuition eauto.
Qed.

Hint Resolve open_tX_rec_degree_tX : lngen.
Hint Rewrite open_tX_rec_degree_tX using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma open_ex_rec_degree_ex_mutual :
(forall e1 e n,
  degree_ex n e1 ->
  open_ex_rec n e e1 = e1).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_ex_rec_degree_ex :
forall e1 e n,
  degree_ex n e1 ->
  open_ex_rec n e e1 = e1.
Proof.
pose proof open_ex_rec_degree_ex_mutual as H; intuition eauto.
Qed.

Hint Resolve open_ex_rec_degree_ex : lngen.
Hint Rewrite open_ex_rec_degree_ex using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma open_eX_rec_degree_eX_mutual :
(forall e t n,
  degree_eX n e ->
  open_eX_rec n t e = e).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_eX_rec_degree_eX :
forall e t n,
  degree_eX n e ->
  open_eX_rec n t e = e.
Proof.
pose proof open_eX_rec_degree_eX_mutual as H; intuition eauto.
Qed.

Hint Resolve open_eX_rec_degree_eX : lngen.
Hint Rewrite open_eX_rec_degree_eX using solve [auto] : lngen.

(* end hide *)

Lemma open_tX_lc_t :
forall t1 t,
  lc_t t1 ->
  open_tX t t1 = t1.
Proof.
unfold open_tX; default_simp.
Qed.

Hint Resolve open_tX_lc_t : lngen.
Hint Rewrite open_tX_lc_t using solve [auto] : lngen.

Lemma open_ex_lc_e :
forall e1 e,
  lc_e e1 ->
  open_ex e e1 = e1.
Proof.
unfold open_ex; default_simp.
Qed.

Hint Resolve open_ex_lc_e : lngen.
Hint Rewrite open_ex_lc_e using solve [auto] : lngen.

Lemma open_eX_lc_e :
forall e t,
  lc_e e ->
  open_eX t e = e.
Proof.
unfold open_eX; default_simp.
Qed.

Hint Resolve open_eX_lc_e : lngen.
Hint Rewrite open_eX_lc_e using solve [auto] : lngen.


(* *********************************************************************** *)
(** * Theorems about [subst] *)

(* begin hide *)

Lemma subst_tX_close_tX_rec_mutual :
(forall t1 t X X1 n,
  degree_tX n t ->
  X <> X1 ->
  X1 `notin` fv_tX t ->
  subst_tX t X (close_tX_rec n X1 t1) = close_tX_rec n X1 (subst_tX t X t1)).
Proof.
apply_mutual_ind typ_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_tX_close_tX_rec :
forall t1 t X X1 n,
  degree_tX n t ->
  X <> X1 ->
  X1 `notin` fv_tX t ->
  subst_tX t X (close_tX_rec n X1 t1) = close_tX_rec n X1 (subst_tX t X t1).
Proof.
pose proof subst_tX_close_tX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_tX_close_tX_rec : lngen.

(* begin hide *)

Lemma subst_ex_close_ex_rec_mutual :
(forall e1 e x x1 n,
  degree_ex n e ->
  x <> x1 ->
  x1 `notin` fv_ex e ->
  subst_ex e x (close_ex_rec n x1 e1) = close_ex_rec n x1 (subst_ex e x e1)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_ex_close_ex_rec :
forall e1 e x x1 n,
  degree_ex n e ->
  x <> x1 ->
  x1 `notin` fv_ex e ->
  subst_ex e x (close_ex_rec n x1 e1) = close_ex_rec n x1 (subst_ex e x e1).
Proof.
pose proof subst_ex_close_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_ex_close_ex_rec : lngen.

(* begin hide *)

Lemma subst_ex_close_eX_rec_mutual :
(forall e1 e X x n,
  degree_eX n e ->
  x `notin` fv_eX e ->
  subst_ex e X (close_eX_rec n x e1) = close_eX_rec n x (subst_ex e X e1)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_ex_close_eX_rec :
forall e1 e X x n,
  degree_eX n e ->
  x `notin` fv_eX e ->
  subst_ex e X (close_eX_rec n x e1) = close_eX_rec n x (subst_ex e X e1).
Proof.
pose proof subst_ex_close_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_ex_close_eX_rec : lngen.

(* begin hide *)

Lemma subst_eX_close_ex_rec_mutual :
(forall e t x X n,
  subst_eX t x (close_ex_rec n X e) = close_ex_rec n X (subst_eX t x e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_eX_close_ex_rec :
forall e t x X n,
  subst_eX t x (close_ex_rec n X e) = close_ex_rec n X (subst_eX t x e).
Proof.
pose proof subst_eX_close_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_eX_close_ex_rec : lngen.

(* begin hide *)

Lemma subst_eX_close_eX_rec_mutual :
(forall e t X X1 n,
  degree_tX n t ->
  X <> X1 ->
  X1 `notin` fv_tX t ->
  subst_eX t X (close_eX_rec n X1 e) = close_eX_rec n X1 (subst_eX t X e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_eX_close_eX_rec :
forall e t X X1 n,
  degree_tX n t ->
  X <> X1 ->
  X1 `notin` fv_tX t ->
  subst_eX t X (close_eX_rec n X1 e) = close_eX_rec n X1 (subst_eX t X e).
Proof.
pose proof subst_eX_close_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_eX_close_eX_rec : lngen.

Lemma subst_tX_close_tX :
forall t1 t X X1,
  lc_t t ->  X <> X1 ->
  X1 `notin` fv_tX t ->
  subst_tX t X (close_tX X1 t1) = close_tX X1 (subst_tX t X t1).
Proof.
unfold close_tX; default_simp.
Qed.

Hint Resolve subst_tX_close_tX : lngen.

Lemma subst_ex_close_ex :
forall e1 e x x1,
  lc_e e ->  x <> x1 ->
  x1 `notin` fv_ex e ->
  subst_ex e x (close_ex x1 e1) = close_ex x1 (subst_ex e x e1).
Proof.
unfold close_ex; default_simp.
Qed.

Hint Resolve subst_ex_close_ex : lngen.

Lemma subst_ex_close_eX :
forall e1 e X x,
  lc_e e ->  x `notin` fv_eX e ->
  subst_ex e X (close_eX x e1) = close_eX x (subst_ex e X e1).
Proof.
unfold close_eX; default_simp.
Qed.

Hint Resolve subst_ex_close_eX : lngen.

Lemma subst_eX_close_ex :
forall e t x X,
  lc_t t ->  subst_eX t x (close_ex X e) = close_ex X (subst_eX t x e).
Proof.
unfold close_ex; default_simp.
Qed.

Hint Resolve subst_eX_close_ex : lngen.

Lemma subst_eX_close_eX :
forall e t X X1,
  lc_t t ->  X <> X1 ->
  X1 `notin` fv_tX t ->
  subst_eX t X (close_eX X1 e) = close_eX X1 (subst_eX t X e).
Proof.
unfold close_eX; default_simp.
Qed.

Hint Resolve subst_eX_close_eX : lngen.

(* begin hide *)

Lemma subst_tX_degree_tX_mutual :
(forall t t1 X n,
  degree_tX n t ->
  degree_tX n t1 ->
  degree_tX n (subst_tX t1 X t)).
Proof.
apply_mutual_ind typ_mutrec;
default_simp.
Qed.

(* end hide *)

Lemma subst_tX_degree_tX :
forall t t1 X n,
  degree_tX n t ->
  degree_tX n t1 ->
  degree_tX n (subst_tX t1 X t).
Proof.
pose proof subst_tX_degree_tX_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_tX_degree_tX : lngen.

(* begin hide *)

Lemma subst_ex_degree_ex_mutual :
(forall e e1 x n,
  degree_ex n e ->
  degree_ex n e1 ->
  degree_ex n (subst_ex e1 x e)).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

Lemma subst_ex_degree_ex :
forall e e1 x n,
  degree_ex n e ->
  degree_ex n e1 ->
  degree_ex n (subst_ex e1 x e).
Proof.
pose proof subst_ex_degree_ex_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_ex_degree_ex : lngen.

(* begin hide *)

Lemma subst_ex_degree_eX_mutual :
(forall e e1 x n,
  degree_eX n e ->
  degree_eX n e1 ->
  degree_eX n (subst_ex e1 x e)).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

Lemma subst_ex_degree_eX :
forall e e1 x n,
  degree_eX n e ->
  degree_eX n e1 ->
  degree_eX n (subst_ex e1 x e).
Proof.
pose proof subst_ex_degree_eX_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_ex_degree_eX : lngen.

(* begin hide *)

Lemma subst_eX_degree_ex_mutual :
(forall e t X n,
  degree_ex n e ->
  degree_ex n (subst_eX t X e)).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

Lemma subst_eX_degree_ex :
forall e t X n,
  degree_ex n e ->
  degree_ex n (subst_eX t X e).
Proof.
pose proof subst_eX_degree_ex_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_eX_degree_ex : lngen.

(* begin hide *)

Lemma subst_eX_degree_eX_mutual :
(forall e t X n,
  degree_eX n e ->
  degree_tX n t ->
  degree_eX n (subst_eX t X e)).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

Lemma subst_eX_degree_eX :
forall e t X n,
  degree_eX n e ->
  degree_tX n t ->
  degree_eX n (subst_eX t X e).
Proof.
pose proof subst_eX_degree_eX_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_eX_degree_eX : lngen.

(* begin hide *)

Lemma subst_tX_fresh_mutual :
(forall t1 t X,
  X `notin` fv_tX t1 ->
  subst_tX t X t1 = t1).
Proof.
apply_mutual_ind typ_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_tX_fresh :
forall t1 t X,
  X `notin` fv_tX t1 ->
  subst_tX t X t1 = t1.
Proof.
pose proof subst_tX_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_tX_fresh : lngen.
Hint Rewrite subst_tX_fresh using solve [auto] : lngen.

(* begin hide *)

Lemma subst_ex_fresh_mutual :
(forall e1 e x,
  x `notin` fv_ex e1 ->
  subst_ex e x e1 = e1).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_ex_fresh :
forall e1 e x,
  x `notin` fv_ex e1 ->
  subst_ex e x e1 = e1.
Proof.
pose proof subst_ex_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_ex_fresh : lngen.
Hint Rewrite subst_ex_fresh using solve [auto] : lngen.

(* begin hide *)

Lemma subst_eX_fresh_mutual :
(forall e t X,
  X `notin` fv_eX e ->
  subst_eX t X e = e).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_eX_fresh :
forall e t X,
  X `notin` fv_eX e ->
  subst_eX t X e = e.
Proof.
pose proof subst_eX_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_eX_fresh : lngen.
Hint Rewrite subst_eX_fresh using solve [auto] : lngen.

Lemma subst_tX_lc_t :
forall t t1 X,
  lc_t t ->
  lc_t t1 ->
  lc_t (subst_tX t1 X t).
Proof.
default_simp.
Qed.

Hint Resolve subst_tX_lc_t : lngen.

Lemma subst_ex_lc_e :
forall e e1 x,
  lc_e e ->
  lc_e e1 ->
  lc_e (subst_ex e1 x e).
Proof.
default_simp.
Qed.

Hint Resolve subst_ex_lc_e : lngen.

Lemma subst_eX_lc_e :
forall e t X,
  lc_e e ->
  lc_t t ->
  lc_e (subst_eX t X e).
Proof.
default_simp.
Qed.

Hint Resolve subst_eX_lc_e : lngen.

(* begin hide *)

Lemma subst_tX_open_tX_rec_mutual :
(forall t2 t t1 X n,
  lc_t t ->
  subst_tX t X (open_tX_rec n t1 t2) = open_tX_rec n (subst_tX t X t1) (subst_tX t X t2)).
Proof.
apply_mutual_ind typ_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_tX_open_tX_rec :
forall t2 t t1 X n,
  lc_t t ->
  subst_tX t X (open_tX_rec n t1 t2) = open_tX_rec n (subst_tX t X t1) (subst_tX t X t2).
Proof.
pose proof subst_tX_open_tX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_tX_open_tX_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma subst_ex_open_ex_rec_mutual :
(forall e2 e e1 x n,
  lc_e e ->
  subst_ex e x (open_ex_rec n e1 e2) = open_ex_rec n (subst_ex e x e1) (subst_ex e x e2)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_ex_open_ex_rec :
forall e2 e e1 x n,
  lc_e e ->
  subst_ex e x (open_ex_rec n e1 e2) = open_ex_rec n (subst_ex e x e1) (subst_ex e x e2).
Proof.
pose proof subst_ex_open_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_ex_open_ex_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma subst_ex_open_eX_rec_mutual :
(forall e1 e t x n,
  lc_e e ->
  subst_ex e x (open_eX_rec n t e1) = open_eX_rec n t (subst_ex e x e1)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_ex_open_eX_rec :
forall e1 e t x n,
  lc_e e ->
  subst_ex e x (open_eX_rec n t e1) = open_eX_rec n t (subst_ex e x e1).
Proof.
pose proof subst_ex_open_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_ex_open_eX_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma subst_eX_open_ex_rec_mutual :
(forall e1 t e X n,
  subst_eX t X (open_ex_rec n e e1) = open_ex_rec n (subst_eX t X e) (subst_eX t X e1)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_eX_open_ex_rec :
forall e1 t e X n,
  subst_eX t X (open_ex_rec n e e1) = open_ex_rec n (subst_eX t X e) (subst_eX t X e1).
Proof.
pose proof subst_eX_open_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_eX_open_ex_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma subst_eX_open_eX_rec_mutual :
(forall e t t1 X n,
  lc_t t ->
  subst_eX t X (open_eX_rec n t1 e) = open_eX_rec n (subst_tX t X t1) (subst_eX t X e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_eX_open_eX_rec :
forall e t t1 X n,
  lc_t t ->
  subst_eX t X (open_eX_rec n t1 e) = open_eX_rec n (subst_tX t X t1) (subst_eX t X e).
Proof.
pose proof subst_eX_open_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_eX_open_eX_rec : lngen.

(* end hide *)

Lemma subst_tX_open_tX :
forall t2 t t1 X,
  lc_t t ->
  subst_tX t X (open_tX t1 t2) = open_tX (subst_tX t X t1) (subst_tX t X t2).
Proof.
unfold open_tX; default_simp.
Qed.

Hint Resolve subst_tX_open_tX : lngen.

Lemma subst_ex_open_ex :
forall e2 e e1 x,
  lc_e e ->
  subst_ex e x (open_ex e1 e2) = open_ex (subst_ex e x e1) (subst_ex e x e2).
Proof.
unfold open_ex; default_simp.
Qed.

Hint Resolve subst_ex_open_ex : lngen.

Lemma subst_ex_open_eX :
forall e1 e t x,
  lc_e e ->
  subst_ex e x (open_eX t e1) = open_eX t (subst_ex e x e1).
Proof.
unfold open_eX; default_simp.
Qed.

Hint Resolve subst_ex_open_eX : lngen.

Lemma subst_eX_open_ex :
forall e1 t e X,
  subst_eX t X (open_ex e e1) = open_ex (subst_eX t X e) (subst_eX t X e1).
Proof.
unfold open_ex; default_simp.
Qed.

Hint Resolve subst_eX_open_ex : lngen.

Lemma subst_eX_open_eX :
forall e t t1 X,
  lc_t t ->
  subst_eX t X (open_eX t1 e) = open_eX (subst_tX t X t1) (subst_eX t X e).
Proof.
unfold open_eX; default_simp.
Qed.

Hint Resolve subst_eX_open_eX : lngen.

Lemma subst_tX_open_tX_var :
forall t1 t X X1,
  X <> X1 ->
  lc_t t ->
  subst_tX t X (open_tX (typ_var X1) t1) = open_tX (typ_var X1) (subst_tX t X t1).
Proof.
intros; rewrite subst_tX_open_tX; default_simp.
Qed.

Hint Resolve subst_tX_open_tX_var : lngen.

Lemma subst_ex_open_ex_var :
forall e1 e x x1,
  x <> x1 ->
  lc_e e ->
  subst_ex e x (open_ex (exp_var x1) e1) = open_ex (exp_var x1) (subst_ex e x e1).
Proof.
intros; rewrite subst_ex_open_ex; default_simp.
Qed.

Hint Resolve subst_ex_open_ex_var : lngen.

Lemma subst_ex_open_eX_var :
forall e1 e x X,
  lc_e e ->
  subst_ex e x (open_eX (typ_var X) e1) = open_eX (typ_var X) (subst_ex e x e1).
Proof.
intros; rewrite subst_ex_open_eX; default_simp.
Qed.

Hint Resolve subst_ex_open_eX_var : lngen.

Lemma subst_eX_open_ex_var :
forall e t X x,
  subst_eX t X (open_ex (exp_var x) e) = open_ex (exp_var x) (subst_eX t X e).
Proof.
intros; rewrite subst_eX_open_ex; default_simp.
Qed.

Hint Resolve subst_eX_open_ex_var : lngen.

Lemma subst_eX_open_eX_var :
forall e t X X1,
  X <> X1 ->
  lc_t t ->
  subst_eX t X (open_eX (typ_var X1) e) = open_eX (typ_var X1) (subst_eX t X e).
Proof.
intros; rewrite subst_eX_open_eX; default_simp.
Qed.

Hint Resolve subst_eX_open_eX_var : lngen.

(* begin hide *)

Lemma subst_tX_spec_rec_mutual :
(forall t t1 X n,
  subst_tX t1 X t = open_tX_rec n t1 (close_tX_rec n X t)).
Proof.
apply_mutual_ind typ_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_tX_spec_rec :
forall t t1 X n,
  subst_tX t1 X t = open_tX_rec n t1 (close_tX_rec n X t).
Proof.
pose proof subst_tX_spec_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_tX_spec_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma subst_ex_spec_rec_mutual :
(forall e e1 x n,
  subst_ex e1 x e = open_ex_rec n e1 (close_ex_rec n x e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_ex_spec_rec :
forall e e1 x n,
  subst_ex e1 x e = open_ex_rec n e1 (close_ex_rec n x e).
Proof.
pose proof subst_ex_spec_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_ex_spec_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma subst_eX_spec_rec_mutual :
(forall e t X n,
  subst_eX t X e = open_eX_rec n t (close_eX_rec n X e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_eX_spec_rec :
forall e t X n,
  subst_eX t X e = open_eX_rec n t (close_eX_rec n X e).
Proof.
pose proof subst_eX_spec_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_eX_spec_rec : lngen.

(* end hide *)

Lemma subst_tX_spec :
forall t t1 X,
  subst_tX t1 X t = open_tX t1 (close_tX X t).
Proof.
unfold close_tX; unfold open_tX; default_simp.
Qed.

Hint Resolve subst_tX_spec : lngen.

Lemma subst_ex_spec :
forall e e1 x,
  subst_ex e1 x e = open_ex e1 (close_ex x e).
Proof.
unfold close_ex; unfold open_ex; default_simp.
Qed.

Hint Resolve subst_ex_spec : lngen.

Lemma subst_eX_spec :
forall e t X,
  subst_eX t X e = open_eX t (close_eX X e).
Proof.
unfold close_eX; unfold open_eX; default_simp.
Qed.

Hint Resolve subst_eX_spec : lngen.

(* begin hide *)

Lemma subst_tX_subst_tX_mutual :
(forall t t1 t2 X1 X,
  X1 `notin` fv_tX t1 ->
  X1 <> X ->
  subst_tX t1 X (subst_tX t2 X1 t) = subst_tX (subst_tX t1 X t2) X1 (subst_tX t1 X t)).
Proof.
apply_mutual_ind typ_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_tX_subst_tX :
forall t t1 t2 X1 X,
  X1 `notin` fv_tX t1 ->
  X1 <> X ->
  subst_tX t1 X (subst_tX t2 X1 t) = subst_tX (subst_tX t1 X t2) X1 (subst_tX t1 X t).
Proof.
pose proof subst_tX_subst_tX_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_tX_subst_tX : lngen.

(* begin hide *)

Lemma subst_ex_subst_ex_mutual :
(forall e e1 e2 x1 x,
  x1 `notin` fv_ex e1 ->
  x1 <> x ->
  subst_ex e1 x (subst_ex e2 x1 e) = subst_ex (subst_ex e1 x e2) x1 (subst_ex e1 x e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_ex_subst_ex :
forall e e1 e2 x1 x,
  x1 `notin` fv_ex e1 ->
  x1 <> x ->
  subst_ex e1 x (subst_ex e2 x1 e) = subst_ex (subst_ex e1 x e2) x1 (subst_ex e1 x e).
Proof.
pose proof subst_ex_subst_ex_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_ex_subst_ex : lngen.

(* begin hide *)

Lemma subst_ex_subst_eX_mutual :
(forall e e1 t X x,
  X `notin` fv_eX e1 ->
  subst_ex e1 x (subst_eX t X e) = subst_eX t X (subst_ex e1 x e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_ex_subst_eX :
forall e e1 t X x,
  X `notin` fv_eX e1 ->
  subst_ex e1 x (subst_eX t X e) = subst_eX t X (subst_ex e1 x e).
Proof.
pose proof subst_ex_subst_eX_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_ex_subst_eX : lngen.

(* begin hide *)

Lemma subst_eX_subst_ex_mutual :
(forall e t e1 x X,
  subst_eX t X (subst_ex e1 x e) = subst_ex (subst_eX t X e1) x (subst_eX t X e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_eX_subst_ex :
forall e t e1 x X,
  subst_eX t X (subst_ex e1 x e) = subst_ex (subst_eX t X e1) x (subst_eX t X e).
Proof.
pose proof subst_eX_subst_ex_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_eX_subst_ex : lngen.

(* begin hide *)

Lemma subst_eX_subst_eX_mutual :
(forall e t t1 X1 X,
  X1 `notin` fv_tX t ->
  X1 <> X ->
  subst_eX t X (subst_eX t1 X1 e) = subst_eX (subst_tX t X t1) X1 (subst_eX t X e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_eX_subst_eX :
forall e t t1 X1 X,
  X1 `notin` fv_tX t ->
  X1 <> X ->
  subst_eX t X (subst_eX t1 X1 e) = subst_eX (subst_tX t X t1) X1 (subst_eX t X e).
Proof.
pose proof subst_eX_subst_eX_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_eX_subst_eX : lngen.


(* *********************************************************************** *)
(** * Theorems about [fv] *)

(* begin hide *)

Lemma fv_tX_close_tX_rec_mutual :
(forall t X n,
  fv_tX (close_tX_rec n X t) [=] remove X (fv_tX t)).
Proof.
apply_mutual_ind typ_mutind;
default_simp; fsetdec typvar.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_tX_close_tX_rec :
forall t X n,
  fv_tX (close_tX_rec n X t) [=] remove X (fv_tX t).
Proof.
pose proof fv_tX_close_tX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_tX_close_tX_rec : lngen.
Hint Rewrite fv_tX_close_tX_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_ex_close_ex_rec_mutual :
(forall e x n,
  fv_ex (close_ex_rec n x e) [=] remove x (fv_ex e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec expvar.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_ex_close_ex_rec :
forall e x n,
  fv_ex (close_ex_rec n x e) [=] remove x (fv_ex e).
Proof.
pose proof fv_ex_close_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_close_ex_rec : lngen.
Hint Rewrite fv_ex_close_ex_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_ex_close_eX_rec_mutual :
(forall e X n,
  fv_ex (close_eX_rec n X e) [=] fv_ex e).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec expvar.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_ex_close_eX_rec :
forall e X n,
  fv_ex (close_eX_rec n X e) [=] fv_ex e.
Proof.
pose proof fv_ex_close_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_close_eX_rec : lngen.
Hint Rewrite fv_ex_close_eX_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_eX_close_ex_rec_mutual :
(forall e x n,
  fv_eX (close_ex_rec n x e) [=] fv_eX e).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec typvar.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_eX_close_ex_rec :
forall e x n,
  fv_eX (close_ex_rec n x e) [=] fv_eX e.
Proof.
pose proof fv_eX_close_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_close_ex_rec : lngen.
Hint Rewrite fv_eX_close_ex_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_eX_close_eX_rec_mutual :
(forall e X n,
  fv_eX (close_eX_rec n X e) [=] remove X (fv_eX e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec typvar.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_eX_close_eX_rec :
forall e X n,
  fv_eX (close_eX_rec n X e) [=] remove X (fv_eX e).
Proof.
pose proof fv_eX_close_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_close_eX_rec : lngen.
Hint Rewrite fv_eX_close_eX_rec using solve [auto] : lngen.

(* end hide *)

Lemma fv_tX_close_tX :
forall t X,
  fv_tX (close_tX X t) [=] remove X (fv_tX t).
Proof.
unfold close_tX; default_simp.
Qed.

Hint Resolve fv_tX_close_tX : lngen.
Hint Rewrite fv_tX_close_tX using solve [auto] : lngen.

Lemma fv_ex_close_ex :
forall e x,
  fv_ex (close_ex x e) [=] remove x (fv_ex e).
Proof.
unfold close_ex; default_simp.
Qed.

Hint Resolve fv_ex_close_ex : lngen.
Hint Rewrite fv_ex_close_ex using solve [auto] : lngen.

Lemma fv_ex_close_eX :
forall e X,
  fv_ex (close_eX X e) [=] fv_ex e.
Proof.
unfold close_eX; default_simp.
Qed.

Hint Resolve fv_ex_close_eX : lngen.
Hint Rewrite fv_ex_close_eX using solve [auto] : lngen.

Lemma fv_eX_close_ex :
forall e x,
  fv_eX (close_ex x e) [=] fv_eX e.
Proof.
unfold close_ex; default_simp.
Qed.

Hint Resolve fv_eX_close_ex : lngen.
Hint Rewrite fv_eX_close_ex using solve [auto] : lngen.

Lemma fv_eX_close_eX :
forall e X,
  fv_eX (close_eX X e) [=] remove X (fv_eX e).
Proof.
unfold close_eX; default_simp.
Qed.

Hint Resolve fv_eX_close_eX : lngen.
Hint Rewrite fv_eX_close_eX using solve [auto] : lngen.

(* begin hide *)

Lemma fv_tX_open_tX_rec_lower_mutual :
(forall t t1 n,
  fv_tX t [<=] fv_tX (open_tX_rec n t1 t)).
Proof.
apply_mutual_ind typ_mutind;
default_simp; fsetdec typvar.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_tX_open_tX_rec_lower :
forall t t1 n,
  fv_tX t [<=] fv_tX (open_tX_rec n t1 t).
Proof.
pose proof fv_tX_open_tX_rec_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_tX_open_tX_rec_lower : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_ex_open_ex_rec_lower_mutual :
(forall e e1 n,
  fv_ex e [<=] fv_ex (open_ex_rec n e1 e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec expvar.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_ex_open_ex_rec_lower :
forall e e1 n,
  fv_ex e [<=] fv_ex (open_ex_rec n e1 e).
Proof.
pose proof fv_ex_open_ex_rec_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_open_ex_rec_lower : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_ex_open_eX_rec_lower_mutual :
(forall e t n,
  fv_ex e [<=] fv_ex (open_eX_rec n t e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec expvar.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_ex_open_eX_rec_lower :
forall e t n,
  fv_ex e [<=] fv_ex (open_eX_rec n t e).
Proof.
pose proof fv_ex_open_eX_rec_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_open_eX_rec_lower : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_eX_open_ex_rec_lower_mutual :
(forall e e1 n,
  fv_eX e [<=] fv_eX (open_ex_rec n e1 e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec typvar.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_eX_open_ex_rec_lower :
forall e e1 n,
  fv_eX e [<=] fv_eX (open_ex_rec n e1 e).
Proof.
pose proof fv_eX_open_ex_rec_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_open_ex_rec_lower : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_eX_open_eX_rec_lower_mutual :
(forall e t n,
  fv_eX e [<=] fv_eX (open_eX_rec n t e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec typvar.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_eX_open_eX_rec_lower :
forall e t n,
  fv_eX e [<=] fv_eX (open_eX_rec n t e).
Proof.
pose proof fv_eX_open_eX_rec_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_open_eX_rec_lower : lngen.

(* end hide *)

Lemma fv_tX_open_tX_lower :
forall t t1,
  fv_tX t [<=] fv_tX (open_tX t1 t).
Proof.
unfold open_tX; default_simp.
Qed.

Hint Resolve fv_tX_open_tX_lower : lngen.

Lemma fv_ex_open_ex_lower :
forall e e1,
  fv_ex e [<=] fv_ex (open_ex e1 e).
Proof.
unfold open_ex; default_simp.
Qed.

Hint Resolve fv_ex_open_ex_lower : lngen.

Lemma fv_ex_open_eX_lower :
forall e t,
  fv_ex e [<=] fv_ex (open_eX t e).
Proof.
unfold open_eX; default_simp.
Qed.

Hint Resolve fv_ex_open_eX_lower : lngen.

Lemma fv_eX_open_ex_lower :
forall e e1,
  fv_eX e [<=] fv_eX (open_ex e1 e).
Proof.
unfold open_ex; default_simp.
Qed.

Hint Resolve fv_eX_open_ex_lower : lngen.

Lemma fv_eX_open_eX_lower :
forall e t,
  fv_eX e [<=] fv_eX (open_eX t e).
Proof.
unfold open_eX; default_simp.
Qed.

Hint Resolve fv_eX_open_eX_lower : lngen.

(* begin hide *)

Lemma fv_tX_open_tX_rec_upper_mutual :
(forall t t1 n,
  fv_tX (open_tX_rec n t1 t) [<=] fv_tX t1 `union` fv_tX t).
Proof.
apply_mutual_ind typ_mutind;
default_simp; fsetdec typvar.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_tX_open_tX_rec_upper :
forall t t1 n,
  fv_tX (open_tX_rec n t1 t) [<=] fv_tX t1 `union` fv_tX t.
Proof.
pose proof fv_tX_open_tX_rec_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_tX_open_tX_rec_upper : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_ex_open_ex_rec_upper_mutual :
(forall e e1 n,
  fv_ex (open_ex_rec n e1 e) [<=] fv_ex e1 `union` fv_ex e).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec expvar.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_ex_open_ex_rec_upper :
forall e e1 n,
  fv_ex (open_ex_rec n e1 e) [<=] fv_ex e1 `union` fv_ex e.
Proof.
pose proof fv_ex_open_ex_rec_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_open_ex_rec_upper : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_ex_open_eX_rec_upper_mutual :
(forall e t n,
  fv_ex (open_eX_rec n t e) [<=] fv_ex e).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec expvar.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_ex_open_eX_rec_upper :
forall e t n,
  fv_ex (open_eX_rec n t e) [<=] fv_ex e.
Proof.
pose proof fv_ex_open_eX_rec_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_open_eX_rec_upper : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_eX_open_ex_rec_upper_mutual :
(forall e e1 n,
  fv_eX (open_ex_rec n e1 e) [<=] fv_eX e1 `union` fv_eX e).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec typvar.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_eX_open_ex_rec_upper :
forall e e1 n,
  fv_eX (open_ex_rec n e1 e) [<=] fv_eX e1 `union` fv_eX e.
Proof.
pose proof fv_eX_open_ex_rec_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_open_ex_rec_upper : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_eX_open_eX_rec_upper_mutual :
(forall e t n,
  fv_eX (open_eX_rec n t e) [<=] fv_tX t `union` fv_eX e).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec typvar.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_eX_open_eX_rec_upper :
forall e t n,
  fv_eX (open_eX_rec n t e) [<=] fv_tX t `union` fv_eX e.
Proof.
pose proof fv_eX_open_eX_rec_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_open_eX_rec_upper : lngen.

(* end hide *)

Lemma fv_tX_open_tX_upper :
forall t t1,
  fv_tX (open_tX t1 t) [<=] fv_tX t1 `union` fv_tX t.
Proof.
unfold open_tX; default_simp.
Qed.

Hint Resolve fv_tX_open_tX_upper : lngen.

Lemma fv_ex_open_ex_upper :
forall e e1,
  fv_ex (open_ex e1 e) [<=] fv_ex e1 `union` fv_ex e.
Proof.
unfold open_ex; default_simp.
Qed.

Hint Resolve fv_ex_open_ex_upper : lngen.

Lemma fv_ex_open_eX_upper :
forall e t,
  fv_ex (open_eX t e) [<=] fv_ex e.
Proof.
unfold open_eX; default_simp.
Qed.

Hint Resolve fv_ex_open_eX_upper : lngen.

Lemma fv_eX_open_ex_upper :
forall e e1,
  fv_eX (open_ex e1 e) [<=] fv_eX e1 `union` fv_eX e.
Proof.
unfold open_ex; default_simp.
Qed.

Hint Resolve fv_eX_open_ex_upper : lngen.

Lemma fv_eX_open_eX_upper :
forall e t,
  fv_eX (open_eX t e) [<=] fv_tX t `union` fv_eX e.
Proof.
unfold open_eX; default_simp.
Qed.

Hint Resolve fv_eX_open_eX_upper : lngen.

(* begin hide *)

Lemma fv_tX_subst_tX_fresh_mutual :
(forall t t1 X,
  X `notin` fv_tX t ->
  fv_tX (subst_tX t1 X t) [=] fv_tX t).
Proof.
apply_mutual_ind typ_mutind;
default_simp; fsetdec typvar.
Qed.

(* end hide *)

Lemma fv_tX_subst_tX_fresh :
forall t t1 X,
  X `notin` fv_tX t ->
  fv_tX (subst_tX t1 X t) [=] fv_tX t.
Proof.
pose proof fv_tX_subst_tX_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_tX_subst_tX_fresh : lngen.
Hint Rewrite fv_tX_subst_tX_fresh using solve [auto] : lngen.

(* begin hide *)

Lemma fv_ex_subst_ex_fresh_mutual :
(forall e e1 x,
  x `notin` fv_ex e ->
  fv_ex (subst_ex e1 x e) [=] fv_ex e).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec expvar.
Qed.

(* end hide *)

Lemma fv_ex_subst_ex_fresh :
forall e e1 x,
  x `notin` fv_ex e ->
  fv_ex (subst_ex e1 x e) [=] fv_ex e.
Proof.
pose proof fv_ex_subst_ex_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_subst_ex_fresh : lngen.
Hint Rewrite fv_ex_subst_ex_fresh using solve [auto] : lngen.

(* begin hide *)

Lemma fv_eX_subst_ex_fresh_mutual :
(forall e t X,
  fv_ex (subst_eX t X e) [=] fv_ex e).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec expvar.
Qed.

(* end hide *)

Lemma fv_eX_subst_ex_fresh :
forall e t X,
  fv_ex (subst_eX t X e) [=] fv_ex e.
Proof.
pose proof fv_eX_subst_ex_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_subst_ex_fresh : lngen.
Hint Rewrite fv_eX_subst_ex_fresh using solve [auto] : lngen.

(* begin hide *)

Lemma fv_eX_subst_eX_fresh_mutual :
(forall e t X,
  X `notin` fv_eX e ->
  fv_eX (subst_eX t X e) [=] fv_eX e).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec typvar.
Qed.

(* end hide *)

Lemma fv_eX_subst_eX_fresh :
forall e t X,
  X `notin` fv_eX e ->
  fv_eX (subst_eX t X e) [=] fv_eX e.
Proof.
pose proof fv_eX_subst_eX_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_subst_eX_fresh : lngen.
Hint Rewrite fv_eX_subst_eX_fresh using solve [auto] : lngen.

(* begin hide *)

Lemma fv_tX_subst_tX_lower_mutual :
(forall t t1 X,
  remove X (fv_tX t) [<=] fv_tX (subst_tX t1 X t)).
Proof.
apply_mutual_ind typ_mutind;
default_simp; fsetdec typvar.
Qed.

(* end hide *)

Lemma fv_tX_subst_tX_lower :
forall t t1 X,
  remove X (fv_tX t) [<=] fv_tX (subst_tX t1 X t).
Proof.
pose proof fv_tX_subst_tX_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_tX_subst_tX_lower : lngen.

(* begin hide *)

Lemma fv_ex_subst_ex_lower_mutual :
(forall e e1 x,
  remove x (fv_ex e) [<=] fv_ex (subst_ex e1 x e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec expvar.
Qed.

(* end hide *)

Lemma fv_ex_subst_ex_lower :
forall e e1 x,
  remove x (fv_ex e) [<=] fv_ex (subst_ex e1 x e).
Proof.
pose proof fv_ex_subst_ex_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_subst_ex_lower : lngen.

(* begin hide *)

Lemma fv_ex_subst_eX_lower_mutual :
(forall e t X,
  fv_ex e [<=] fv_ex (subst_eX t X e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec expvar.
Qed.

(* end hide *)

Lemma fv_ex_subst_eX_lower :
forall e t X,
  fv_ex e [<=] fv_ex (subst_eX t X e).
Proof.
pose proof fv_ex_subst_eX_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_subst_eX_lower : lngen.

(* begin hide *)

Lemma fv_eX_subst_ex_lower_mutual :
(forall e e1 x,
  fv_eX e [<=] fv_eX (subst_ex e1 x e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec typvar.
Qed.

(* end hide *)

Lemma fv_eX_subst_ex_lower :
forall e e1 x,
  fv_eX e [<=] fv_eX (subst_ex e1 x e).
Proof.
pose proof fv_eX_subst_ex_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_subst_ex_lower : lngen.

(* begin hide *)

Lemma fv_eX_subst_eX_lower_mutual :
(forall e t X,
  remove X (fv_eX e) [<=] fv_eX (subst_eX t X e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec typvar.
Qed.

(* end hide *)

Lemma fv_eX_subst_eX_lower :
forall e t X,
  remove X (fv_eX e) [<=] fv_eX (subst_eX t X e).
Proof.
pose proof fv_eX_subst_eX_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_subst_eX_lower : lngen.

(* begin hide *)

Lemma fv_tX_subst_tX_notin_mutual :
(forall t t1 X X1,
  X1 `notin` fv_tX t ->
  X1 `notin` fv_tX t1 ->
  X1 `notin` fv_tX (subst_tX t1 X t)).
Proof.
apply_mutual_ind typ_mutind;
default_simp; fsetdec typvar.
Qed.

(* end hide *)

Lemma fv_tX_subst_tX_notin :
forall t t1 X X1,
  X1 `notin` fv_tX t ->
  X1 `notin` fv_tX t1 ->
  X1 `notin` fv_tX (subst_tX t1 X t).
Proof.
pose proof fv_tX_subst_tX_notin_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_tX_subst_tX_notin : lngen.

(* begin hide *)

Lemma fv_ex_subst_ex_notin_mutual :
(forall e e1 x x1,
  x1 `notin` fv_ex e ->
  x1 `notin` fv_ex e1 ->
  x1 `notin` fv_ex (subst_ex e1 x e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec expvar.
Qed.

(* end hide *)

Lemma fv_ex_subst_ex_notin :
forall e e1 x x1,
  x1 `notin` fv_ex e ->
  x1 `notin` fv_ex e1 ->
  x1 `notin` fv_ex (subst_ex e1 x e).
Proof.
pose proof fv_ex_subst_ex_notin_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_subst_ex_notin : lngen.

(* begin hide *)

Lemma fv_ex_subst_eX_notin_mutual :
(forall e t X x,
  x `notin` fv_ex e ->
  x `notin` fv_ex (subst_eX t X e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec expvar.
Qed.

(* end hide *)

Lemma fv_ex_subst_eX_notin :
forall e t X x,
  x `notin` fv_ex e ->
  x `notin` fv_ex (subst_eX t X e).
Proof.
pose proof fv_ex_subst_eX_notin_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_subst_eX_notin : lngen.

(* begin hide *)

Lemma fv_eX_subst_ex_notin_mutual :
(forall e e1 x X,
  X `notin` fv_eX e ->
  X `notin` fv_eX e1 ->
  X `notin` fv_eX (subst_ex e1 x e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec typvar.
Qed.

(* end hide *)

Lemma fv_eX_subst_ex_notin :
forall e e1 x X,
  X `notin` fv_eX e ->
  X `notin` fv_eX e1 ->
  X `notin` fv_eX (subst_ex e1 x e).
Proof.
pose proof fv_eX_subst_ex_notin_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_subst_ex_notin : lngen.

(* begin hide *)

Lemma fv_eX_subst_eX_notin_mutual :
(forall e t X X1,
  X1 `notin` fv_eX e ->
  X1 `notin` fv_tX t ->
  X1 `notin` fv_eX (subst_eX t X e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec typvar.
Qed.

(* end hide *)

Lemma fv_eX_subst_eX_notin :
forall e t X X1,
  X1 `notin` fv_eX e ->
  X1 `notin` fv_tX t ->
  X1 `notin` fv_eX (subst_eX t X e).
Proof.
pose proof fv_eX_subst_eX_notin_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_subst_eX_notin : lngen.

(* begin hide *)

Lemma fv_tX_subst_tX_upper_mutual :
(forall t t1 X,
  fv_tX (subst_tX t1 X t) [<=] fv_tX t1 `union` remove X (fv_tX t)).
Proof.
apply_mutual_ind typ_mutind;
default_simp; fsetdec typvar.
Qed.

(* end hide *)

Lemma fv_tX_subst_tX_upper :
forall t t1 X,
  fv_tX (subst_tX t1 X t) [<=] fv_tX t1 `union` remove X (fv_tX t).
Proof.
pose proof fv_tX_subst_tX_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_tX_subst_tX_upper : lngen.

(* begin hide *)

Lemma fv_ex_subst_ex_upper_mutual :
(forall e e1 x,
  fv_ex (subst_ex e1 x e) [<=] fv_ex e1 `union` remove x (fv_ex e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec expvar.
Qed.

(* end hide *)

Lemma fv_ex_subst_ex_upper :
forall e e1 x,
  fv_ex (subst_ex e1 x e) [<=] fv_ex e1 `union` remove x (fv_ex e).
Proof.
pose proof fv_ex_subst_ex_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_subst_ex_upper : lngen.

(* begin hide *)

Lemma fv_ex_subst_eX_upper_mutual :
(forall e t X,
  fv_ex (subst_eX t X e) [<=] fv_ex e).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec expvar.
Qed.

(* end hide *)

Lemma fv_ex_subst_eX_upper :
forall e t X,
  fv_ex (subst_eX t X e) [<=] fv_ex e.
Proof.
pose proof fv_ex_subst_eX_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_ex_subst_eX_upper : lngen.

(* begin hide *)

Lemma fv_eX_subst_ex_upper_mutual :
(forall e e1 x,
  fv_eX (subst_ex e1 x e) [<=] fv_eX e1 `union` fv_eX e).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec typvar.
Qed.

(* end hide *)

Lemma fv_eX_subst_ex_upper :
forall e e1 x,
  fv_eX (subst_ex e1 x e) [<=] fv_eX e1 `union` fv_eX e.
Proof.
pose proof fv_eX_subst_ex_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_subst_ex_upper : lngen.

(* begin hide *)

Lemma fv_eX_subst_eX_upper_mutual :
(forall e t X,
  fv_eX (subst_eX t X e) [<=] fv_tX t `union` remove X (fv_eX e)).
Proof.
apply_mutual_ind exp_mutind;
default_simp; fsetdec typvar.
Qed.

(* end hide *)

Lemma fv_eX_subst_eX_upper :
forall e t X,
  fv_eX (subst_eX t X e) [<=] fv_tX t `union` remove X (fv_eX e).
Proof.
pose proof fv_eX_subst_eX_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_eX_subst_eX_upper : lngen.


(* *********************************************************************** *)
(** * More theorems about [subst] *)

(* begin hide *)

Lemma subst_tX_close_tX_rec_open_tX_rec_mutual :
(forall t1 t X X1 n,
  X1 `notin` fv_tX t1 ->
  X1 `notin` fv_tX t ->
  X1 <> X ->
  degree_tX n t ->
  subst_tX t X t1 = close_tX_rec n X1 (subst_tX t X (open_tX_rec n (typ_var X1) t1))).
Proof.
apply_mutual_ind typ_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_tX_close_tX_rec_open_tX_rec :
forall t1 t X X1 n,
  X1 `notin` fv_tX t1 ->
  X1 `notin` fv_tX t ->
  X1 <> X ->
  degree_tX n t ->
  subst_tX t X t1 = close_tX_rec n X1 (subst_tX t X (open_tX_rec n (typ_var X1) t1)).
Proof.
pose proof subst_tX_close_tX_rec_open_tX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_tX_close_tX_rec_open_tX_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma subst_ex_close_ex_rec_open_ex_rec_mutual :
(forall e1 e x x1 n,
  x1 `notin` fv_ex e1 ->
  x1 `notin` fv_ex e ->
  x1 <> x ->
  degree_ex n e ->
  subst_ex e x e1 = close_ex_rec n x1 (subst_ex e x (open_ex_rec n (exp_var x1) e1))).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_ex_close_ex_rec_open_ex_rec :
forall e1 e x x1 n,
  x1 `notin` fv_ex e1 ->
  x1 `notin` fv_ex e ->
  x1 <> x ->
  degree_ex n e ->
  subst_ex e x e1 = close_ex_rec n x1 (subst_ex e x (open_ex_rec n (exp_var x1) e1)).
Proof.
pose proof subst_ex_close_ex_rec_open_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_ex_close_ex_rec_open_ex_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma subst_ex_close_eX_rec_open_eX_rec_mutual :
(forall e1 e x X n,
  X `notin` fv_eX e1 ->
  X `notin` fv_eX e ->
  degree_eX n e ->
  subst_ex e x e1 = close_eX_rec n X (subst_ex e x (open_eX_rec n (typ_var X) e1))).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_ex_close_eX_rec_open_eX_rec :
forall e1 e x X n,
  X `notin` fv_eX e1 ->
  X `notin` fv_eX e ->
  degree_eX n e ->
  subst_ex e x e1 = close_eX_rec n X (subst_ex e x (open_eX_rec n (typ_var X) e1)).
Proof.
pose proof subst_ex_close_eX_rec_open_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_ex_close_eX_rec_open_eX_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma subst_eX_close_ex_rec_open_ex_rec_mutual :
(forall e t X x n,
  x `notin` fv_ex e ->
  subst_eX t X e = close_ex_rec n x (subst_eX t X (open_ex_rec n (exp_var x) e))).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_eX_close_ex_rec_open_ex_rec :
forall e t X x n,
  x `notin` fv_ex e ->
  subst_eX t X e = close_ex_rec n x (subst_eX t X (open_ex_rec n (exp_var x) e)).
Proof.
pose proof subst_eX_close_ex_rec_open_ex_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_eX_close_ex_rec_open_ex_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma subst_eX_close_eX_rec_open_eX_rec_mutual :
(forall e t X X1 n,
  X1 `notin` fv_eX e ->
  X1 `notin` fv_tX t ->
  X1 <> X ->
  degree_tX n t ->
  subst_eX t X e = close_eX_rec n X1 (subst_eX t X (open_eX_rec n (typ_var X1) e))).
Proof.
apply_mutual_ind exp_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_eX_close_eX_rec_open_eX_rec :
forall e t X X1 n,
  X1 `notin` fv_eX e ->
  X1 `notin` fv_tX t ->
  X1 <> X ->
  degree_tX n t ->
  subst_eX t X e = close_eX_rec n X1 (subst_eX t X (open_eX_rec n (typ_var X1) e)).
Proof.
pose proof subst_eX_close_eX_rec_open_eX_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_eX_close_eX_rec_open_eX_rec : lngen.

(* end hide *)

Lemma subst_tX_close_tX_open_tX :
forall t1 t X X1,
  X1 `notin` fv_tX t1 ->
  X1 `notin` fv_tX t ->
  X1 <> X ->
  lc_t t ->
  subst_tX t X t1 = close_tX X1 (subst_tX t X (open_tX (typ_var X1) t1)).
Proof.
unfold close_tX; unfold open_tX; default_simp.
Qed.

Hint Resolve subst_tX_close_tX_open_tX : lngen.

Lemma subst_ex_close_ex_open_ex :
forall e1 e x x1,
  x1 `notin` fv_ex e1 ->
  x1 `notin` fv_ex e ->
  x1 <> x ->
  lc_e e ->
  subst_ex e x e1 = close_ex x1 (subst_ex e x (open_ex (exp_var x1) e1)).
Proof.
unfold close_ex; unfold open_ex; default_simp.
Qed.

Hint Resolve subst_ex_close_ex_open_ex : lngen.

Lemma subst_ex_close_eX_open_eX :
forall e1 e x X,
  X `notin` fv_eX e1 ->
  X `notin` fv_eX e ->
  lc_e e ->
  subst_ex e x e1 = close_eX X (subst_ex e x (open_eX (typ_var X) e1)).
Proof.
unfold close_eX; unfold open_eX; default_simp.
Qed.

Hint Resolve subst_ex_close_eX_open_eX : lngen.

Lemma subst_eX_close_ex_open_ex :
forall e t X x,
  x `notin` fv_ex e ->
  lc_t t ->
  subst_eX t X e = close_ex x (subst_eX t X (open_ex (exp_var x) e)).
Proof.
unfold close_ex; unfold open_ex; default_simp.
Qed.

Hint Resolve subst_eX_close_ex_open_ex : lngen.

Lemma subst_eX_close_eX_open_eX :
forall e t X X1,
  X1 `notin` fv_eX e ->
  X1 `notin` fv_tX t ->
  X1 <> X ->
  lc_t t ->
  subst_eX t X e = close_eX X1 (subst_eX t X (open_eX (typ_var X1) e)).
Proof.
unfold close_eX; unfold open_eX; default_simp.
Qed.

Hint Resolve subst_eX_close_eX_open_eX : lngen.

Lemma subst_tX_typ_all :
forall X1 t1 t2 t X,
  lc_t t ->
  X1 `notin` fv_tX t `union` fv_tX t2 `union` singleton X ->
  subst_tX t X (typ_all t1 t2) = typ_all (subst_tX t X t1) (close_tX X1 (subst_tX t X (open_tX (typ_var X1) t2))).
Proof.
default_simp.
Qed.

Hint Resolve subst_tX_typ_all : lngen.

Lemma subst_ex_exp_abs :
forall x1 t e1 e x,
  lc_e e ->
  x1 `notin` fv_ex e `union` fv_ex e1 `union` singleton x ->
  subst_ex e x (exp_abs t e1) = exp_abs (t) (close_ex x1 (subst_ex e x (open_ex (exp_var x1) e1))).
Proof.
default_simp.
Qed.

Hint Resolve subst_ex_exp_abs : lngen.

Lemma subst_ex_exp_tabs :
forall X t e1 e x,
  lc_e e ->
  X `notin` fv_eX e `union` fv_eX e1 ->
  subst_ex e x (exp_tabs t e1) = exp_tabs (t) (close_eX X (subst_ex e x (open_eX (typ_var X) e1))).
Proof.
default_simp.
Qed.

Hint Resolve subst_ex_exp_tabs : lngen.

Lemma subst_eX_exp_abs :
forall x t1 e t X,
  lc_t t ->
  x `notin` fv_ex e ->
  subst_eX t X (exp_abs t1 e) = exp_abs (subst_tX t X t1) (close_ex x (subst_eX t X (open_ex (exp_var x) e))).
Proof.
default_simp.
Qed.

Hint Resolve subst_eX_exp_abs : lngen.

Lemma subst_eX_exp_tabs :
forall X1 t1 e t X,
  lc_t t ->
  X1 `notin` fv_tX t `union` fv_eX e `union` singleton X ->
  subst_eX t X (exp_tabs t1 e) = exp_tabs (subst_tX t X t1) (close_eX X1 (subst_eX t X (open_eX (typ_var X1) e))).
Proof.
default_simp.
Qed.

Hint Resolve subst_eX_exp_tabs : lngen.

