Add LoadPath "metatheory-8.2".
Require Export Coq.Program.Equality.
Require Export Metatheory.
Require Export LNgenFacts.
Require Import Wf_nat.

(** NOTE: Auxiliary theorems are hidden in generated documentation.
    In general, there is a [_rec] version of every lemma involving
    [open] and [close]. *)

(** For generated HTML documentation. *)

(** remove printing <= *)

(** For proving the uniqueness of [lc] proofs. *)

Axiom axiom_fun_ext_eq :
  forall (A : Type) (B : A -> Type) (f g : forall x : A, B x),
    (forall x, f x = g x) ->
    f = g.


(* *********************************************************************** *)
(** * Ott "Phantom" declarations *)

Parameter family_constant : Set.

Parameter object_constant : Set.


(* *********************************************************************** *)
(** * Metavariable declarations *)

(** The rest of this file assumes that [EqDec] and [Atom] instances
    are already declared for the implementation types below. *)

Notation var := (string).


(* *********************************************************************** *)
(** * Nonterminal declarations *)

Inductive family : Set :=
  | family_const : family_constant -> family
  | family_pi : family -> family -> family
  | family_abs : family -> family -> family
  | family_app : family -> object -> family

with object : Set :=
  | object_const : object_constant -> object
  | object_var : var -> object
  | object_var_ind : nat -> object
  | object_abs : family -> object -> object
  | object_app : object -> object -> object.

Scheme family_ind' := Induction for family Sort Prop
  with object_ind' := Induction for object Sort Prop.

Definition family_object_mutind :=
  fun H H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 =>
  (conj (family_ind' H H1 H2 H3 H4 H5 H6 H7 H8 H9 H10)
  (object_ind' H H1 H2 H3 H4 H5 H6 H7 H8 H9 H10)).

Scheme family_rec' := Induction for family Sort Set
  with object_rec' := Induction for object Sort Set.

Definition family_object_mutrec :=
  fun H H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 =>
  (pair (family_rec' H H1 H2 H3 H4 H5 H6 H7 H8 H9 H10)
  (object_rec' H H1 H2 H3 H4 H5 H6 H7 H8 H9 H10)).

Inductive kind : Set :=
  | kind_type : kind
  | kind_pi : family -> kind -> kind.

Scheme kind_ind' := Induction for kind Sort Prop.

Definition kind_mutind :=
  fun H H1 H2 =>
  kind_ind' H H1 H2.

Scheme kind_rec' := Induction for kind Sort Set.

Definition kind_mutrec :=
  fun H H1 H2 =>
  kind_rec' H H1 H2.


(* *********************************************************************** *)
(** * Close *)

Fixpoint close_Ax_rec (n : nat) (x : var) (A : family) {struct A} : family :=
  match A with
    | family_const a => family_const a
    | family_pi A1 B => family_pi (close_Ax_rec n x A1) (close_Ax_rec (S n) x B)
    | family_abs A1 B => family_abs (close_Ax_rec n x A1) (close_Ax_rec (S n) x B)
    | family_app A1 M => family_app (close_Ax_rec n x A1) (close_Mx_rec n x M)
  end

with close_Mx_rec (n : nat) (x : var) (M : object) {struct M} : object :=
  match M with
    | object_const c => object_const c
    | object_var x1 => if (x == x1) then (object_var_ind n) else (object_var x1)
    | object_var_ind n1 => if (lt_ge_dec n1 n) then (object_var_ind n1) else (object_var_ind (S n1))
    | object_abs A M1 => object_abs (close_Ax_rec n x A) (close_Mx_rec (S n) x M1)
    | object_app M1 N => object_app (close_Mx_rec n x M1) (close_Mx_rec n x N)
  end.

Definition close_Ax x A := close_Ax_rec 0 x A.

Definition close_Mx x M := close_Mx_rec 0 x M.

Fixpoint close_Kx_rec (n : nat) (x : var) (K : kind) {struct K} : kind :=
  match K with
    | kind_type => kind_type
    | kind_pi A K1 => kind_pi (close_Ax_rec n x A) (close_Kx_rec (S n) x K1)
  end.

Definition close_Kx x K := close_Kx_rec 0 x K.


(* *********************************************************************** *)
(** * Free variables *)

Fixpoint fv_Ax (A : family) {struct A} : fset var :=
  match A with
    | family_const a => empty
    | family_pi A1 B => (fv_Ax A1) `union` (fv_Ax B)
    | family_abs A1 B => (fv_Ax A1) `union` (fv_Ax B)
    | family_app A1 M => (fv_Ax A1) `union` (fv_Mx M)
  end

with fv_Mx (M : object) {struct M} : fset var :=
  match M with
    | object_const c => empty
    | object_var x => (singleton x)
    | object_var_ind n => empty
    | object_abs A M1 => (fv_Ax A) `union` (fv_Mx M1)
    | object_app M1 N => (fv_Mx M1) `union` (fv_Mx N)
  end.

Fixpoint fv_Kx (K : kind) {struct K} : fset var :=
  match K with
    | kind_type => empty
    | kind_pi A K1 => (fv_Ax A) `union` (fv_Kx K1)
  end.


(* *********************************************************************** *)
(** * Open *)

(** The definitions below are only correct under the following
    condition: when substituting for index [n], the [degree] of
    the term being substituted in should also be [n]. *)

Fixpoint open_Ax_rec (n : nat) (M : object) (A : family) {struct A} : family :=
  match A with
    | family_const a => family_const a
    | family_pi A1 B => family_pi (open_Ax_rec n M A1) (open_Ax_rec (S n) M B)
    | family_abs A1 B => family_abs (open_Ax_rec n M A1) (open_Ax_rec (S n) M B)
    | family_app A1 M1 => family_app (open_Ax_rec n M A1) (open_Mx_rec n M M1)
  end

with open_Mx_rec (n : nat) (M : object) (M1 : object) {struct M1} : object :=
  match M1 with
    | object_const c => object_const c
    | object_var x => object_var x
    | object_var_ind n1 =>
      match lt_eq_lt_dec n1 n with
        | inleft (left _)  => object_var_ind n1
        | inleft (right _) => M
        | inright _        => object_var_ind (n1 - 1)
      end
    | object_abs A M2 => object_abs (open_Ax_rec n M A) (open_Mx_rec (S n) M M2)
    | object_app M2 N => object_app (open_Mx_rec n M M2) (open_Mx_rec n M N)
  end.

Definition open_Ax M A := open_Ax_rec 0 M A.

Definition open_Mx M1 M := open_Mx_rec 0 M1 M.

Fixpoint open_Kx_rec (n : nat) (M : object) (K : kind) {struct K} : kind :=
  match K with
    | kind_type => kind_type
    | kind_pi A K1 => kind_pi (open_Ax_rec n M A) (open_Kx_rec (S n) M K1)
  end.

Definition open_Kx M K := open_Kx_rec 0 M K.


(* *********************************************************************** *)
(** * Size *)

Fixpoint size_A (A : family) {struct A} : nat :=
  match A with
    | family_const a => 1
    | family_pi A1 B => 1 + (size_A A1) + (size_A B)
    | family_abs A1 B => 1 + (size_A A1) + (size_A B)
    | family_app A1 M => 1 + (size_A A1) + (size_M M)
  end

with size_M (M : object) {struct M} : nat :=
  match M with
    | object_const c => 1
    | object_var x => 1
    | object_var_ind n => 1
    | object_abs A M1 => 1 + (size_A A) + (size_M M1)
    | object_app M1 N => 1 + (size_M M1) + (size_M N)
  end.

Fixpoint size_K (K : kind) {struct K} : nat :=
  match K with
    | kind_type => 1
    | kind_pi A K1 => 1 + (size_A A) + (size_K K1)
  end.


(* *********************************************************************** *)
(** * Substitution *)

Fixpoint subst_Ax (M : object) (x : var) (A : family) {struct A} : family :=
  match A with
    | family_const a => family_const a
    | family_pi A1 B => family_pi (subst_Ax M x A1) (subst_Ax M x B)
    | family_abs A1 B => family_abs (subst_Ax M x A1) (subst_Ax M x B)
    | family_app A1 M1 => family_app (subst_Ax M x A1) (subst_Mx M x M1)
  end

with subst_Mx (M : object) (x : var) (M1 : object) {struct M1} : object :=
  match M1 with
    | object_const c => object_const c
    | object_var x1 => if (x == x1) then (M) else (object_var x1)
    | object_var_ind n => object_var_ind n
    | object_abs A M2 => object_abs (subst_Ax M x A) (subst_Mx M x M2)
    | object_app M2 N => object_app (subst_Mx M x M2) (subst_Mx M x N)
  end.

Fixpoint subst_Kx (M : object) (x : var) (K : kind) {struct K} : kind :=
  match K with
    | kind_type => kind_type
    | kind_pi A K1 => kind_pi (subst_Ax M x A) (subst_Kx M x K1)
  end.


(* *********************************************************************** *)
(** * Degree *)

(** These define only an upper bound, not a strict upper bound. *)

Inductive degree_Ax : nat -> family -> Set :=
  | degree_Ax_const : forall n a,
    degree_Ax n (family_const a)
  | degree_Ax_pi : forall n A B,
    degree_Ax n A ->
    degree_Ax (S n) B ->
    degree_Ax n (family_pi A B)
  | degree_Ax_abs : forall n A B,
    degree_Ax n A ->
    degree_Ax (S n) B ->
    degree_Ax n (family_abs A B)
  | degree_Ax_app : forall n A M,
    degree_Ax n A ->
    degree_Mx n M ->
    degree_Ax n (family_app A M)

with degree_Mx : nat -> object -> Set :=
  | degree_Mx_const : forall n c,
    degree_Mx n (object_const c)
  | degree_Mx_var : forall n x,
    degree_Mx n (object_var x)
  | degree_Mx_var_ind : forall n n1,
    lt n1 n ->
    degree_Mx n (object_var_ind n1)
  | degree_Mx_abs : forall n A M,
    degree_Ax n A ->
    degree_Mx (S n) M ->
    degree_Mx n (object_abs A M)
  | degree_Mx_app : forall n M N,
    degree_Mx n M ->
    degree_Mx n N ->
    degree_Mx n (object_app M N).

Scheme degree_Ax_ind' := Induction for degree_Ax Sort Prop
  with degree_Mx_ind' := Induction for degree_Mx Sort Prop.

Definition degree_Ax_degree_Mx_mutind :=
  fun H H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 =>
  (conj (degree_Ax_ind' H H1 H2 H3 H4 H5 H6 H7 H8 H9 H10)
  (degree_Mx_ind' H H1 H2 H3 H4 H5 H6 H7 H8 H9 H10)).

Scheme degree_Ax_rec' := Induction for degree_Ax Sort Set
  with degree_Mx_rec' := Induction for degree_Mx Sort Set.

Definition degree_Ax_degree_Mx_mutrec :=
  fun H H1 H2 H3 H4 H5 H6 H7 H8 H9 H10 =>
  (pair (degree_Ax_rec' H H1 H2 H3 H4 H5 H6 H7 H8 H9 H10)
  (degree_Mx_rec' H H1 H2 H3 H4 H5 H6 H7 H8 H9 H10)).

Hint Constructors degree_Ax : core lngen.

Hint Constructors degree_Mx : core lngen.

Inductive degree_Kx : nat -> kind -> Set :=
  | degree_Kx_type : forall n,
    degree_Kx n (kind_type)
  | degree_Kx_pi : forall n A K,
    degree_Ax n A ->
    degree_Kx (S n) K ->
    degree_Kx n (kind_pi A K).

Scheme degree_Kx_ind' := Induction for degree_Kx Sort Prop.

Definition degree_Kx_mutind :=
  fun H H1 H2 =>
  degree_Kx_ind' H H1 H2.

Scheme degree_Kx_rec' := Induction for degree_Kx Sort Set.

Definition degree_Kx_mutrec :=
  fun H H1 H2 =>
  degree_Kx_rec' H H1 H2.

Hint Constructors degree_Kx : core lngen.


(* *********************************************************************** *)
(** * Local closure *)

Inductive lc_A : family -> Set :=
  | lc_A_const : forall a,
    lc_A (family_const a)
  | lc_A_pi : forall A B,
    lc_A A ->
    (forall x : var, lc_A (open_Ax (object_var x) B)) ->
    lc_A (family_pi A B)
  | lc_A_abs : forall A B,
    lc_A A ->
    (forall x : var, lc_A (open_Ax (object_var x) B)) ->
    lc_A (family_abs A B)
  | lc_A_app : forall A M,
    lc_A A ->
    lc_M M ->
    lc_A (family_app A M)

with lc_M : object -> Set :=
  | lc_M_const : forall c,
    lc_M (object_const c)
  | lc_M_var : forall x,
    lc_M (object_var x)
  | lc_M_abs : forall A M,
    lc_A A ->
    (forall x : var, lc_M (open_Mx (object_var x) M)) ->
    lc_M (object_abs A M)
  | lc_M_app : forall M N,
    lc_M M ->
    lc_M N ->
    lc_M (object_app M N).

Scheme lc_A_ind' := Induction for lc_A Sort Prop
  with lc_M_ind' := Induction for lc_M Sort Prop.

Definition lc_A_lc_M_mutind :=
  fun H H1 H2 H3 H4 H5 H6 H7 H8 H9 =>
  (conj (lc_A_ind' H H1 H2 H3 H4 H5 H6 H7 H8 H9)
  (lc_M_ind' H H1 H2 H3 H4 H5 H6 H7 H8 H9)).

Scheme lc_A_rec' := Induction for lc_A Sort Set
  with lc_M_rec' := Induction for lc_M Sort Set.

Definition lc_A_lc_M_mutrec :=
  fun H H1 H2 H3 H4 H5 H6 H7 H8 H9 =>
  (pair (lc_A_rec' H H1 H2 H3 H4 H5 H6 H7 H8 H9)
  (lc_M_rec' H H1 H2 H3 H4 H5 H6 H7 H8 H9)).

Hint Constructors lc_A : core lngen.

Hint Constructors lc_M : core lngen.

Inductive lc_K : kind -> Set :=
  | lc_K_type :
    lc_K (kind_type)
  | lc_K_pi : forall A K,
    lc_A A ->
    (forall x : var, lc_K (open_Kx (object_var x) K)) ->
    lc_K (kind_pi A K).

Scheme lc_K_ind' := Induction for lc_K Sort Prop.

Definition lc_K_mutind :=
  fun H H1 H2 =>
  lc_K_ind' H H1 H2.

Scheme lc_K_rec' := Induction for lc_K Sort Set.

Definition lc_K_mutrec :=
  fun H H1 H2 =>
  lc_K_rec' H H1 H2.

Hint Constructors lc_K : core lngen.


(* *********************************************************************** *)
(** * Tactic support *)

(** Tactic definitions. *)

Ltac gather_strings :=
  let S := ltac_map (fun x : string => singleton x) in
  let S1 := ltac_map (fun x : fset string => x) in
  let S2 := ltac_map (fun x : family => fv_Ax x) in
  let S3 := ltac_map (fun x : kind => fv_Kx x) in
  let S4 := ltac_map (fun x : object => fv_Mx x) in
  simplify_list_of_atom_sets (S ++ S1 ++ S2 ++ S3 ++ S4).

Ltac pick_fresh_var x :=
  let L := gather_strings in
  (pick fresh x for L).

Tactic Notation "pick" "fresh" "var" ident(x) :=
  let L := gather_strings in
  (pick fresh x for L).

(** Additional hint declarations. *)

Hint Resolve @plus_le_compat : lngen. (* from Coq.Arith.Plus *)

(** Redefine [default_auto] and [default_autorewrite]. *)

Ltac default_auto ::= auto with arith set brute_force lngen.
Ltac default_autorewrite ::= autorewrite with lngen in *.


(* *********************************************************************** *)
(** * Theorems about [size] *)

(* begin hide *)

Lemma size_A_min_size_M_min_mutual :
(forall A, 1 <= size_A A) /\
(forall M, 1 <= size_M M).
Proof.
apply_mutual_ind family_object_mutind;
default_simp.
Qed.

(* end hide *)

Lemma size_A_min :
forall A, 1 <= size_A A.
Proof.
pose proof size_A_min_size_M_min_mutual as H; intuition eauto.
Qed.

Hint Resolve size_A_min : lngen.

Lemma size_M_min :
forall M, 1 <= size_M M.
Proof.
pose proof size_A_min_size_M_min_mutual as H; intuition eauto.
Qed.

Hint Resolve size_M_min : lngen.

(* begin hide *)

Lemma size_K_min_mutual :
(forall K, 1 <= size_K K).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

Lemma size_K_min :
forall K, 1 <= size_K K.
Proof.
pose proof size_K_min_mutual as H; intuition eauto.
Qed.

Hint Resolve size_K_min : lngen.

(* begin hide *)

Lemma size_A_close_Ax_rec_size_M_close_Mx_rec_mutual :
(forall A x n,
  size_A (close_Ax_rec n x A) = size_A A) /\
(forall M x n,
  size_M (close_Mx_rec n x M) = size_M M).
Proof.
apply_mutual_ind family_object_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_A_close_Ax_rec :
forall A x n,
  size_A (close_Ax_rec n x A) = size_A A.
Proof.
pose proof size_A_close_Ax_rec_size_M_close_Mx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_A_close_Ax_rec : lngen.
Hint Rewrite size_A_close_Ax_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma size_M_close_Mx_rec :
forall M x n,
  size_M (close_Mx_rec n x M) = size_M M.
Proof.
pose proof size_A_close_Ax_rec_size_M_close_Mx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_M_close_Mx_rec : lngen.
Hint Rewrite size_M_close_Mx_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma size_K_close_Kx_rec_mutual :
(forall K x n,
  size_K (close_Kx_rec n x K) = size_K K).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_K_close_Kx_rec :
forall K x n,
  size_K (close_Kx_rec n x K) = size_K K.
Proof.
pose proof size_K_close_Kx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_K_close_Kx_rec : lngen.
Hint Rewrite size_K_close_Kx_rec using solve [auto] : lngen.

(* end hide *)

Lemma size_A_close_Ax :
forall A x,
  size_A (close_Ax x A) = size_A A.
Proof.
unfold close_Ax; default_simp.
Qed.

Hint Resolve size_A_close_Ax : lngen.
Hint Rewrite size_A_close_Ax using solve [auto] : lngen.

Lemma size_M_close_Mx :
forall M x,
  size_M (close_Mx x M) = size_M M.
Proof.
unfold close_Mx; default_simp.
Qed.

Hint Resolve size_M_close_Mx : lngen.
Hint Rewrite size_M_close_Mx using solve [auto] : lngen.

Lemma size_K_close_Kx :
forall K x,
  size_K (close_Kx x K) = size_K K.
Proof.
unfold close_Kx; default_simp.
Qed.

Hint Resolve size_K_close_Kx : lngen.
Hint Rewrite size_K_close_Kx using solve [auto] : lngen.

(* begin hide *)

Lemma size_A_open_Ax_rec_size_M_open_Mx_rec_mutual :
(forall A M n,
  size_A A <= size_A (open_Ax_rec n M A)) /\
(forall M M1 n,
  size_M M <= size_M (open_Mx_rec n M1 M)).
Proof.
apply_mutual_ind family_object_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_A_open_Ax_rec :
forall A M n,
  size_A A <= size_A (open_Ax_rec n M A).
Proof.
pose proof size_A_open_Ax_rec_size_M_open_Mx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_A_open_Ax_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma size_M_open_Mx_rec :
forall M M1 n,
  size_M M <= size_M (open_Mx_rec n M1 M).
Proof.
pose proof size_A_open_Ax_rec_size_M_open_Mx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_M_open_Mx_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma size_K_open_Kx_rec_mutual :
(forall K M n,
  size_K K <= size_K (open_Kx_rec n M K)).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_K_open_Kx_rec :
forall K M n,
  size_K K <= size_K (open_Kx_rec n M K).
Proof.
pose proof size_K_open_Kx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve size_K_open_Kx_rec : lngen.

(* end hide *)

Lemma size_A_open_Ax :
forall A M,
  size_A A <= size_A (open_Ax M A).
Proof.
unfold open_Ax; default_simp.
Qed.

Hint Resolve size_A_open_Ax : lngen.

Lemma size_M_open_Mx :
forall M M1,
  size_M M <= size_M (open_Mx M1 M).
Proof.
unfold open_Mx; default_simp.
Qed.

Hint Resolve size_M_open_Mx : lngen.

Lemma size_K_open_Kx :
forall K M,
  size_K K <= size_K (open_Kx M K).
Proof.
unfold open_Kx; default_simp.
Qed.

Hint Resolve size_K_open_Kx : lngen.

(* begin hide *)

Lemma size_A_open_Ax_rec_var_size_M_open_Mx_rec_var_mutual :
(forall A x n,
  size_A (open_Ax_rec n (object_var x) A) = size_A A) /\
(forall M x n,
  size_M (open_Mx_rec n (object_var x) M) = size_M M).
Proof.
apply_mutual_ind family_object_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_A_open_Ax_rec_var :
forall A x n,
  size_A (open_Ax_rec n (object_var x) A) = size_A A.
Proof.
pose proof size_A_open_Ax_rec_var_size_M_open_Mx_rec_var_mutual as H; intuition eauto.
Qed.

Hint Resolve size_A_open_Ax_rec_var : lngen.
Hint Rewrite size_A_open_Ax_rec_var using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma size_M_open_Mx_rec_var :
forall M x n,
  size_M (open_Mx_rec n (object_var x) M) = size_M M.
Proof.
pose proof size_A_open_Ax_rec_var_size_M_open_Mx_rec_var_mutual as H; intuition eauto.
Qed.

Hint Resolve size_M_open_Mx_rec_var : lngen.
Hint Rewrite size_M_open_Mx_rec_var using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma size_K_open_Kx_rec_var_mutual :
(forall K x n,
  size_K (open_Kx_rec n (object_var x) K) = size_K K).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma size_K_open_Kx_rec_var :
forall K x n,
  size_K (open_Kx_rec n (object_var x) K) = size_K K.
Proof.
pose proof size_K_open_Kx_rec_var_mutual as H; intuition eauto.
Qed.

Hint Resolve size_K_open_Kx_rec_var : lngen.
Hint Rewrite size_K_open_Kx_rec_var using solve [auto] : lngen.

(* end hide *)

Lemma size_A_open_Ax_var :
forall A x,
  size_A (open_Ax (object_var x) A) = size_A A.
Proof.
unfold open_Ax; default_simp.
Qed.

Hint Resolve size_A_open_Ax_var : lngen.
Hint Rewrite size_A_open_Ax_var using solve [auto] : lngen.

Lemma size_M_open_Mx_var :
forall M x,
  size_M (open_Mx (object_var x) M) = size_M M.
Proof.
unfold open_Mx; default_simp.
Qed.

Hint Resolve size_M_open_Mx_var : lngen.
Hint Rewrite size_M_open_Mx_var using solve [auto] : lngen.

Lemma size_K_open_Kx_var :
forall K x,
  size_K (open_Kx (object_var x) K) = size_K K.
Proof.
unfold open_Kx; default_simp.
Qed.

Hint Resolve size_K_open_Kx_var : lngen.
Hint Rewrite size_K_open_Kx_var using solve [auto] : lngen.


(* *********************************************************************** *)
(** * Theorems about [degree] *)

(* begin hide *)

Lemma degree_Ax_S_degree_Mx_S_mutual :
(forall n A,
  degree_Ax n A ->
  degree_Ax (S n) A) *
(forall n M,
  degree_Mx n M ->
  degree_Mx (S n) M).
Proof.
apply_mutual_ind degree_Ax_degree_Mx_mutrec;
default_simp.
Qed.

(* end hide *)

Lemma degree_Ax_S :
forall n A,
  degree_Ax n A ->
  degree_Ax (S n) A.
Proof.
pose proof degree_Ax_S_degree_Mx_S_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_Ax_S : lngen.

Lemma degree_Mx_S :
forall n M,
  degree_Mx n M ->
  degree_Mx (S n) M.
Proof.
pose proof degree_Ax_S_degree_Mx_S_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_Mx_S : lngen.

(* begin hide *)

Lemma degree_Kx_S_mutual :
(forall n K,
  degree_Kx n K ->
  degree_Kx (S n) K).
Proof.
apply_mutual_ind degree_Kx_mutrec;
default_simp.
Qed.

(* end hide *)

Lemma degree_Kx_S :
forall n K,
  degree_Kx n K ->
  degree_Kx (S n) K.
Proof.
pose proof degree_Kx_S_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_Kx_S : lngen.

Lemma degree_Ax_O :
forall n A,
  degree_Ax O A ->
  degree_Ax n A.
Proof.
induction n; default_simp.
Qed.

Hint Resolve degree_Ax_O : lngen.

Lemma degree_Mx_O :
forall n M,
  degree_Mx O M ->
  degree_Mx n M.
Proof.
induction n; default_simp.
Qed.

Hint Resolve degree_Mx_O : lngen.

Lemma degree_Kx_O :
forall n K,
  degree_Kx O K ->
  degree_Kx n K.
Proof.
induction n; default_simp.
Qed.

Hint Resolve degree_Kx_O : lngen.

(* begin hide *)

Lemma degree_Ax_close_Ax_rec_degree_Mx_close_Mx_rec_mutual :
(forall A x n,
  degree_Ax n A ->
  degree_Ax (S n) (close_Ax_rec n x A)) *
(forall M x n,
  degree_Mx n M ->
  degree_Mx (S n) (close_Mx_rec n x M)).
Proof.
apply_mutual_ind family_object_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_Ax_close_Ax_rec :
forall A x n,
  degree_Ax n A ->
  degree_Ax (S n) (close_Ax_rec n x A).
Proof.
pose proof degree_Ax_close_Ax_rec_degree_Mx_close_Mx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_Ax_close_Ax_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_Mx_close_Mx_rec :
forall M x n,
  degree_Mx n M ->
  degree_Mx (S n) (close_Mx_rec n x M).
Proof.
pose proof degree_Ax_close_Ax_rec_degree_Mx_close_Mx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_Mx_close_Mx_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_Kx_close_Kx_rec_mutual :
(forall K x n,
  degree_Kx n K ->
  degree_Kx (S n) (close_Kx_rec n x K)).
Proof.
apply_mutual_ind kind_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_Kx_close_Kx_rec :
forall K x n,
  degree_Kx n K ->
  degree_Kx (S n) (close_Kx_rec n x K).
Proof.
pose proof degree_Kx_close_Kx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_Kx_close_Kx_rec : lngen.

(* end hide *)

Lemma degree_Ax_close_Ax :
forall A x,
  degree_Ax 0 A ->
  degree_Ax 1 (close_Ax x A).
Proof.
unfold close_Ax; default_simp.
Qed.

Hint Resolve degree_Ax_close_Ax : lngen.

Lemma degree_Mx_close_Mx :
forall M x,
  degree_Mx 0 M ->
  degree_Mx 1 (close_Mx x M).
Proof.
unfold close_Mx; default_simp.
Qed.

Hint Resolve degree_Mx_close_Mx : lngen.

Lemma degree_Kx_close_Kx :
forall K x,
  degree_Kx 0 K ->
  degree_Kx 1 (close_Kx x K).
Proof.
unfold close_Kx; default_simp.
Qed.

Hint Resolve degree_Kx_close_Kx : lngen.

(* begin hide *)

Lemma degree_Ax_close_Ax_rec_inv_degree_Mx_close_Mx_rec_inv_mutual :
(forall A x n,
  degree_Ax (S n) (close_Ax_rec n x A) ->
  degree_Ax n A) *
(forall M x n,
  degree_Mx (S n) (close_Mx_rec n x M) ->
  degree_Mx n M).
Proof.
apply_mutual_ind family_object_mutrec;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_Ax_close_Ax_rec_inv :
forall A x n,
  degree_Ax (S n) (close_Ax_rec n x A) ->
  degree_Ax n A.
Proof.
pose proof degree_Ax_close_Ax_rec_inv_degree_Mx_close_Mx_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_Ax_close_Ax_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_Mx_close_Mx_rec_inv :
forall M x n,
  degree_Mx (S n) (close_Mx_rec n x M) ->
  degree_Mx n M.
Proof.
pose proof degree_Ax_close_Ax_rec_inv_degree_Mx_close_Mx_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_Mx_close_Mx_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_Kx_close_Kx_rec_inv_mutual :
(forall K x n,
  degree_Kx (S n) (close_Kx_rec n x K) ->
  degree_Kx n K).
Proof.
apply_mutual_ind kind_mutrec;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_Kx_close_Kx_rec_inv :
forall K x n,
  degree_Kx (S n) (close_Kx_rec n x K) ->
  degree_Kx n K.
Proof.
pose proof degree_Kx_close_Kx_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_Kx_close_Kx_rec_inv : lngen.

(* end hide *)

Lemma degree_Ax_close_Ax_inv :
forall A x,
  degree_Ax 1 (close_Ax x A) ->
  degree_Ax 0 A.
Proof.
unfold close_Ax; eauto 2 with lngen.
Qed.

Hint Immediate degree_Ax_close_Ax_inv : lngen.

Lemma degree_Mx_close_Mx_inv :
forall M x,
  degree_Mx 1 (close_Mx x M) ->
  degree_Mx 0 M.
Proof.
unfold close_Mx; eauto 2 with lngen.
Qed.

Hint Immediate degree_Mx_close_Mx_inv : lngen.

Lemma degree_Kx_close_Kx_inv :
forall K x,
  degree_Kx 1 (close_Kx x K) ->
  degree_Kx 0 K.
Proof.
unfold close_Kx; eauto 2 with lngen.
Qed.

Hint Immediate degree_Kx_close_Kx_inv : lngen.

(* begin hide *)

Lemma degree_Ax_open_Ax_rec_degree_Mx_open_Mx_rec_mutual :
(forall A M n,
  degree_Ax (S n) A ->
  degree_Mx n M ->
  degree_Ax n (open_Ax_rec n M A)) *
(forall M M1 n,
  degree_Mx (S n) M ->
  degree_Mx n M1 ->
  degree_Mx n (open_Mx_rec n M1 M)).
Proof.
apply_mutual_ind family_object_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_Ax_open_Ax_rec :
forall A M n,
  degree_Ax (S n) A ->
  degree_Mx n M ->
  degree_Ax n (open_Ax_rec n M A).
Proof.
pose proof degree_Ax_open_Ax_rec_degree_Mx_open_Mx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_Ax_open_Ax_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_Mx_open_Mx_rec :
forall M M1 n,
  degree_Mx (S n) M ->
  degree_Mx n M1 ->
  degree_Mx n (open_Mx_rec n M1 M).
Proof.
pose proof degree_Ax_open_Ax_rec_degree_Mx_open_Mx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_Mx_open_Mx_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_Kx_open_Kx_rec_mutual :
(forall K M n,
  degree_Kx (S n) K ->
  degree_Mx n M ->
  degree_Kx n (open_Kx_rec n M K)).
Proof.
apply_mutual_ind kind_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_Kx_open_Kx_rec :
forall K M n,
  degree_Kx (S n) K ->
  degree_Mx n M ->
  degree_Kx n (open_Kx_rec n M K).
Proof.
pose proof degree_Kx_open_Kx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_Kx_open_Kx_rec : lngen.

(* end hide *)

Lemma degree_Ax_open_Ax :
forall A M,
  degree_Ax 1 A ->
  degree_Mx 0 M ->
  degree_Ax 0 (open_Ax M A).
Proof.
unfold open_Ax; default_simp.
Qed.

Hint Resolve degree_Ax_open_Ax : lngen.

Lemma degree_Mx_open_Mx :
forall M M1,
  degree_Mx 1 M ->
  degree_Mx 0 M1 ->
  degree_Mx 0 (open_Mx M1 M).
Proof.
unfold open_Mx; default_simp.
Qed.

Hint Resolve degree_Mx_open_Mx : lngen.

Lemma degree_Kx_open_Kx :
forall K M,
  degree_Kx 1 K ->
  degree_Mx 0 M ->
  degree_Kx 0 (open_Kx M K).
Proof.
unfold open_Kx; default_simp.
Qed.

Hint Resolve degree_Kx_open_Kx : lngen.

(* begin hide *)

Lemma degree_Ax_open_Ax_rec_inv_degree_Mx_open_Mx_rec_inv_mutual :
(forall A M n,
  degree_Ax n (open_Ax_rec n M A) ->
  degree_Ax (S n) A) *
(forall M M1 n,
  degree_Mx n (open_Mx_rec n M1 M) ->
  degree_Mx (S n) M).
Proof.
apply_mutual_ind family_object_mutrec;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_Ax_open_Ax_rec_inv :
forall A M n,
  degree_Ax n (open_Ax_rec n M A) ->
  degree_Ax (S n) A.
Proof.
pose proof degree_Ax_open_Ax_rec_inv_degree_Mx_open_Mx_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_Ax_open_Ax_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_Mx_open_Mx_rec_inv :
forall M M1 n,
  degree_Mx n (open_Mx_rec n M1 M) ->
  degree_Mx (S n) M.
Proof.
pose proof degree_Ax_open_Ax_rec_inv_degree_Mx_open_Mx_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_Mx_open_Mx_rec_inv : lngen.

(* end hide *)

(* begin hide *)

Lemma degree_Kx_open_Kx_rec_inv_mutual :
(forall K M n,
  degree_Kx n (open_Kx_rec n M K) ->
  degree_Kx (S n) K).
Proof.
apply_mutual_ind kind_mutrec;
default_simp; eauto with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma degree_Kx_open_Kx_rec_inv :
forall K M n,
  degree_Kx n (open_Kx_rec n M K) ->
  degree_Kx (S n) K.
Proof.
pose proof degree_Kx_open_Kx_rec_inv_mutual as H; intuition eauto.
Qed.

Hint Immediate degree_Kx_open_Kx_rec_inv : lngen.

(* end hide *)

Lemma degree_Ax_open_Ax_inv :
forall A M,
  degree_Ax 0 (open_Ax M A) ->
  degree_Ax 1 A.
Proof.
unfold open_Ax; eauto 2 with lngen.
Qed.

Hint Resolve degree_Ax_open_Ax_inv : lngen.

Lemma degree_Mx_open_Mx_inv :
forall M M1,
  degree_Mx 0 (open_Mx M1 M) ->
  degree_Mx 1 M.
Proof.
unfold open_Mx; eauto 2 with lngen.
Qed.

Hint Resolve degree_Mx_open_Mx_inv : lngen.

Lemma degree_Kx_open_Kx_inv :
forall K M,
  degree_Kx 0 (open_Kx M K) ->
  degree_Kx 1 K.
Proof.
unfold open_Kx; eauto 2 with lngen.
Qed.

Hint Resolve degree_Kx_open_Kx_inv : lngen.


(* *********************************************************************** *)
(** * Theorems about [open] and [close] *)

(* begin hide *)

Lemma close_Ax_rec_inj_close_Mx_rec_inj_mutual :
(forall A1 A x n,
  close_Ax_rec n x A1 = close_Ax_rec n x A ->
  A1 = A) /\
(forall M1 M x n,
  close_Mx_rec n x M1 = close_Mx_rec n x M ->
  M1 = M).
Proof.
apply_mutual_ind family_object_mutind;
intros; (destruct A || destruct M);
default_simp; eauto 2 with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_Ax_rec_inj :
forall A1 A x n,
  close_Ax_rec n x A1 = close_Ax_rec n x A ->
  A1 = A.
Proof.
pose proof close_Ax_rec_inj_close_Mx_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate close_Ax_rec_inj : lngen.

(* end hide *)

(* begin hide *)

Lemma close_Mx_rec_inj :
forall M1 M x n,
  close_Mx_rec n x M1 = close_Mx_rec n x M ->
  M1 = M.
Proof.
pose proof close_Ax_rec_inj_close_Mx_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate close_Mx_rec_inj : lngen.

(* end hide *)

(* begin hide *)

Lemma close_Kx_rec_inj_mutual :
(forall K1 K x n,
  close_Kx_rec n x K1 = close_Kx_rec n x K ->
  K1 = K).
Proof.
apply_mutual_ind kind_mutind;
intros; (destruct K);
default_simp; eauto 2 with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_Kx_rec_inj :
forall K1 K x n,
  close_Kx_rec n x K1 = close_Kx_rec n x K ->
  K1 = K.
Proof.
pose proof close_Kx_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate close_Kx_rec_inj : lngen.

(* end hide *)

Lemma close_Ax_inj :
forall A1 A x,
  close_Ax x A1 = close_Ax x A ->
  A1 = A.
Proof.
unfold close_Ax; eauto 2 with lngen.
Qed.

Hint Immediate close_Ax_inj : lngen.

Lemma close_Mx_inj :
forall M1 M x,
  close_Mx x M1 = close_Mx x M ->
  M1 = M.
Proof.
unfold close_Mx; eauto 2 with lngen.
Qed.

Hint Immediate close_Mx_inj : lngen.

Lemma close_Kx_inj :
forall K1 K x,
  close_Kx x K1 = close_Kx x K ->
  K1 = K.
Proof.
unfold close_Kx; eauto 2 with lngen.
Qed.

Hint Immediate close_Kx_inj : lngen.

(* begin hide *)

Lemma close_Ax_rec_open_Ax_rec_close_Mx_rec_open_Mx_rec_mutual :
(forall A x n,
  x `notin` fv_Ax A ->
  close_Ax_rec n x (open_Ax_rec n (object_var x) A) = A) /\
(forall M x n,
  x `notin` fv_Mx M ->
  close_Mx_rec n x (open_Mx_rec n (object_var x) M) = M).
Proof.
apply_mutual_ind family_object_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_Ax_rec_open_Ax_rec :
forall A x n,
  x `notin` fv_Ax A ->
  close_Ax_rec n x (open_Ax_rec n (object_var x) A) = A.
Proof.
pose proof close_Ax_rec_open_Ax_rec_close_Mx_rec_open_Mx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve close_Ax_rec_open_Ax_rec : lngen.
Hint Rewrite close_Ax_rec_open_Ax_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma close_Mx_rec_open_Mx_rec :
forall M x n,
  x `notin` fv_Mx M ->
  close_Mx_rec n x (open_Mx_rec n (object_var x) M) = M.
Proof.
pose proof close_Ax_rec_open_Ax_rec_close_Mx_rec_open_Mx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve close_Mx_rec_open_Mx_rec : lngen.
Hint Rewrite close_Mx_rec_open_Mx_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma close_Kx_rec_open_Kx_rec_mutual :
(forall K x n,
  x `notin` fv_Kx K ->
  close_Kx_rec n x (open_Kx_rec n (object_var x) K) = K).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_Kx_rec_open_Kx_rec :
forall K x n,
  x `notin` fv_Kx K ->
  close_Kx_rec n x (open_Kx_rec n (object_var x) K) = K.
Proof.
pose proof close_Kx_rec_open_Kx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve close_Kx_rec_open_Kx_rec : lngen.
Hint Rewrite close_Kx_rec_open_Kx_rec using solve [auto] : lngen.

(* end hide *)

Lemma close_Ax_open_Ax :
forall A x,
  x `notin` fv_Ax A ->
  close_Ax x (open_Ax (object_var x) A) = A.
Proof.
unfold close_Ax; unfold open_Ax; default_simp.
Qed.

Hint Resolve close_Ax_open_Ax : lngen.
Hint Rewrite close_Ax_open_Ax using solve [auto] : lngen.

Lemma close_Mx_open_Mx :
forall M x,
  x `notin` fv_Mx M ->
  close_Mx x (open_Mx (object_var x) M) = M.
Proof.
unfold close_Mx; unfold open_Mx; default_simp.
Qed.

Hint Resolve close_Mx_open_Mx : lngen.
Hint Rewrite close_Mx_open_Mx using solve [auto] : lngen.

Lemma close_Kx_open_Kx :
forall K x,
  x `notin` fv_Kx K ->
  close_Kx x (open_Kx (object_var x) K) = K.
Proof.
unfold close_Kx; unfold open_Kx; default_simp.
Qed.

Hint Resolve close_Kx_open_Kx : lngen.
Hint Rewrite close_Kx_open_Kx using solve [auto] : lngen.

(* begin hide *)

Lemma open_Ax_rec_close_Ax_rec_open_Mx_rec_close_Mx_rec_mutual :
(forall A x n,
  open_Ax_rec n (object_var x) (close_Ax_rec n x A) = A) /\
(forall M x n,
  open_Mx_rec n (object_var x) (close_Mx_rec n x M) = M).
Proof.
apply_mutual_ind family_object_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_Ax_rec_close_Ax_rec :
forall A x n,
  open_Ax_rec n (object_var x) (close_Ax_rec n x A) = A.
Proof.
pose proof open_Ax_rec_close_Ax_rec_open_Mx_rec_close_Mx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve open_Ax_rec_close_Ax_rec : lngen.
Hint Rewrite open_Ax_rec_close_Ax_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma open_Mx_rec_close_Mx_rec :
forall M x n,
  open_Mx_rec n (object_var x) (close_Mx_rec n x M) = M.
Proof.
pose proof open_Ax_rec_close_Ax_rec_open_Mx_rec_close_Mx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve open_Mx_rec_close_Mx_rec : lngen.
Hint Rewrite open_Mx_rec_close_Mx_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma open_Kx_rec_close_Kx_rec_mutual :
(forall K x n,
  open_Kx_rec n (object_var x) (close_Kx_rec n x K) = K).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_Kx_rec_close_Kx_rec :
forall K x n,
  open_Kx_rec n (object_var x) (close_Kx_rec n x K) = K.
Proof.
pose proof open_Kx_rec_close_Kx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve open_Kx_rec_close_Kx_rec : lngen.
Hint Rewrite open_Kx_rec_close_Kx_rec using solve [auto] : lngen.

(* end hide *)

Lemma open_Ax_close_Ax :
forall A x,
  open_Ax (object_var x) (close_Ax x A) = A.
Proof.
unfold close_Ax; unfold open_Ax; default_simp.
Qed.

Hint Resolve open_Ax_close_Ax : lngen.
Hint Rewrite open_Ax_close_Ax using solve [auto] : lngen.

Lemma open_Mx_close_Mx :
forall M x,
  open_Mx (object_var x) (close_Mx x M) = M.
Proof.
unfold close_Mx; unfold open_Mx; default_simp.
Qed.

Hint Resolve open_Mx_close_Mx : lngen.
Hint Rewrite open_Mx_close_Mx using solve [auto] : lngen.

Lemma open_Kx_close_Kx :
forall K x,
  open_Kx (object_var x) (close_Kx x K) = K.
Proof.
unfold close_Kx; unfold open_Kx; default_simp.
Qed.

Hint Resolve open_Kx_close_Kx : lngen.
Hint Rewrite open_Kx_close_Kx using solve [auto] : lngen.

(* begin hide *)

Lemma open_Ax_rec_inj_open_Mx_rec_inj_mutual :
(forall A1 A x n,
  x `notin` fv_Ax A1 ->
  x `notin` fv_Ax A ->
  open_Ax_rec n (object_var x) A1 = open_Ax_rec n (object_var x) A ->
  A1 = A) /\
(forall M1 M x n,
  x `notin` fv_Mx M1 ->
  x `notin` fv_Mx M ->
  open_Mx_rec n (object_var x) M1 = open_Mx_rec n (object_var x) M ->
  M1 = M).
Proof.
apply_mutual_ind family_object_mutind;
intros; (destruct A || destruct M);
default_simp; eauto 2 with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_Ax_rec_inj :
forall A1 A x n,
  x `notin` fv_Ax A1 ->
  x `notin` fv_Ax A ->
  open_Ax_rec n (object_var x) A1 = open_Ax_rec n (object_var x) A ->
  A1 = A.
Proof.
pose proof open_Ax_rec_inj_open_Mx_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate open_Ax_rec_inj : lngen.

(* end hide *)

(* begin hide *)

Lemma open_Mx_rec_inj :
forall M1 M x n,
  x `notin` fv_Mx M1 ->
  x `notin` fv_Mx M ->
  open_Mx_rec n (object_var x) M1 = open_Mx_rec n (object_var x) M ->
  M1 = M.
Proof.
pose proof open_Ax_rec_inj_open_Mx_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate open_Mx_rec_inj : lngen.

(* end hide *)

(* begin hide *)

Lemma open_Kx_rec_inj_mutual :
(forall K1 K x n,
  x `notin` fv_Kx K1 ->
  x `notin` fv_Kx K ->
  open_Kx_rec n (object_var x) K1 = open_Kx_rec n (object_var x) K ->
  K1 = K).
Proof.
apply_mutual_ind kind_mutind;
intros; (destruct K);
default_simp; eauto 2 with lngen.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_Kx_rec_inj :
forall K1 K x n,
  x `notin` fv_Kx K1 ->
  x `notin` fv_Kx K ->
  open_Kx_rec n (object_var x) K1 = open_Kx_rec n (object_var x) K ->
  K1 = K.
Proof.
pose proof open_Kx_rec_inj_mutual as H; intuition eauto.
Qed.

Hint Immediate open_Kx_rec_inj : lngen.

(* end hide *)

Lemma open_Ax_inj :
forall A1 A x,
  x `notin` fv_Ax A1 ->
  x `notin` fv_Ax A ->
  open_Ax (object_var x) A1 = open_Ax (object_var x) A ->
  A1 = A.
Proof.
unfold open_Ax; eauto 2 with lngen.
Qed.

Hint Immediate open_Ax_inj : lngen.

Lemma open_Mx_inj :
forall M1 M x,
  x `notin` fv_Mx M1 ->
  x `notin` fv_Mx M ->
  open_Mx (object_var x) M1 = open_Mx (object_var x) M ->
  M1 = M.
Proof.
unfold open_Mx; eauto 2 with lngen.
Qed.

Hint Immediate open_Mx_inj : lngen.

Lemma open_Kx_inj :
forall K1 K x,
  x `notin` fv_Kx K1 ->
  x `notin` fv_Kx K ->
  open_Kx (object_var x) K1 = open_Kx (object_var x) K ->
  K1 = K.
Proof.
unfold open_Kx; eauto 2 with lngen.
Qed.

Hint Immediate open_Kx_inj : lngen.


(* *********************************************************************** *)
(** * Theorems about [lc] *)

(* begin hide *)

Lemma degree_Ax_of_lc_A_degree_Mx_of_lc_M_mutual :
(forall A,
  lc_A A ->
  degree_Ax 0 A) *
(forall M,
  lc_M M ->
  degree_Mx 0 M).
Proof.
apply_mutual_ind lc_A_lc_M_mutrec;
intros;
let x := fresh "x" in pick_fresh_var x;
repeat (match goal with
          | H : _, H1 : _ |- _ => specialize H with H1
        end);
default_simp; eauto with lngen.
Qed.

(* end hide *)

Lemma degree_Ax_of_lc_A :
forall A,
  lc_A A ->
  degree_Ax 0 A.
Proof.
pose proof degree_Ax_of_lc_A_degree_Mx_of_lc_M_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_Ax_of_lc_A : lngen.

Lemma degree_Mx_of_lc_M :
forall M,
  lc_M M ->
  degree_Mx 0 M.
Proof.
pose proof degree_Ax_of_lc_A_degree_Mx_of_lc_M_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_Mx_of_lc_M : lngen.

(* begin hide *)

Lemma degree_Kx_of_lc_K_mutual :
(forall K,
  lc_K K ->
  degree_Kx 0 K).
Proof.
apply_mutual_ind lc_K_mutrec;
intros;
let x := fresh "x" in pick_fresh_var x;
repeat (match goal with
          | H : _, H1 : _ |- _ => specialize H with H1
        end);
default_simp; eauto with lngen.
Qed.

(* end hide *)

Lemma degree_Kx_of_lc_K :
forall K,
  lc_K K ->
  degree_Kx 0 K.
Proof.
pose proof degree_Kx_of_lc_K_mutual as H; intuition eauto.
Qed.

Hint Resolve degree_Kx_of_lc_K : lngen.

(* begin hide *)

Lemma lc_A_of_degree_lc_M_of_degree_size_mutual :
forall i,
(forall A,
  size_A A = i ->
  degree_Ax 0 A ->
  lc_A A) *
(forall M,
  size_M M = i ->
  degree_Mx 0 M ->
  lc_M M).
Proof.
intros i; pattern i; apply lt_wf_rec;
clear i; intros i H;
apply_mutual_ind family_object_mutrec;
default_simp;
(* non-trivial cases *)
constructor; default_simp; eapply_first_hyp;
(* instantiate the size *)
match goal with
  | |- _ = _ => reflexivity
  | _ => idtac
end;
(* everything should be easy now *)
default_simp.
Qed.

(* end hide *)

Lemma lc_A_of_degree :
forall A,
  degree_Ax 0 A ->
  lc_A A.
Proof.
intros A; intros;
pose proof (lc_A_of_degree_lc_M_of_degree_size_mutual (size_A A));
intuition eauto.
Qed.

Hint Resolve lc_A_of_degree : lngen.

Lemma lc_M_of_degree :
forall M,
  degree_Mx 0 M ->
  lc_M M.
Proof.
intros M; intros;
pose proof (lc_A_of_degree_lc_M_of_degree_size_mutual (size_M M));
intuition eauto.
Qed.

Hint Resolve lc_M_of_degree : lngen.

(* begin hide *)

Lemma lc_K_of_degree_size_mutual :
forall i,
(forall K,
  size_K K = i ->
  degree_Kx 0 K ->
  lc_K K).
Proof.
intros i; pattern i; apply lt_wf_rec;
clear i; intros i H;
apply_mutual_ind kind_mutrec;
default_simp;
(* non-trivial cases *)
constructor; default_simp; eapply_first_hyp;
(* instantiate the size *)
match goal with
  | |- _ = _ => reflexivity
  | _ => idtac
end;
(* everything should be easy now *)
default_simp.
Qed.

(* end hide *)

Lemma lc_K_of_degree :
forall K,
  degree_Kx 0 K ->
  lc_K K.
Proof.
intros K; intros;
pose proof (lc_K_of_degree_size_mutual (size_K K));
intuition eauto.
Qed.

Hint Resolve lc_K_of_degree : lngen.

Lemma lc_A_pi_exists :
forall x A B,
  lc_A A ->
  lc_A (open_Ax (object_var x) B) ->
  lc_A (family_pi A B).
Proof.
default_simp; eauto 10 with lngen.
Qed.

Hint Resolve lc_A_pi_exists : lngen.

Lemma lc_A_abs_exists :
forall x A B,
  lc_A A ->
  lc_A (open_Ax (object_var x) B) ->
  lc_A (family_abs A B).
Proof.
default_simp; eauto 10 with lngen.
Qed.

Hint Resolve lc_A_abs_exists : lngen.

Lemma lc_M_abs_exists :
forall x A M,
  lc_A A ->
  lc_M (open_Mx (object_var x) M) ->
  lc_M (object_abs A M).
Proof.
default_simp; eauto 10 with lngen.
Qed.

Hint Resolve lc_M_abs_exists : lngen.

Lemma lc_K_pi_exists :
forall x A K,
  lc_A A ->
  lc_K (open_Kx (object_var x) K) ->
  lc_K (kind_pi A K).
Proof.
default_simp; eauto 10 with lngen.
Qed.

Hint Resolve lc_K_pi_exists : lngen.

(* begin hide *)

Lemma lc_A_unique_lc_M_unique_mutual :
(forall A (proof proof1 : lc_A A), proof = proof1) /\
(forall M (proof proof1 : lc_M M), proof = proof1).
Proof.
apply_mutual_ind lc_A_lc_M_mutind;
intros; dependent destruction proof1; default_simp; auto using axiom_fun_ext_eq.
Qed.

(* end hide *)

Lemma lc_A_unique :
forall A (proof proof1 : lc_A A), proof = proof1.
Proof.
pose proof lc_A_unique_lc_M_unique_mutual as H; intuition eauto.
Qed.

Hint Resolve lc_A_unique : lngen.

Lemma lc_M_unique :
forall M (proof proof1 : lc_M M), proof = proof1.
Proof.
pose proof lc_A_unique_lc_M_unique_mutual as H; intuition eauto.
Qed.

Hint Resolve lc_M_unique : lngen.

(* begin hide *)

Lemma lc_K_unique_mutual :
(forall K (proof proof1 : lc_K K), proof = proof1).
Proof.
apply_mutual_ind lc_K_mutind;
intros; dependent destruction proof1; default_simp; auto using axiom_fun_ext_eq.
Qed.

(* end hide *)

Lemma lc_K_unique :
forall K (proof proof1 : lc_K K), proof = proof1.
Proof.
pose proof lc_K_unique_mutual as H; intuition eauto.
Qed.

Hint Resolve lc_K_unique : lngen.


(* *********************************************************************** *)
(** * More theorems about [open] and [close] *)

(* begin hide *)

Lemma close_Ax_rec_degree_Ax_close_Mx_rec_degree_Mx_mutual :
(forall A x n,
  degree_Ax n A ->
  x `notin` fv_Ax A ->
  close_Ax_rec n x A = A) /\
(forall M x n,
  degree_Mx n M ->
  x `notin` fv_Mx M ->
  close_Mx_rec n x M = M).
Proof.
apply_mutual_ind family_object_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_Ax_rec_degree_Ax :
forall A x n,
  degree_Ax n A ->
  x `notin` fv_Ax A ->
  close_Ax_rec n x A = A.
Proof.
pose proof close_Ax_rec_degree_Ax_close_Mx_rec_degree_Mx_mutual as H; intuition eauto.
Qed.

Hint Resolve close_Ax_rec_degree_Ax : lngen.
Hint Rewrite close_Ax_rec_degree_Ax using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma close_Mx_rec_degree_Mx :
forall M x n,
  degree_Mx n M ->
  x `notin` fv_Mx M ->
  close_Mx_rec n x M = M.
Proof.
pose proof close_Ax_rec_degree_Ax_close_Mx_rec_degree_Mx_mutual as H; intuition eauto.
Qed.

Hint Resolve close_Mx_rec_degree_Mx : lngen.
Hint Rewrite close_Mx_rec_degree_Mx using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma close_Kx_rec_degree_Kx_mutual :
(forall K x n,
  degree_Kx n K ->
  x `notin` fv_Kx K ->
  close_Kx_rec n x K = K).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma close_Kx_rec_degree_Kx :
forall K x n,
  degree_Kx n K ->
  x `notin` fv_Kx K ->
  close_Kx_rec n x K = K.
Proof.
pose proof close_Kx_rec_degree_Kx_mutual as H; intuition eauto.
Qed.

Hint Resolve close_Kx_rec_degree_Kx : lngen.
Hint Rewrite close_Kx_rec_degree_Kx using solve [auto] : lngen.

(* end hide *)

Lemma close_Ax_lc_A :
forall A x,
  lc_A A ->
  x `notin` fv_Ax A ->
  close_Ax x A = A.
Proof.
unfold close_Ax; default_simp.
Qed.

Hint Resolve close_Ax_lc_A : lngen.
Hint Rewrite close_Ax_lc_A using solve [auto] : lngen.

Lemma close_Mx_lc_M :
forall M x,
  lc_M M ->
  x `notin` fv_Mx M ->
  close_Mx x M = M.
Proof.
unfold close_Mx; default_simp.
Qed.

Hint Resolve close_Mx_lc_M : lngen.
Hint Rewrite close_Mx_lc_M using solve [auto] : lngen.

Lemma close_Kx_lc_K :
forall K x,
  lc_K K ->
  x `notin` fv_Kx K ->
  close_Kx x K = K.
Proof.
unfold close_Kx; default_simp.
Qed.

Hint Resolve close_Kx_lc_K : lngen.
Hint Rewrite close_Kx_lc_K using solve [auto] : lngen.

(* begin hide *)

Lemma open_Ax_rec_degree_Ax_open_Mx_rec_degree_Mx_mutual :
(forall A M n,
  degree_Ax n A ->
  open_Ax_rec n M A = A) /\
(forall M1 M n,
  degree_Mx n M1 ->
  open_Mx_rec n M M1 = M1).
Proof.
apply_mutual_ind family_object_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_Ax_rec_degree_Ax :
forall A M n,
  degree_Ax n A ->
  open_Ax_rec n M A = A.
Proof.
pose proof open_Ax_rec_degree_Ax_open_Mx_rec_degree_Mx_mutual as H; intuition eauto.
Qed.

Hint Resolve open_Ax_rec_degree_Ax : lngen.
Hint Rewrite open_Ax_rec_degree_Ax using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma open_Mx_rec_degree_Mx :
forall M1 M n,
  degree_Mx n M1 ->
  open_Mx_rec n M M1 = M1.
Proof.
pose proof open_Ax_rec_degree_Ax_open_Mx_rec_degree_Mx_mutual as H; intuition eauto.
Qed.

Hint Resolve open_Mx_rec_degree_Mx : lngen.
Hint Rewrite open_Mx_rec_degree_Mx using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma open_Kx_rec_degree_Kx_mutual :
(forall K M n,
  degree_Kx n K ->
  open_Kx_rec n M K = K).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma open_Kx_rec_degree_Kx :
forall K M n,
  degree_Kx n K ->
  open_Kx_rec n M K = K.
Proof.
pose proof open_Kx_rec_degree_Kx_mutual as H; intuition eauto.
Qed.

Hint Resolve open_Kx_rec_degree_Kx : lngen.
Hint Rewrite open_Kx_rec_degree_Kx using solve [auto] : lngen.

(* end hide *)

Lemma open_Ax_lc_A :
forall A M,
  lc_A A ->
  open_Ax M A = A.
Proof.
unfold open_Ax; default_simp.
Qed.

Hint Resolve open_Ax_lc_A : lngen.
Hint Rewrite open_Ax_lc_A using solve [auto] : lngen.

Lemma open_Mx_lc_M :
forall M1 M,
  lc_M M1 ->
  open_Mx M M1 = M1.
Proof.
unfold open_Mx; default_simp.
Qed.

Hint Resolve open_Mx_lc_M : lngen.
Hint Rewrite open_Mx_lc_M using solve [auto] : lngen.

Lemma open_Kx_lc_K :
forall K M,
  lc_K K ->
  open_Kx M K = K.
Proof.
unfold open_Kx; default_simp.
Qed.

Hint Resolve open_Kx_lc_K : lngen.
Hint Rewrite open_Kx_lc_K using solve [auto] : lngen.


(* *********************************************************************** *)
(** * Theorems about [subst] *)

(* begin hide *)

Lemma subst_Ax_close_Ax_rec_subst_Mx_close_Mx_rec_mutual :
(forall A M x x1 n,
  degree_Mx n M ->
  x <> x1 ->
  x1 `notin` fv_Mx M ->
  subst_Ax M x (close_Ax_rec n x1 A) = close_Ax_rec n x1 (subst_Ax M x A)) /\
(forall M1 M x x1 n,
  degree_Mx n M ->
  x <> x1 ->
  x1 `notin` fv_Mx M ->
  subst_Mx M x (close_Mx_rec n x1 M1) = close_Mx_rec n x1 (subst_Mx M x M1)).
Proof.
apply_mutual_ind family_object_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_Ax_close_Ax_rec :
forall A M x x1 n,
  degree_Mx n M ->
  x <> x1 ->
  x1 `notin` fv_Mx M ->
  subst_Ax M x (close_Ax_rec n x1 A) = close_Ax_rec n x1 (subst_Ax M x A).
Proof.
pose proof subst_Ax_close_Ax_rec_subst_Mx_close_Mx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_Ax_close_Ax_rec : lngen.

Lemma subst_Mx_close_Mx_rec :
forall M1 M x x1 n,
  degree_Mx n M ->
  x <> x1 ->
  x1 `notin` fv_Mx M ->
  subst_Mx M x (close_Mx_rec n x1 M1) = close_Mx_rec n x1 (subst_Mx M x M1).
Proof.
pose proof subst_Ax_close_Ax_rec_subst_Mx_close_Mx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_Mx_close_Mx_rec : lngen.

(* begin hide *)

Lemma subst_Kx_close_Kx_rec_mutual :
(forall K M x x1 n,
  degree_Mx n M ->
  x <> x1 ->
  x1 `notin` fv_Mx M ->
  subst_Kx M x (close_Kx_rec n x1 K) = close_Kx_rec n x1 (subst_Kx M x K)).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_Kx_close_Kx_rec :
forall K M x x1 n,
  degree_Mx n M ->
  x <> x1 ->
  x1 `notin` fv_Mx M ->
  subst_Kx M x (close_Kx_rec n x1 K) = close_Kx_rec n x1 (subst_Kx M x K).
Proof.
pose proof subst_Kx_close_Kx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_Kx_close_Kx_rec : lngen.

Lemma subst_Ax_close_Ax :
forall A M x x1,
  lc_M M ->  x <> x1 ->
  x1 `notin` fv_Mx M ->
  subst_Ax M x (close_Ax x1 A) = close_Ax x1 (subst_Ax M x A).
Proof.
unfold close_Ax; default_simp.
Qed.

Hint Resolve subst_Ax_close_Ax : lngen.

Lemma subst_Mx_close_Mx :
forall M1 M x x1,
  lc_M M ->  x <> x1 ->
  x1 `notin` fv_Mx M ->
  subst_Mx M x (close_Mx x1 M1) = close_Mx x1 (subst_Mx M x M1).
Proof.
unfold close_Mx; default_simp.
Qed.

Hint Resolve subst_Mx_close_Mx : lngen.

Lemma subst_Kx_close_Kx :
forall K M x x1,
  lc_M M ->  x <> x1 ->
  x1 `notin` fv_Mx M ->
  subst_Kx M x (close_Kx x1 K) = close_Kx x1 (subst_Kx M x K).
Proof.
unfold close_Kx; default_simp.
Qed.

Hint Resolve subst_Kx_close_Kx : lngen.

(* begin hide *)

Lemma subst_Ax_degree_Ax_subst_Mx_degree_Mx_mutual :
(forall A M x n,
  degree_Ax n A ->
  degree_Mx n M ->
  degree_Ax n (subst_Ax M x A)) *
(forall M M1 x n,
  degree_Mx n M ->
  degree_Mx n M1 ->
  degree_Mx n (subst_Mx M1 x M)).
Proof.
apply_mutual_ind family_object_mutrec;
default_simp.
Qed.

(* end hide *)

Lemma subst_Ax_degree_Ax :
forall A M x n,
  degree_Ax n A ->
  degree_Mx n M ->
  degree_Ax n (subst_Ax M x A).
Proof.
pose proof subst_Ax_degree_Ax_subst_Mx_degree_Mx_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_Ax_degree_Ax : lngen.

Lemma subst_Mx_degree_Mx :
forall M M1 x n,
  degree_Mx n M ->
  degree_Mx n M1 ->
  degree_Mx n (subst_Mx M1 x M).
Proof.
pose proof subst_Ax_degree_Ax_subst_Mx_degree_Mx_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_Mx_degree_Mx : lngen.

(* begin hide *)

Lemma subst_Kx_degree_Kx_mutual :
(forall K M x n,
  degree_Kx n K ->
  degree_Mx n M ->
  degree_Kx n (subst_Kx M x K)).
Proof.
apply_mutual_ind kind_mutrec;
default_simp.
Qed.

(* end hide *)

Lemma subst_Kx_degree_Kx :
forall K M x n,
  degree_Kx n K ->
  degree_Mx n M ->
  degree_Kx n (subst_Kx M x K).
Proof.
pose proof subst_Kx_degree_Kx_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_Kx_degree_Kx : lngen.

(* begin hide *)

Lemma subst_Ax_fresh_subst_Mx_fresh_mutual :
(forall A M x,
  x `notin` fv_Ax A ->
  subst_Ax M x A = A) /\
(forall M1 M x,
  x `notin` fv_Mx M1 ->
  subst_Mx M x M1 = M1).
Proof.
apply_mutual_ind family_object_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_Ax_fresh :
forall A M x,
  x `notin` fv_Ax A ->
  subst_Ax M x A = A.
Proof.
pose proof subst_Ax_fresh_subst_Mx_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_Ax_fresh : lngen.
Hint Rewrite subst_Ax_fresh using solve [auto] : lngen.

Lemma subst_Mx_fresh :
forall M1 M x,
  x `notin` fv_Mx M1 ->
  subst_Mx M x M1 = M1.
Proof.
pose proof subst_Ax_fresh_subst_Mx_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_Mx_fresh : lngen.
Hint Rewrite subst_Mx_fresh using solve [auto] : lngen.

(* begin hide *)

Lemma subst_Kx_fresh_mutual :
(forall K M x,
  x `notin` fv_Kx K ->
  subst_Kx M x K = K).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_Kx_fresh :
forall K M x,
  x `notin` fv_Kx K ->
  subst_Kx M x K = K.
Proof.
pose proof subst_Kx_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_Kx_fresh : lngen.
Hint Rewrite subst_Kx_fresh using solve [auto] : lngen.

Lemma subst_Ax_lc_A :
forall A M x,
  lc_A A ->
  lc_M M ->
  lc_A (subst_Ax M x A).
Proof.
default_simp.
Qed.

Hint Resolve subst_Ax_lc_A : lngen.

Lemma subst_Mx_lc_M :
forall M M1 x,
  lc_M M ->
  lc_M M1 ->
  lc_M (subst_Mx M1 x M).
Proof.
default_simp.
Qed.

Hint Resolve subst_Mx_lc_M : lngen.

Lemma subst_Kx_lc_K :
forall K M x,
  lc_K K ->
  lc_M M ->
  lc_K (subst_Kx M x K).
Proof.
default_simp.
Qed.

Hint Resolve subst_Kx_lc_K : lngen.

(* begin hide *)

Lemma subst_Ax_open_Ax_rec_subst_Mx_open_Mx_rec_mutual :
(forall A M M1 x n,
  lc_M M ->
  subst_Ax M x (open_Ax_rec n M1 A) = open_Ax_rec n (subst_Mx M x M1) (subst_Ax M x A)) /\
(forall M2 M M1 x n,
  lc_M M ->
  subst_Mx M x (open_Mx_rec n M1 M2) = open_Mx_rec n (subst_Mx M x M1) (subst_Mx M x M2)).
Proof.
apply_mutual_ind family_object_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_Ax_open_Ax_rec :
forall A M M1 x n,
  lc_M M ->
  subst_Ax M x (open_Ax_rec n M1 A) = open_Ax_rec n (subst_Mx M x M1) (subst_Ax M x A).
Proof.
pose proof subst_Ax_open_Ax_rec_subst_Mx_open_Mx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_Ax_open_Ax_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma subst_Mx_open_Mx_rec :
forall M2 M M1 x n,
  lc_M M ->
  subst_Mx M x (open_Mx_rec n M1 M2) = open_Mx_rec n (subst_Mx M x M1) (subst_Mx M x M2).
Proof.
pose proof subst_Ax_open_Ax_rec_subst_Mx_open_Mx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_Mx_open_Mx_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma subst_Kx_open_Kx_rec_mutual :
(forall K M M1 x n,
  lc_M M ->
  subst_Kx M x (open_Kx_rec n M1 K) = open_Kx_rec n (subst_Mx M x M1) (subst_Kx M x K)).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_Kx_open_Kx_rec :
forall K M M1 x n,
  lc_M M ->
  subst_Kx M x (open_Kx_rec n M1 K) = open_Kx_rec n (subst_Mx M x M1) (subst_Kx M x K).
Proof.
pose proof subst_Kx_open_Kx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_Kx_open_Kx_rec : lngen.

(* end hide *)

Lemma subst_Ax_open_Ax :
forall A M M1 x,
  lc_M M ->
  subst_Ax M x (open_Ax M1 A) = open_Ax (subst_Mx M x M1) (subst_Ax M x A).
Proof.
unfold open_Ax; default_simp.
Qed.

Hint Resolve subst_Ax_open_Ax : lngen.

Lemma subst_Mx_open_Mx :
forall M2 M M1 x,
  lc_M M ->
  subst_Mx M x (open_Mx M1 M2) = open_Mx (subst_Mx M x M1) (subst_Mx M x M2).
Proof.
unfold open_Mx; default_simp.
Qed.

Hint Resolve subst_Mx_open_Mx : lngen.

Lemma subst_Kx_open_Kx :
forall K M M1 x,
  lc_M M ->
  subst_Kx M x (open_Kx M1 K) = open_Kx (subst_Mx M x M1) (subst_Kx M x K).
Proof.
unfold open_Kx; default_simp.
Qed.

Hint Resolve subst_Kx_open_Kx : lngen.

Lemma subst_Ax_open_Ax_var :
forall A M x x1,
  x <> x1 ->
  lc_M M ->
  subst_Ax M x (open_Ax (object_var x1) A) = open_Ax (object_var x1) (subst_Ax M x A).
Proof.
intros; rewrite subst_Ax_open_Ax; default_simp.
Qed.

Hint Resolve subst_Ax_open_Ax_var : lngen.

Lemma subst_Mx_open_Mx_var :
forall M1 M x x1,
  x <> x1 ->
  lc_M M ->
  subst_Mx M x (open_Mx (object_var x1) M1) = open_Mx (object_var x1) (subst_Mx M x M1).
Proof.
intros; rewrite subst_Mx_open_Mx; default_simp.
Qed.

Hint Resolve subst_Mx_open_Mx_var : lngen.

Lemma subst_Kx_open_Kx_var :
forall K M x x1,
  x <> x1 ->
  lc_M M ->
  subst_Kx M x (open_Kx (object_var x1) K) = open_Kx (object_var x1) (subst_Kx M x K).
Proof.
intros; rewrite subst_Kx_open_Kx; default_simp.
Qed.

Hint Resolve subst_Kx_open_Kx_var : lngen.

(* begin hide *)

Lemma subst_Ax_spec_rec_subst_Mx_spec_rec_mutual :
(forall A M x n,
  subst_Ax M x A = open_Ax_rec n M (close_Ax_rec n x A)) /\
(forall M M1 x n,
  subst_Mx M1 x M = open_Mx_rec n M1 (close_Mx_rec n x M)).
Proof.
apply_mutual_ind family_object_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_Ax_spec_rec :
forall A M x n,
  subst_Ax M x A = open_Ax_rec n M (close_Ax_rec n x A).
Proof.
pose proof subst_Ax_spec_rec_subst_Mx_spec_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_Ax_spec_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma subst_Mx_spec_rec :
forall M M1 x n,
  subst_Mx M1 x M = open_Mx_rec n M1 (close_Mx_rec n x M).
Proof.
pose proof subst_Ax_spec_rec_subst_Mx_spec_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_Mx_spec_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma subst_Kx_spec_rec_mutual :
(forall K M x n,
  subst_Kx M x K = open_Kx_rec n M (close_Kx_rec n x K)).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_Kx_spec_rec :
forall K M x n,
  subst_Kx M x K = open_Kx_rec n M (close_Kx_rec n x K).
Proof.
pose proof subst_Kx_spec_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_Kx_spec_rec : lngen.

(* end hide *)

Lemma subst_Ax_spec :
forall A M x,
  subst_Ax M x A = open_Ax M (close_Ax x A).
Proof.
unfold close_Ax; unfold open_Ax; default_simp.
Qed.

Hint Resolve subst_Ax_spec : lngen.

Lemma subst_Mx_spec :
forall M M1 x,
  subst_Mx M1 x M = open_Mx M1 (close_Mx x M).
Proof.
unfold close_Mx; unfold open_Mx; default_simp.
Qed.

Hint Resolve subst_Mx_spec : lngen.

Lemma subst_Kx_spec :
forall K M x,
  subst_Kx M x K = open_Kx M (close_Kx x K).
Proof.
unfold close_Kx; unfold open_Kx; default_simp.
Qed.

Hint Resolve subst_Kx_spec : lngen.

(* begin hide *)

Lemma subst_Ax_subst_Ax_subst_Mx_subst_Mx_mutual :
(forall A M M1 x1 x,
  x1 `notin` fv_Mx M ->
  x1 <> x ->
  subst_Ax M x (subst_Ax M1 x1 A) = subst_Ax (subst_Mx M x M1) x1 (subst_Ax M x A)) /\
(forall M M1 M2 x1 x,
  x1 `notin` fv_Mx M1 ->
  x1 <> x ->
  subst_Mx M1 x (subst_Mx M2 x1 M) = subst_Mx (subst_Mx M1 x M2) x1 (subst_Mx M1 x M)).
Proof.
apply_mutual_ind family_object_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_Ax_subst_Ax :
forall A M M1 x1 x,
  x1 `notin` fv_Mx M ->
  x1 <> x ->
  subst_Ax M x (subst_Ax M1 x1 A) = subst_Ax (subst_Mx M x M1) x1 (subst_Ax M x A).
Proof.
pose proof subst_Ax_subst_Ax_subst_Mx_subst_Mx_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_Ax_subst_Ax : lngen.

Lemma subst_Mx_subst_Mx :
forall M M1 M2 x1 x,
  x1 `notin` fv_Mx M1 ->
  x1 <> x ->
  subst_Mx M1 x (subst_Mx M2 x1 M) = subst_Mx (subst_Mx M1 x M2) x1 (subst_Mx M1 x M).
Proof.
pose proof subst_Ax_subst_Ax_subst_Mx_subst_Mx_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_Mx_subst_Mx : lngen.

(* begin hide *)

Lemma subst_Kx_subst_Kx_mutual :
(forall K M M1 x1 x,
  x1 `notin` fv_Mx M ->
  x1 <> x ->
  subst_Kx M x (subst_Kx M1 x1 K) = subst_Kx (subst_Mx M x M1) x1 (subst_Kx M x K)).
Proof.
apply_mutual_ind kind_mutind;
default_simp.
Qed.

(* end hide *)

Lemma subst_Kx_subst_Kx :
forall K M M1 x1 x,
  x1 `notin` fv_Mx M ->
  x1 <> x ->
  subst_Kx M x (subst_Kx M1 x1 K) = subst_Kx (subst_Mx M x M1) x1 (subst_Kx M x K).
Proof.
pose proof subst_Kx_subst_Kx_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_Kx_subst_Kx : lngen.


(* *********************************************************************** *)
(** * Theorems about [fv] *)

(* begin hide *)

Lemma fv_Ax_close_Ax_rec_fv_Mx_close_Mx_rec_mutual :
(forall A x n,
  fv_Ax (close_Ax_rec n x A) [=] remove x (fv_Ax A)) /\
(forall M x n,
  fv_Mx (close_Mx_rec n x M) [=] remove x (fv_Mx M)).
Proof.
apply_mutual_ind family_object_mutind;
default_simp; fsetdec var.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_Ax_close_Ax_rec :
forall A x n,
  fv_Ax (close_Ax_rec n x A) [=] remove x (fv_Ax A).
Proof.
pose proof fv_Ax_close_Ax_rec_fv_Mx_close_Mx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_Ax_close_Ax_rec : lngen.
Hint Rewrite fv_Ax_close_Ax_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_Mx_close_Mx_rec :
forall M x n,
  fv_Mx (close_Mx_rec n x M) [=] remove x (fv_Mx M).
Proof.
pose proof fv_Ax_close_Ax_rec_fv_Mx_close_Mx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_Mx_close_Mx_rec : lngen.
Hint Rewrite fv_Mx_close_Mx_rec using solve [auto] : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_Kx_close_Kx_rec_mutual :
(forall K x n,
  fv_Kx (close_Kx_rec n x K) [=] remove x (fv_Kx K)).
Proof.
apply_mutual_ind kind_mutind;
default_simp; fsetdec var.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_Kx_close_Kx_rec :
forall K x n,
  fv_Kx (close_Kx_rec n x K) [=] remove x (fv_Kx K).
Proof.
pose proof fv_Kx_close_Kx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_Kx_close_Kx_rec : lngen.
Hint Rewrite fv_Kx_close_Kx_rec using solve [auto] : lngen.

(* end hide *)

Lemma fv_Ax_close_Ax :
forall A x,
  fv_Ax (close_Ax x A) [=] remove x (fv_Ax A).
Proof.
unfold close_Ax; default_simp.
Qed.

Hint Resolve fv_Ax_close_Ax : lngen.
Hint Rewrite fv_Ax_close_Ax using solve [auto] : lngen.

Lemma fv_Mx_close_Mx :
forall M x,
  fv_Mx (close_Mx x M) [=] remove x (fv_Mx M).
Proof.
unfold close_Mx; default_simp.
Qed.

Hint Resolve fv_Mx_close_Mx : lngen.
Hint Rewrite fv_Mx_close_Mx using solve [auto] : lngen.

Lemma fv_Kx_close_Kx :
forall K x,
  fv_Kx (close_Kx x K) [=] remove x (fv_Kx K).
Proof.
unfold close_Kx; default_simp.
Qed.

Hint Resolve fv_Kx_close_Kx : lngen.
Hint Rewrite fv_Kx_close_Kx using solve [auto] : lngen.

(* begin hide *)

Lemma fv_Ax_open_Ax_rec_lower_fv_Mx_open_Mx_rec_lower_mutual :
(forall A M n,
  fv_Ax A [<=] fv_Ax (open_Ax_rec n M A)) /\
(forall M M1 n,
  fv_Mx M [<=] fv_Mx (open_Mx_rec n M1 M)).
Proof.
apply_mutual_ind family_object_mutind;
default_simp; fsetdec var.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_Ax_open_Ax_rec_lower :
forall A M n,
  fv_Ax A [<=] fv_Ax (open_Ax_rec n M A).
Proof.
pose proof fv_Ax_open_Ax_rec_lower_fv_Mx_open_Mx_rec_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_Ax_open_Ax_rec_lower : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_Mx_open_Mx_rec_lower :
forall M M1 n,
  fv_Mx M [<=] fv_Mx (open_Mx_rec n M1 M).
Proof.
pose proof fv_Ax_open_Ax_rec_lower_fv_Mx_open_Mx_rec_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_Mx_open_Mx_rec_lower : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_Kx_open_Kx_rec_lower_mutual :
(forall K M n,
  fv_Kx K [<=] fv_Kx (open_Kx_rec n M K)).
Proof.
apply_mutual_ind kind_mutind;
default_simp; fsetdec var.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_Kx_open_Kx_rec_lower :
forall K M n,
  fv_Kx K [<=] fv_Kx (open_Kx_rec n M K).
Proof.
pose proof fv_Kx_open_Kx_rec_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_Kx_open_Kx_rec_lower : lngen.

(* end hide *)

Lemma fv_Ax_open_Ax_lower :
forall A M,
  fv_Ax A [<=] fv_Ax (open_Ax M A).
Proof.
unfold open_Ax; default_simp.
Qed.

Hint Resolve fv_Ax_open_Ax_lower : lngen.

Lemma fv_Mx_open_Mx_lower :
forall M M1,
  fv_Mx M [<=] fv_Mx (open_Mx M1 M).
Proof.
unfold open_Mx; default_simp.
Qed.

Hint Resolve fv_Mx_open_Mx_lower : lngen.

Lemma fv_Kx_open_Kx_lower :
forall K M,
  fv_Kx K [<=] fv_Kx (open_Kx M K).
Proof.
unfold open_Kx; default_simp.
Qed.

Hint Resolve fv_Kx_open_Kx_lower : lngen.

(* begin hide *)

Lemma fv_Ax_open_Ax_rec_upper_fv_Mx_open_Mx_rec_upper_mutual :
(forall A M n,
  fv_Ax (open_Ax_rec n M A) [<=] fv_Mx M `union` fv_Ax A) /\
(forall M M1 n,
  fv_Mx (open_Mx_rec n M1 M) [<=] fv_Mx M1 `union` fv_Mx M).
Proof.
apply_mutual_ind family_object_mutind;
default_simp; fsetdec var.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_Ax_open_Ax_rec_upper :
forall A M n,
  fv_Ax (open_Ax_rec n M A) [<=] fv_Mx M `union` fv_Ax A.
Proof.
pose proof fv_Ax_open_Ax_rec_upper_fv_Mx_open_Mx_rec_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_Ax_open_Ax_rec_upper : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_Mx_open_Mx_rec_upper :
forall M M1 n,
  fv_Mx (open_Mx_rec n M1 M) [<=] fv_Mx M1 `union` fv_Mx M.
Proof.
pose proof fv_Ax_open_Ax_rec_upper_fv_Mx_open_Mx_rec_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_Mx_open_Mx_rec_upper : lngen.

(* end hide *)

(* begin hide *)

Lemma fv_Kx_open_Kx_rec_upper_mutual :
(forall K M n,
  fv_Kx (open_Kx_rec n M K) [<=] fv_Mx M `union` fv_Kx K).
Proof.
apply_mutual_ind kind_mutind;
default_simp; fsetdec var.
Qed.

(* end hide *)

(* begin hide *)

Lemma fv_Kx_open_Kx_rec_upper :
forall K M n,
  fv_Kx (open_Kx_rec n M K) [<=] fv_Mx M `union` fv_Kx K.
Proof.
pose proof fv_Kx_open_Kx_rec_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_Kx_open_Kx_rec_upper : lngen.

(* end hide *)

Lemma fv_Ax_open_Ax_upper :
forall A M,
  fv_Ax (open_Ax M A) [<=] fv_Mx M `union` fv_Ax A.
Proof.
unfold open_Ax; default_simp.
Qed.

Hint Resolve fv_Ax_open_Ax_upper : lngen.

Lemma fv_Mx_open_Mx_upper :
forall M M1,
  fv_Mx (open_Mx M1 M) [<=] fv_Mx M1 `union` fv_Mx M.
Proof.
unfold open_Mx; default_simp.
Qed.

Hint Resolve fv_Mx_open_Mx_upper : lngen.

Lemma fv_Kx_open_Kx_upper :
forall K M,
  fv_Kx (open_Kx M K) [<=] fv_Mx M `union` fv_Kx K.
Proof.
unfold open_Kx; default_simp.
Qed.

Hint Resolve fv_Kx_open_Kx_upper : lngen.

(* begin hide *)

Lemma fv_Ax_subst_Ax_fresh_fv_Mx_subst_Mx_fresh_mutual :
(forall A M x,
  x `notin` fv_Ax A ->
  fv_Ax (subst_Ax M x A) [=] fv_Ax A) /\
(forall M M1 x,
  x `notin` fv_Mx M ->
  fv_Mx (subst_Mx M1 x M) [=] fv_Mx M).
Proof.
apply_mutual_ind family_object_mutind;
default_simp; fsetdec var.
Qed.

(* end hide *)

Lemma fv_Ax_subst_Ax_fresh :
forall A M x,
  x `notin` fv_Ax A ->
  fv_Ax (subst_Ax M x A) [=] fv_Ax A.
Proof.
pose proof fv_Ax_subst_Ax_fresh_fv_Mx_subst_Mx_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_Ax_subst_Ax_fresh : lngen.
Hint Rewrite fv_Ax_subst_Ax_fresh using solve [auto] : lngen.

Lemma fv_Mx_subst_Mx_fresh :
forall M M1 x,
  x `notin` fv_Mx M ->
  fv_Mx (subst_Mx M1 x M) [=] fv_Mx M.
Proof.
pose proof fv_Ax_subst_Ax_fresh_fv_Mx_subst_Mx_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_Mx_subst_Mx_fresh : lngen.
Hint Rewrite fv_Mx_subst_Mx_fresh using solve [auto] : lngen.

(* begin hide *)

Lemma fv_Kx_subst_Kx_fresh_mutual :
(forall K M x,
  x `notin` fv_Kx K ->
  fv_Kx (subst_Kx M x K) [=] fv_Kx K).
Proof.
apply_mutual_ind kind_mutind;
default_simp; fsetdec var.
Qed.

(* end hide *)

Lemma fv_Kx_subst_Kx_fresh :
forall K M x,
  x `notin` fv_Kx K ->
  fv_Kx (subst_Kx M x K) [=] fv_Kx K.
Proof.
pose proof fv_Kx_subst_Kx_fresh_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_Kx_subst_Kx_fresh : lngen.
Hint Rewrite fv_Kx_subst_Kx_fresh using solve [auto] : lngen.

(* begin hide *)

Lemma fv_Ax_subst_Ax_lower_fv_Mx_subst_Mx_lower_mutual :
(forall A M x,
  remove x (fv_Ax A) [<=] fv_Ax (subst_Ax M x A)) /\
(forall M M1 x,
  remove x (fv_Mx M) [<=] fv_Mx (subst_Mx M1 x M)).
Proof.
apply_mutual_ind family_object_mutind;
default_simp; fsetdec var.
Qed.

(* end hide *)

Lemma fv_Ax_subst_Ax_lower :
forall A M x,
  remove x (fv_Ax A) [<=] fv_Ax (subst_Ax M x A).
Proof.
pose proof fv_Ax_subst_Ax_lower_fv_Mx_subst_Mx_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_Ax_subst_Ax_lower : lngen.

Lemma fv_Mx_subst_Mx_lower :
forall M M1 x,
  remove x (fv_Mx M) [<=] fv_Mx (subst_Mx M1 x M).
Proof.
pose proof fv_Ax_subst_Ax_lower_fv_Mx_subst_Mx_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_Mx_subst_Mx_lower : lngen.

(* begin hide *)

Lemma fv_Kx_subst_Kx_lower_mutual :
(forall K M x,
  remove x (fv_Kx K) [<=] fv_Kx (subst_Kx M x K)).
Proof.
apply_mutual_ind kind_mutind;
default_simp; fsetdec var.
Qed.

(* end hide *)

Lemma fv_Kx_subst_Kx_lower :
forall K M x,
  remove x (fv_Kx K) [<=] fv_Kx (subst_Kx M x K).
Proof.
pose proof fv_Kx_subst_Kx_lower_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_Kx_subst_Kx_lower : lngen.

(* begin hide *)

Lemma fv_Ax_subst_Ax_notin_fv_Mx_subst_Mx_notin_mutual :
(forall A M x x1,
  x1 `notin` fv_Ax A ->
  x1 `notin` fv_Mx M ->
  x1 `notin` fv_Ax (subst_Ax M x A)) /\
(forall M M1 x x1,
  x1 `notin` fv_Mx M ->
  x1 `notin` fv_Mx M1 ->
  x1 `notin` fv_Mx (subst_Mx M1 x M)).
Proof.
apply_mutual_ind family_object_mutind;
default_simp; fsetdec var.
Qed.

(* end hide *)

Lemma fv_Ax_subst_Ax_notin :
forall A M x x1,
  x1 `notin` fv_Ax A ->
  x1 `notin` fv_Mx M ->
  x1 `notin` fv_Ax (subst_Ax M x A).
Proof.
pose proof fv_Ax_subst_Ax_notin_fv_Mx_subst_Mx_notin_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_Ax_subst_Ax_notin : lngen.

Lemma fv_Mx_subst_Mx_notin :
forall M M1 x x1,
  x1 `notin` fv_Mx M ->
  x1 `notin` fv_Mx M1 ->
  x1 `notin` fv_Mx (subst_Mx M1 x M).
Proof.
pose proof fv_Ax_subst_Ax_notin_fv_Mx_subst_Mx_notin_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_Mx_subst_Mx_notin : lngen.

(* begin hide *)

Lemma fv_Kx_subst_Kx_notin_mutual :
(forall K M x x1,
  x1 `notin` fv_Kx K ->
  x1 `notin` fv_Mx M ->
  x1 `notin` fv_Kx (subst_Kx M x K)).
Proof.
apply_mutual_ind kind_mutind;
default_simp; fsetdec var.
Qed.

(* end hide *)

Lemma fv_Kx_subst_Kx_notin :
forall K M x x1,
  x1 `notin` fv_Kx K ->
  x1 `notin` fv_Mx M ->
  x1 `notin` fv_Kx (subst_Kx M x K).
Proof.
pose proof fv_Kx_subst_Kx_notin_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_Kx_subst_Kx_notin : lngen.

(* begin hide *)

Lemma fv_Ax_subst_Ax_upper_fv_Mx_subst_Mx_upper_mutual :
(forall A M x,
  fv_Ax (subst_Ax M x A) [<=] fv_Mx M `union` remove x (fv_Ax A)) /\
(forall M M1 x,
  fv_Mx (subst_Mx M1 x M) [<=] fv_Mx M1 `union` remove x (fv_Mx M)).
Proof.
apply_mutual_ind family_object_mutind;
default_simp; fsetdec var.
Qed.

(* end hide *)

Lemma fv_Ax_subst_Ax_upper :
forall A M x,
  fv_Ax (subst_Ax M x A) [<=] fv_Mx M `union` remove x (fv_Ax A).
Proof.
pose proof fv_Ax_subst_Ax_upper_fv_Mx_subst_Mx_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_Ax_subst_Ax_upper : lngen.

Lemma fv_Mx_subst_Mx_upper :
forall M M1 x,
  fv_Mx (subst_Mx M1 x M) [<=] fv_Mx M1 `union` remove x (fv_Mx M).
Proof.
pose proof fv_Ax_subst_Ax_upper_fv_Mx_subst_Mx_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_Mx_subst_Mx_upper : lngen.

(* begin hide *)

Lemma fv_Kx_subst_Kx_upper_mutual :
(forall K M x,
  fv_Kx (subst_Kx M x K) [<=] fv_Mx M `union` remove x (fv_Kx K)).
Proof.
apply_mutual_ind kind_mutind;
default_simp; fsetdec var.
Qed.

(* end hide *)

Lemma fv_Kx_subst_Kx_upper :
forall K M x,
  fv_Kx (subst_Kx M x K) [<=] fv_Mx M `union` remove x (fv_Kx K).
Proof.
pose proof fv_Kx_subst_Kx_upper_mutual as H; intuition eauto.
Qed.

Hint Resolve fv_Kx_subst_Kx_upper : lngen.


(* *********************************************************************** *)
(** * More theorems about [subst] *)

(* begin hide *)

Lemma subst_Ax_close_Ax_rec_open_Ax_rec_subst_Mx_close_Mx_rec_open_Mx_rec_mutual :
(forall A M x x1 n,
  x1 `notin` fv_Ax A ->
  x1 `notin` fv_Mx M ->
  x1 <> x ->
  degree_Mx n M ->
  subst_Ax M x A = close_Ax_rec n x1 (subst_Ax M x (open_Ax_rec n (object_var x1) A))) *
(forall M1 M x x1 n,
  x1 `notin` fv_Mx M1 ->
  x1 `notin` fv_Mx M ->
  x1 <> x ->
  degree_Mx n M ->
  subst_Mx M x M1 = close_Mx_rec n x1 (subst_Mx M x (open_Mx_rec n (object_var x1) M1))).
Proof.
apply_mutual_ind family_object_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_Ax_close_Ax_rec_open_Ax_rec :
forall A M x x1 n,
  x1 `notin` fv_Ax A ->
  x1 `notin` fv_Mx M ->
  x1 <> x ->
  degree_Mx n M ->
  subst_Ax M x A = close_Ax_rec n x1 (subst_Ax M x (open_Ax_rec n (object_var x1) A)).
Proof.
pose proof subst_Ax_close_Ax_rec_open_Ax_rec_subst_Mx_close_Mx_rec_open_Mx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_Ax_close_Ax_rec_open_Ax_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma subst_Mx_close_Mx_rec_open_Mx_rec :
forall M1 M x x1 n,
  x1 `notin` fv_Mx M1 ->
  x1 `notin` fv_Mx M ->
  x1 <> x ->
  degree_Mx n M ->
  subst_Mx M x M1 = close_Mx_rec n x1 (subst_Mx M x (open_Mx_rec n (object_var x1) M1)).
Proof.
pose proof subst_Ax_close_Ax_rec_open_Ax_rec_subst_Mx_close_Mx_rec_open_Mx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_Mx_close_Mx_rec_open_Mx_rec : lngen.

(* end hide *)

(* begin hide *)

Lemma subst_Kx_close_Kx_rec_open_Kx_rec_mutual :
(forall K M x x1 n,
  x1 `notin` fv_Kx K ->
  x1 `notin` fv_Mx M ->
  x1 <> x ->
  degree_Mx n M ->
  subst_Kx M x K = close_Kx_rec n x1 (subst_Kx M x (open_Kx_rec n (object_var x1) K))).
Proof.
apply_mutual_ind kind_mutrec;
default_simp.
Qed.

(* end hide *)

(* begin hide *)

Lemma subst_Kx_close_Kx_rec_open_Kx_rec :
forall K M x x1 n,
  x1 `notin` fv_Kx K ->
  x1 `notin` fv_Mx M ->
  x1 <> x ->
  degree_Mx n M ->
  subst_Kx M x K = close_Kx_rec n x1 (subst_Kx M x (open_Kx_rec n (object_var x1) K)).
Proof.
pose proof subst_Kx_close_Kx_rec_open_Kx_rec_mutual as H; intuition eauto.
Qed.

Hint Resolve subst_Kx_close_Kx_rec_open_Kx_rec : lngen.

(* end hide *)

Lemma subst_Ax_close_Ax_open_Ax :
forall A M x x1,
  x1 `notin` fv_Ax A ->
  x1 `notin` fv_Mx M ->
  x1 <> x ->
  lc_M M ->
  subst_Ax M x A = close_Ax x1 (subst_Ax M x (open_Ax (object_var x1) A)).
Proof.
unfold close_Ax; unfold open_Ax; default_simp.
Qed.

Hint Resolve subst_Ax_close_Ax_open_Ax : lngen.

Lemma subst_Mx_close_Mx_open_Mx :
forall M1 M x x1,
  x1 `notin` fv_Mx M1 ->
  x1 `notin` fv_Mx M ->
  x1 <> x ->
  lc_M M ->
  subst_Mx M x M1 = close_Mx x1 (subst_Mx M x (open_Mx (object_var x1) M1)).
Proof.
unfold close_Mx; unfold open_Mx; default_simp.
Qed.

Hint Resolve subst_Mx_close_Mx_open_Mx : lngen.

Lemma subst_Kx_close_Kx_open_Kx :
forall K M x x1,
  x1 `notin` fv_Kx K ->
  x1 `notin` fv_Mx M ->
  x1 <> x ->
  lc_M M ->
  subst_Kx M x K = close_Kx x1 (subst_Kx M x (open_Kx (object_var x1) K)).
Proof.
unfold close_Kx; unfold open_Kx; default_simp.
Qed.

Hint Resolve subst_Kx_close_Kx_open_Kx : lngen.

Lemma subst_Ax_family_pi :
forall x1 A B M x,
  lc_M M ->
  x1 `notin` fv_Mx M `union` fv_Ax B `union` singleton x ->
  subst_Ax M x (family_pi A B) = family_pi (subst_Ax M x A) (close_Ax x1 (subst_Ax M x (open_Ax (object_var x1) B))).
Proof.
default_simp.
Qed.

Hint Resolve subst_Ax_family_pi : lngen.

Lemma subst_Ax_family_abs :
forall x1 A B M x,
  lc_M M ->
  x1 `notin` fv_Mx M `union` fv_Ax B `union` singleton x ->
  subst_Ax M x (family_abs A B) = family_abs (subst_Ax M x A) (close_Ax x1 (subst_Ax M x (open_Ax (object_var x1) B))).
Proof.
default_simp.
Qed.

Hint Resolve subst_Ax_family_abs : lngen.

Lemma subst_Mx_object_abs :
forall x1 A M1 M x,
  lc_M M ->
  x1 `notin` fv_Mx M `union` fv_Mx M1 `union` singleton x ->
  subst_Mx M x (object_abs A M1) = object_abs (subst_Ax M x A) (close_Mx x1 (subst_Mx M x (open_Mx (object_var x1) M1))).
Proof.
default_simp.
Qed.

Hint Resolve subst_Mx_object_abs : lngen.

Lemma subst_Kx_kind_pi :
forall x1 A K M x,
  lc_M M ->
  x1 `notin` fv_Mx M `union` fv_Kx K `union` singleton x ->
  subst_Kx M x (kind_pi A K) = kind_pi (subst_Ax M x A) (close_Kx x1 (subst_Kx M x (open_Kx (object_var x1) K))).
Proof.
default_simp.
Qed.

Hint Resolve subst_Kx_kind_pi : lngen.

