(* This file is distributed under the terms of the MIT License, also
   known as the X11 Licence.  A copy of this license is in the README
   file that accompanied the original distribution of this file.

   Based on code written by:
     Brian Aydemir
     Aaron Bohannon
     Arthur Charg\'eraud

   $Id: AssocList.v 2332 2009-01-12 17:34:43Z baydemir $ *)

(** A library for association lists, i.e., lists of pairs *)


Require Import Coq.Lists.List.
Require Import Coq.Logic.Decidable.
Require Import Coq.Classes.Equivalence.
Require Import Coq.Classes.EquivDec.

Require Import AdditionalTactics.
Require Import AssocList.
Require Import FSetsClass.
Require Import FSetsClassDecide.
Require Import FSetsClassFacts.
Require Import ListFacts.


(* ********************************************************************* *)
(** * List properties *)

(** The following properties are an assortment of structural
    properties about association lists. *)

Section AssortedListProperties.
  Variable  X : Type.
  Variables x : X.
  Variables xs ys zs : list X.

  Lemma one_eq_app :
    one x ++ xs = ys ++ zs ->
    (exists qs, ys = x :: qs /\ xs = qs ++ zs) \/
    (ys = nil /\ zs = x :: xs).
  Proof. auto using ListFacts.cons_eq_app. Qed.

  Lemma app_eq_one :
    ys ++ zs = one x ++ xs ->
    (exists qs, ys = x :: qs /\ xs = qs ++ zs) \/
    (ys = nil /\ zs = x :: xs).
  Proof. auto using ListFacts.app_eq_cons. Qed.

  Lemma nil_neq_one_mid :
    nil <> xs ++ one x ++ ys.
  Proof.
    Strategy transparent [ one ].
    simpl. apply List.app_cons_not_nil.
    Strategy opaque [ one ].
  Qed.

  Lemma one_mid_neq_nil :
    xs ++ one x ++ ys <> nil.
  Proof. intros H. symmetry in H. auto using nil_neq_one_mid. Qed.

End AssortedListProperties.


(* ********************************************************************* *)
(** * The [solve_uniq] tactic *)

(** Rewriting hints. *)

Hint Rewrite disjoint_one_l disjoint_one_r : rewr_uniq.
Hint Rewrite disjoint_app_l disjoint_app_r : rewr_uniq.
Hint Rewrite disjoint_map_l disjoint_map_r : rewr_uniq.
Hint Rewrite uniq_app uniq_map             : rewr_uniq.

(** This tactic attempts to solve goals about [uniq].  It takes as an
    argument the type of keys.  It is intended only to work when the
    hypotheses of the goal already contain all the relevant [uniq]
    propositions.

    Implementation note: The second [simpl_alist] in the definition
    is because of [disjoint_one_{l,r}].  The "[|| fail]" at the end
    is so that in the case of failure, the error message reported to
    the user is not the one from [fsetdec]. *)

Ltac solve_uniq key :=
  try trivial;
  simpl_alist in *;
  autorewrite with rewr_uniq in *;
  simpl_alist in *;
  intuition;
  auto with alist
  || (unfold disjoint in *; fsetdec key)
  || fail "did not solve the goal".


(* ********************************************************************* *)
(** * Facts about [uniq] *)

Section UniqDerived.
  Variable  A key : Type.
  Variables x y   : key.
  Variables a b   : A.
  Variables E F G : list (key*A).
  Variable  R     : relation key.
  Variable  Equiv : @Equivalence key R.
  Variable  Eqdec : @EqDec key R Equiv.
  Variable  Fset  : @FSet key R Equiv Eqdec.

  Lemma uniq_cons :
    uniq E ->
    x `notin` dom E ->
    uniq ((x, a) :: E).
  Proof. clear. solve_uniq key. Qed.

  Lemma uniq_insert_mid :
    uniq (G ++ E) ->
    x `notin` dom (G ++ E) ->
    uniq (G ++ (x ~ a) ++ E).
  Proof. clear. solve_uniq key. Qed.

  Lemma uniq_remove_mid :
    uniq (E ++ F ++ G) ->
    uniq (E ++ G).
  Proof. clear. solve_uniq key. Qed.

  Lemma uniq_map_app_l : forall (f : A -> A),
    uniq (F ++ E) ->
    uniq (map f F ++ E).
  Proof. clear. intros. solve_uniq key. Qed.

  Lemma fresh_mid_tail :
    uniq (F ++ (x ~ a) ++ E) ->
    x `notin` dom E.
  Proof. clear. solve_uniq key. Qed.

  Lemma fresh_mid_head :
    uniq (F ++ (x ~ a) ++ E) ->
    x `notin` dom E.
  Proof. clear. solve_uniq key. Qed.

End UniqDerived.


(* ********************************************************************* *)
(** * Additional lemmas and tactics for working with [binds] *)

(** Rewriting hints. *)

Hint Rewrite binds_nil binds_one binds_app : rewr_binds.
Hint Rewrite binds_nil binds_one           : rewr_binds_uniq.

Hint Rewrite
  binds_app_uniq using intuition auto with alist
  : rewr_binds_uniq.

(** The [solve_binds] tactic attempts to solve goals about [binds].
    It takes as an argument the type of keys.  Given its definition,
    it's likely to work only when the hypotheses in the goal already
    contain all the relevant [binds] propositions.  Thus, the tactic
    may not be generally useful.  It is useful, however, for proving
    facts about [binds] such as the ones below. *)

Ltac solve_binds key :=
  simpl_alist in *;
  autorewrite with rewr_binds in *;
  intuition eauto with alist;
  fail "did not solve the goal".

(** The tactics [analyze_binds] and [analyze_binds_uniq] tactics
    take as an argument a hypothesis about [binds] and perform a
    case analysis based on the structure of the association list.
    In the case of [analyze_binds_uniq], the analysis is performed
    assuming that the association list is [uniq].  The lemmas
    [binds_app] and [binds_app_uniq] are examples of such case
    analysis.

    Implementation notes:

      - The initial calls to [simpl_alist] put the relevant
        association lists into normal form.
      - In the case of [binds_app_uniq], we assert that the
        association list is [uniq].
      - Rather than work on [H] itself, we copy it and work on the
        copy.  Thus, in instances where the analysis leaves unsolved
        subgoals, it is still possible to see the original
        hypothesis.
      - The rest of the tactic breaks apart the copy of H and tries
        several simple means for solving the resulting subgoals. *)

Ltac analyze_binds_final H :=
  simpl_alist in H;
  try (progress decompose [and or] H; clear H);
  try solve [ trivial | discriminate | intuition auto with alist ].

Ltac analyze_binds H :=
  simpl_alist;
  simpl_alist in H;
  match type of H with
    | @binds ?key _ ?x ?a ?E =>
      let H' := fresh in
      pose proof H as H';
      autorewrite with rewr_binds in H';
      analyze_binds_final H'
  end.

Ltac analyze_binds_uniq H :=
  simpl_alist;
  simpl_alist in H;
  match type of H with
    | @binds ?key _ ?x ?a ?E =>
      let H' := fresh in
      let U  := fresh in
      match goal with
        | H : uniq ?E |- _ => pose proof H as U
        | _ => assert (U : uniq E); [ try solve_uniq key | ]
      end;
      pose proof H as H';
      autorewrite with rewr_uniq in U;
      autorewrite with rewr_binds_uniq in H';
      analyze_binds_final H';
      clear U
  end.


(* ********************************************************************* *)
(** * Facts about [binds] *)

(** Implementation note: Observe that here, we finally need to require
    that equality on keys is [eq]. *)

Section BindsDerived.
  Variables A B key : Type.
  Variables f       : A -> B.
  Variables x y     : key.
  Variables a b     : A.
  Variables E F G   : list (key*A).
  Variable  Eqdec   : EqDec key eq.
  Variable  Fset    : @FSet key _ _ Eqdec.

  Lemma binds_dec :
    (forall a b : A, {a = b} + {a <> b}) ->
    {binds x a E} + {~ binds x a E}.
  Proof.
    clear - Eqdec. intros. unfold binds. apply List.In_dec.
    decide equality.
    apply equiv_dec.
  Defined.

  Lemma binds_lookup_dec_specif :
    {a : A | binds x a E} + (forall a, ~ binds x a E).
  Proof.
    clear - Eqdec. induction E as [ | [x1 a1] E' [[a' H] | H] ].
    right. intros x' H. inversion H.
    left. exists a'. solve_binds key.
    destruct (x == x1).
      left. exists a1. solve_binds key.
      right. unfold binds. simpl. intros a' [K | K].
        injection K. intuition.
        apply (H a'). assumption.
  Defined.

  Lemma binds_lookup_dec :
    decidable (exists a, binds x a E).
  Proof.
    clear - Eqdec. unfold decidable.
    destruct binds_lookup_dec_specif as [[? ?] | ?].
      left. eauto.
      right. intros [? ?]. intuition eauto.
  Defined.

  Lemma binds_weaken :
    binds x a (E ++ G) ->
    binds x a (E ++ F ++ G).
  Proof. clear. intros. solve_binds key. Qed.

  Lemma binds_mid_eq :
    binds x a (F ++ (x ~ b) ++ E) ->
    uniq (F ++ (x ~ b) ++ E) ->
    a = b.
  Proof. clear. intros H ?. analyze_binds_uniq H; fsetdec key. Qed.

  Lemma binds_remove_mid :
    binds x a (F ++ (y ~ b) ++ G) ->
    x <> y ->
    binds x a (F ++ G).
  Proof. clear. intros. solve_binds key. Qed.

  Lemma binds_In : forall x a (E : list (key*A)),
    binds x a E ->
    x `in` dom E.
  Proof.
    clear. intros x' a' E'.
    induction E' as [ | [? ?] ]; intros H; analyze_binds H; auto with set.
  Qed.

  Lemma binds_unique :
    binds x a E ->
    binds x b E ->
    uniq E ->
    a = b.
  Proof.
    clear. induction E as [ | [x1 a1] E' ].
    inversion 1.
    unfold binds. simpl. intros [H | H] [J | J] K; simpl_alist in *.
      Case "left / left".
        congruence.
      Case "left / right".
        assert (x `in` dom E'). eapply binds_In. unfold binds. apply J.
        injection H; intros -> ->.
        solve_uniq key.
      Case "right / left".
        assert (x `in` dom E'). eapply binds_In. unfold binds. apply H.
        injection J; intros -> ->.
        solve_uniq key.
      Case "right / right".
        unfold binds in *.
        apply IHE'; eauto.
        solve_uniq key.
  Qed.

  Lemma fresh_app_l :
    uniq (F ++ E) ->
    binds x a E ->
    x `notin` dom F.
  Proof.
    clear. intros.
    assert (x `in` dom E) by eauto using binds_In.
    solve_uniq key.
  Qed.

  Lemma fresh_app_r :
    uniq (F ++ E) ->
    binds x a F ->
    x `notin` dom E.
  Proof.
    clear. intros.
    assert (x `in` dom F) by eauto using binds_In.
    solve_uniq key.
  Qed.

End BindsDerived.


(* ********************************************************************* *)
(** * Hints *)

Hint Resolve nil_neq_one_mid one_mid_neq_nil : alist.

Hint Resolve uniq_insert_mid                 : alist.
Hint Immediate uniq_remove_mid               : alist.
Hint Resolve uniq_map_app_l                  : alist.
Hint Immediate fresh_mid_tail fresh_mid_head : alist.

Hint Resolve binds_weaken                : alist.
Hint Immediate binds_remove_mid binds_In : alist.
