(* This file is distributed under the terms of the MIT License, also
   known as the X11 Licence.  A copy of this license is in the README
   file that accompanied the original distribution of this file.

   Based on code written by:
     Brian Aydemir
     Aaron Bohannon
     Arthur Charg\'eraud

   $Id: Atom.v 2731 2009-02-24 17:29:24Z baydemir $ *)

(** A library that defines atoms, objects with decidable equality and
    such that their type contains infinitely many objects. *)

Require Import Coq.Arith.Arith.
Require Import Coq.Arith.Max.
Require Export Coq.Strings.String.
Require Import Coq.Lists.List.
Require Import Coq.Lists.SetoidList.
Require Import Coq.Classes.Equivalence.
Require Import Coq.Classes.EquivDec.
Require Import Omega.

Require Import AdditionalTactics.
Require Import DefaultSolver.
Require Import EqDecInstances.
Require Import FSetsClass.
Require Import FSetsClassImpl. (* BEA: needed for the example below. *)
Require Import ListFacts.


(* *********************************************************************** *)
(** * Class definition *)

(** Atoms are objects with decidable equality and such that their type
    contains infinitely many objects.  We require that [eq] be the
    equality on atoms; the generality of an arbitrary equivalence
    relation does not seem necessary. *)

Class Atom (A : Type) `(EqDec A eq) :=
  atom_fresh_for_list : forall (l : list A), { a : A | ~ List.In a l }.


(* *********************************************************************** *)
(** * Tactic support *)

(* The following tactic may be helpful for declaring [Atom] instances
   for types declared in a fashion similar to a Haskell [newtype]
   declaration.  See the [string'] instance below for an example of
   its use. *)

Ltac Atom_tac constr proj :=
  unfold Atom;
  let l   := fresh in
  let a'  := fresh in
  let pf' := fresh in
  let H   := fresh in
  intros l;
  destruct (atom_fresh_for_list (map proj l)) as [a' pf'];
  exists (constr a');
  intros H;
  assert (List.In (proj (constr a')) (map proj l))
    by auto using ListFacts.In_map;
  simpl in *;
  intuition.


(* *********************************************************************** *)
(** * Class implementations: Base types *)

(** Natural numbers are atoms. *)

Instance Atom_nat : Atom nat _.
Proof with default_simp.
  unfold Atom.

  assert (max_lt : forall x y z, x <= z -> x <= max y z).
    induction x...
    destruct y...
    destruct z...

  assert (get_max : forall xs,
            { n : nat | forall m, InA equiv m xs -> m <= n }).
    induction xs as [ | x xs [y H] ].
    exists 0...
    exists (max x y); intros m J. inversion J...

  intros l.
    destruct (get_max l) as [x H].
    exists (S x). intros J. lapply (H (S x))...
    apply SetoidList.In_InA; auto.
Defined.

(** Strings are atoms. *)

Instance Atom_string : Atom string _.
Proof with default_simp.
  unfold Atom.

  assert (max_1 : forall n m, max n m = n -> m <= n).
    induction n...
    destruct m...

  assert (max_2 : forall n m, max n m = m -> n <= m).
    induction n...
    destruct m...

  assert (get_max : forall xs,
            { s : string | forall t, InA equiv t xs ->
                             String.length t <= String.length s }).
    induction xs as [ | x xs [y H] ].
    exists ""...
    destruct (max_dec (String.length x) (String.length y)) as [J | J].
      exists x; intros t K. inversion K... eauto using le_trans.
      exists y; intros t K. inversion K...

  intros l.
    destruct (get_max l) as [l' H].
    exists (String "a" l'). intros J. lapply (H (String "a" l'))...
    apply SetoidList.In_InA; auto.
Defined.


(* *********************************************************************** *)
(** * Class implementations: Haskell-style "newtypes" *)

(** Using [Atom_tac], it is straightforward to define new atom types
    from existing ones.  For this example, we use [string] as the
    existing atom type.  First, we wrap it similar to a Haskell
    [newtype] declaration, calling the new type [string']. *)

Inductive string' :=
  | string'_intro : string -> string'.

(** Second, we provide an [EqDec] instance for [string']. *)

Instance new_string_eq : EqDec string' eq.
Proof. EqDec_eq_tac. Defined.

(** Third, we define the [Atom] instance for [string'].  We supply to
    [Atom_tac] the name of the single constructor for our new atom
    type and a projection function to the original atom type. *)

Instance Atom_string' : Atom string' _.
(* begin show *)
Proof.
  set (f := fun x => match x with string'_intro s => s end).
  Atom_tac string'_intro f.
Defined.
(* end show *)


(* *********************************************************************** *)
(** * Properties *)

(** We can always generate an atom fresh from a finite set of atoms. *)

Fact atom_fresh_for_set :
  forall `{eqa : EqDec A eq, ! Atom _ eqa, ! FSet _ eqa} (s : fset A),
  { a : A | ~ In a s }.
Proof.
  intros until s.
  destruct (atom_fresh_for_list (elements s)) as [a H].
  exists a.
  intros J.
  contradiction H.
  auto using ListFacts.InA_In, elements_1.
Defined.


(* *********************************************************************** *)
(** * Tactics for picking fresh atoms *)

(** The tactic [(pick fresh Y for L)] takes a finite set (or list) of
    atoms [L] and a fresh name [Y], and adds to the context an atom
    with name [Y] and a proof that [(~ In Y L)], i.e., that [Y] is
    fresh for [L].  The tactic will fail if [Y] is already declared in
    the context. *)

Tactic Notation "pick" "fresh" ident(Y) "for" constr(L) :=
  let Fr := fresh "Fr" in
  (destruct (atom_fresh_for_set  L) as [Y Fr]) ||
  (destruct (atom_fresh_for_list L) as [Y Fr]).


(* ********************************************************************** *)
(** ** Example *)

(** The [example_pick_fresh] tactic below illustrates the general
    pattern for collecting together atoms in a context and picking an
    atom fresh for them.  To make the example concrete, we use
    [string] as the type of atoms. *)

(** We begin by defining a generic tactic that takes a list with type
    [(list (fset A))], for some [A], and returns the union of the
    elmenets of the list. *)

Ltac simplify_list_of_atom_sets L :=
  match type of L with
    | list (fset ?A) =>
      ltac_fold_and_simpl (@union A _ _ _ _) (@empty A _ _ _ _) L
  end.

(** We next define a tactic that gathers the atoms (e.g., strings) in
    the current context and returns them as a finite set.  We
    repeatedly invoke the tactic [ltac_map], using each time a unary
    function that returns a [fset string].  The argument type of the
    function determines the hypotheses that are processed.  Each
    invocation returns a list of finite sets of atoms.  We use
    [simplify_list_of_atom_sets], defined above, to obtain a single
    finite set. *)

Ltac example_gather_atoms :=
  let A := ltac_map (fun x : fset string => x) in
  let B := ltac_map (fun x : string => singleton x) in
  simplify_list_of_atom_sets (A ++ B).

(** Given [example_gather_atoms], we can implement a specialized
    version of "[pick fresh]". *)

Ltac example_pick_fresh Y :=
  let L := example_gather_atoms in
  pick fresh Y for L.

(** Finally, we gave a trivial example of using [example_pick_fresh]. *)

Lemma example_pick_fresh_use :
  forall (x y z : string) (L1 L2 L3: fset string), True.
(* begin show *)
Proof.
  intros.
  example_pick_fresh k.

  (** At this point in the proof, we have a new atom [k] and a
      hypothesis [~ In k (union L1 (union L2 (union L3 (union
      (singleton x) (union (singleton y) (union (singleton z)
      empty)))))]. *)

  trivial.
Qed.
(* end show *)


(* *********************************************************************** *)
(** * Swapping *)

(** The definition is straight forward. *)

Definition swap_atom `{Atom A} (ab : A * A) (x : A) : A :=
 let (a, b) := ab in
 if a == x then b else (if b == x then a else x).

(** The following lemmas enable us to write tactics for simplifying
    applications of [swap_atom]. *)

Lemma swap_atom_left : forall `{Atom A} (a b c : A),
  swap_atom (a, b) a = b.
Proof. default_simp. Qed.

Lemma swap_atom_left' : forall `{Atom A} (a b c : A),
  a = c ->
  swap_atom (a, b) c = b.
Proof. default_simp. Qed.

Lemma swap_atom_right : forall `{Atom A} (a b c : A),
  swap_atom (a, b) b = a.
Proof. default_simp. Qed.

Lemma swap_atom_right' : forall `{Atom A} (a b c : A),
  b = c ->
  swap_atom (a, b) c = a.
Proof. default_simp. Qed.

Lemma swap_atom_neither : forall `{Atom A} (a b c : A),
  a <> c ->
  b <> c ->
  swap_atom (a, b) c = c.
Proof. default_simp. Qed.

(** [swap_atom_simpl_once] and [swap_atom_simpl_hyp] are auxilary tactics. *)

Ltac swap_atom_simpl_once a b c :=
  (rewrite (swap_atom_left' a b c);
    [idtac | solve [congruence]]) ||
  (rewrite (swap_atom_right' a b c);
    [idtac | solve [congruence]]) ||
  (rewrite (swap_atom_neither a b c);
    [idtac | solve [congruence] | solve [congruence]]).

Ltac swap_atom_simpl_hyp H a b c :=
  (rewrite (swap_atom_left' a b c) in H;
    [idtac | solve [congruence]]) ||
  (rewrite (swap_atom_right' a b c) in H;
    [idtac | solve [congruence]]) ||
  (rewrite (swap_atom_neither a b c) in H;
    [idtac | solve [congruence] | solve [congruence]]).

(** [swap_atom_simpl] performs all reductions of atom swaps (in the
    goal or assumptions) for which the result is inferrable from the
    context using the [congruence] tactic. *)

Ltac swap_atom_simpl :=
  repeat (
    match goal with
    | |- context [swap_atom (?a, ?b) ?c] =>
        swap_atom_simpl_once a b c
    | H : context [swap_atom (?a, ?b) ?c] |- _ =>
        swap_atom_simpl_hyp H a b c
    end).

(** [swap_atom_case H1 H2 a b c] will reduce all expressions of the
    form [swap_atom (a, b) c] in the goal to either [a], [b], or [c],
    splitting the goal as necessary, when the relationships between
    [a] and [c] and between [b] and [c] cannot be inferred from the
    context.  [H1] will be used as the name for a newly introduced
    assumption about the relationship between [a] and [c], and
    respectively [H2] for [b] and [c].  Also, any additional
    reductions of atom swaps that are inferrable from the context will
    be performed. *)

Ltac swap_atom_case H1 H2 a b c :=
  swap_atom_simpl;
  try match goal with
  | |- context [swap_atom (a, b) c] =>
      case (a == c); intro H1; try congruence;
      swap_atom_simpl;
      try match goal with
      | |- context [swap_atom (a, b) c] =>
          case (b == c); intro H2; try congruence;
          swap_atom_simpl
      end
  end.

(** The following three lemmas are essential facts about swapping.
    From them, we can derive many other facts. *)

Lemma swap_atom_same : forall `{Atom A} (a b : A),
  swap_atom (a, a) b = b.
Proof. intros. swap_atom_case K1 K2 a a b; trivial. Qed.

Lemma swap_atom_invol : forall `{Atom A} (ab : A * A) (c : A),
  swap_atom ab (swap_atom ab c) = c.
Proof. intros ? ? ? ? [a b] c. swap_atom_case K1 K2 a b c; trivial. Qed.

Lemma swap_atom_distrib : forall `{Atom A} (ab : A * A) (c d e : A),
  swap_atom ab (swap_atom (c, d) e) =
  swap_atom (swap_atom ab c, swap_atom ab d) (swap_atom ab e).
Proof.
  intros ? ? ? ? [a b] c d e.
  swap_atom_case ac bc a b c;
  swap_atom_case ad bd a b d;
  swap_atom_case ae be a b e;
  swap_atom_case ce de c d e;
  trivial.
Qed.


(* *********************************************************************** *)
(** * Hints *)

Hint Resolve @swap_atom_same @swap_atom_invol : swap_distrib swap.
Hint Resolve @swap_atom_distrib               : swap_distrib.
