(* $Id: FSetsClass.v 2834 2009-03-04 03:06:06Z baydemir $

   Includes code written by:
     Brian Aydemir
     Aaron Bohannon *)

(** A type class based implementation of finite sets.  The design and
    implementation is inspired by (and uses code from) Coq's FSets
    library. *)


(***********************************************************************)
(*  v      *   The Coq Proof Assistant  /  The Coq Development Team    *)
(* <O___,, *        INRIA-Rocquencourt  &  LRI-CNRS-Orsay              *)
(*   \VV/  *************************************************************)
(*    //   *      This file is distributed under the terms of the      *)
(*         *       GNU Lesser General Public License Version 2.1       *)
(***********************************************************************)

(* Coq SVN: FSetInterface.v 10616 2008-03-04 17:33:35Z letouzey *)


(* Implementation notes:

   1) Coq.Classes.Equivalence: We want the definition [equiv] from here.

   2) EqDecInstances: The hints it declares are useful here. *)

Require Import Coq.Lists.List.
Require Import Coq.Lists.SetoidList.
Require Import Coq.Logic.Decidable.
Require Import Coq.Classes.Equivalence.
Require Import Coq.Classes.EquivDec.

Require Import EqDecInstances.


(* *********************************************************************** *)
(** * Class definition *)

(** A type class for finite sets where equality on elements is
    decidable.

    Compared to the FSets library, this class does not define the
    following fields: Equal, Subset, Empty, For_all, Exists, is_empty,
    singleton, eq, equal, subset, fold, for_all, exists_, filter,
    partition , cardinal, choose.

    Some of these will be defined below, as they do not need to be
    taken as "primitive." *)

Class FSet (A : Type) `(EqDec A) := {

  (** The type of finite sets of [A]s. *)
  fset : Type;

  (** Membership predicate. *)
  In   : A -> fset -> Prop;
  In_1 : forall x y s, x === y -> In x s -> In y s;

  (** Membership test. *)
  mem   : A -> fset -> bool;
  mem_1 : forall x s, In x s -> mem x s = true;
  mem_2 : forall x s, mem x s = true -> In x s;

  (** The empty set. *)
  empty : fset;
  empty_intro : forall x, ~ In x empty;

  (** Add an element to a set. *)
  add   : A -> fset -> fset;
  add_1 : forall y x s, x === y -> In y (add x s);
  add_2 : forall y x s, In y s  -> In y (add x s);
  add_3 : forall x y s, x =/= y -> In y (add x s) -> In y s;

  (** Remove an element from a set. *)
  remove   : A -> fset -> fset;
  remove_1 : forall y x s, x === y -> ~ In y (remove x s);
  remove_2 : forall y x s, x =/= y -> In y s -> In y (remove x s);
  remove_3 : forall x y s, In y (remove x s) -> In y s;

  (** Set union. *)
  union   : fset -> fset -> fset;
  union_1 : forall x s s', In x (union s s') -> In x s \/ In x s';
  union_2 : forall x s s', In x s -> In x (union s s');
  union_3 : forall x s s', In x s' -> In x (union s s');

  (** Set intersection. *)
  inter   : fset -> fset -> fset;
  inter_1 : forall s' x s, In x (inter s s') -> In x s;
  inter_2 : forall s x s', In x (inter s s') -> In x s';
  inter_3 : forall x s s', In x s -> In x s' -> In x (inter s s');

  (** [diff s1 s2] is the elements of [s1] that are not in [s2]. *)
  diff   : fset -> fset -> fset;
  diff_1 : forall s' x s, In x (diff s s') -> In x s;
  diff_2 : forall s x s', In x (diff s s') -> ~ In x s';
  diff_3 : forall x s s', In x s -> ~ In x s' -> In x (diff s s');

  (** The elements of a set as a list. *)
  elements    : fset -> list A;
  elements_1  : forall x s, In x s -> InA equiv x (elements s);
  elements_2  : forall x s, InA equiv x (elements s) -> In x s;
  elements_3w : forall s, NoDupA equiv (elements s)
}.

(** Too many arguments to [fset] are implicit.  We want the type of
    elements to be explicit, since this makes code easier to read. *)

Implicit Arguments fset [ [R] [equiv0] [H] [FSet] ].


(* *********************************************************************** *)
(** * Additional definitions and notations *)

Definition singleton `{FSet A} (x : A) : fset A := add x empty.

Definition Empty `{FSet A} (s : fset A) : Prop := forall x, ~ In x s.

Definition Equal `{FSet A} (s1 s2 : fset A) : Prop :=
  forall x, In x s1 <-> In x s2.

Definition Subset `{FSet A} (s1 s2 : fset A) : Prop :=
  forall x, In x s1 -> In x s2.

Notation "s  [=]  t" :=
  (Equal s t) (at level 70, no associativity) : set_scope.
Notation "s  [<=]  t" :=
  (Subset s t) (at level 70, no associativity) : set_scope.
Notation "x `in` s" :=
  (In x s) (at level 70, no associativity) : set_scope.
Notation "x `notin` s" :=
  (~ In x s) (at level 70, no associativity) : set_scope.
Notation "s `union` t" :=
  (union s t) (at level 65, right associativity) : set_scope.
Notation "s `inter` t" :=
  (inter s t) (at level 65, right associativity) : set_scope.

Open Scope set_scope.

Bind Scope set_scope with fset.

Arguments Scope In [ type_scope _ _ _ _ _ set_scope ].
Arguments Scope add [ type_scope _ _ _ _ _ set_scope ].
Arguments Scope remove [ type_scope _ _ _ _ _ set_scope ].
Arguments Scope union [ type_scope _ _ _ _ set_scope set_scope ].
Arguments Scope inter [ type_scope _ _ _ _ set_scope set_scope ].
Arguments Scope diff [ type_scope _ _ _ _ set_scope set_scope ].
Arguments Scope elements [ type_scope _ _ _ _ set_scope ].
Arguments Scope Empty [ type_scope _ _ _ _ set_scope ].
Arguments Scope Equal [ type_scope _ _ _ _ set_scope set_scope ].
Arguments Scope Subset [ type_scope _ _ _ _ set_scope set_scope ].


(* *********************************************************************** *)
(** * Decidability *)

Lemma In_dec : forall `{FSet A} (x : A) (s : fset A),
  {In x s} + {~ In x s}.
Proof.
  intros ? ? ? ? ? x s.
  case_eq (mem x s); intros eq.
    left; auto using mem_2.
    right; intros J.
      assert (mem x s = true) by auto using mem_1.
      congruence.
Defined.

Lemma dec_In : forall `{FSet A} (x : A) (s : fset A),
  decidable (In x s).
Proof.
  unfold decidable.
  intros ? ? ? ? ? x s.
  destruct (In_dec x s); auto.
Defined.


(* *********************************************************************** *)
(** * Properties *)

Lemma singleton_1 : forall `{FSet A} (y x : A),
  In y (singleton x) ->
  x === y.
Proof.
  unfold singleton.
  intros ? ? ? ? ? y x H.
  destruct (equiv_dec x y) as [J | J]; [ now auto | ].
  assert (In y empty) by eauto using add_3.
  assert (~ In y empty) by auto using empty_intro.
  contradiction.
Qed.

Lemma singleton_2 : forall `{FSet A} (y x : A),
  x === y ->
  In y (singleton x).
Proof. unfold singleton. intros. auto using add_1. Qed.

Lemma Equal_refl : forall `{FSet A} (s1 : fset A),
  Equal s1 s1.
Proof. unfold Equal. intuition. Qed.

Lemma Equal_sym : forall `{FSet A} (s1 s2 : fset A),
  Equal s1 s2 ->
  Equal s2 s1.
Proof. unfold Equal. firstorder. Qed.

Lemma Equal_trans : forall `{FSet A} (s1 s2 s3 : fset A),
  Equal s1 s2 ->
  Equal s2 s3 ->
  Equal s1 s3.
Proof. unfold Equal. firstorder. Qed.

Lemma Subset_refl : forall `{FSet A} (s1 : fset A),
  Subset s1 s1.
Proof. unfold Subset. intuition. Qed.

Lemma Subset_trans : forall `{FSet A} (s1 s2 s3 : fset A),
  Subset s1 s2 ->
  Subset s2 s3 ->
  Subset s1 s3.
Proof. unfold Subset. intuition. Qed.


(* *********************************************************************** *)
(** * Set operations and propositional connectives *)

Lemma empty_iff : forall `{FSet A} (x : A),
  In x empty <-> False.
Proof. split. fold (~ In x empty); auto using empty_intro. inversion 1. Qed.

Lemma add_iff : forall `{FSet A} (x y : A) (s : fset A),
  In y (add x s) <-> x === y \/ In y s.
Proof.
  split.
  destruct (y == x); eauto using add_3.
  intuition (auto using add_1, add_2).
Qed.

Lemma remove_iff : forall `{FSet A} (x y : A) (s : fset A),
  In y (remove x s) <-> In y s /\ x =/= y.
Proof.
  split.
  intros K; split.
    eauto using remove_3.
    intros ?. contradict K. auto using remove_1.
  intuition (auto using remove_2).
Qed.

Lemma union_iff : forall `{FSet A} (x : A) (s1 s2 : fset A),
  In x (union s1 s2) <-> In x s1 \/ In x s2.
Proof.
  split.
  auto using union_1.
  intuition (auto using union_2, union_3).
Qed.

Lemma inter_iff : forall `{FSet A} (x : A) (s1 s2 : fset A),
  In x (inter s1 s2) <-> In x s1 /\ In x s2.
Proof.
  split.
  intuition (eauto using inter_1, inter_2).
  intuition (auto using inter_3).
Qed.

Lemma diff_iff : forall `{FSet A} (x : A) (s1 s2 : fset A),
  In x (diff s1 s2) <-> In x s1 /\ ~ In x s2.
Proof.
  split.
  split. eauto using diff_1. eauto using diff_2.
  intuition (auto using diff_3).
Qed.

Lemma singleton_iff : forall `{FSet A} (x y : A),
  In y (singleton x) <-> x === y.
Proof. intuition (eauto using singleton_1, singleton_2). Qed.


(* *********************************************************************** *)
(** * Morphisms *)

(** Additional morphisms are defined in the FSetsClassFacts library.
    These are merely the morphisms required to implement the [fsetdec]
    tactic in the FSetsClassTactics library. *)

Add Parametric Relation `(FSet A) : (fset A) Equal
  reflexivity proved by Equal_refl
  symmetry proved by Equal_sym
  transitivity proved by Equal_trans
  as relation_fset_Equal.

Add Parametric Morphism `(FSet A) : (@In A _ _ _ _)
  with signature equiv ==> Equal ==> iff as morphism_In.
Proof. unfold Equal. firstorder (eauto using In_1). Qed.


(* *********************************************************************** *)
(** * Hints *)

Hint Resolve
  @mem_1 @mem_2
  @empty_intro
  @add_1 @add_2
  @remove_1 @remove_2
  @union_1 @union_2 @union_3
  @inter_3
  @diff_3
  @elements_1 @elements_2 @elements_3w
  @singleton_1 @singleton_2
  @Equal_refl @Equal_sym
  @Subset_refl
  : set.

Hint Immediate
  @In_1
  @add_3
  @remove_3
  @inter_1 @inter_2
  @diff_1 @diff_2
  @elements_2
  @Equal_trans
  @Subset_trans
  : set.

Hint Resolve @In_dec      : set.
Hint Resolve @singleton_1 : set.
Hint Resolve @singleton_2 : set.
