(* This file is distributed under the terms of the MIT License, also
   known as the X11 Licence.  A copy of this license is in the README
   file that accompanied the original distribution of this file.

   Based on code written by:
     Brian Aydemir

   $Id: FSetsClassFacts.v 2834 2009-03-04 03:06:06Z baydemir $ *)

(** A library contains facts about finite sets. *)

Require Import Coq.Setoids.Setoid.
Require Import Coq.Classes.Equivalence.
Require Import Coq.Classes.EquivDec.

Require Import FSetsClass.
Require Import FSetsClassDecide.


(* *********************************************************************** *)
(** * Morphisms *)

(** Some morphisms are already declared in the FSetsClass library. *)

Add Parametric Morphism `(FSet A) : (@add A _ _ _ _)
  with signature equiv ==> Equal ==> Equal as morphism_add.
Proof. fsetdec A. Qed.

Add Parametric Morphism `(FSet A) : (@remove A _ _ _ _)
  with signature equiv ==> Equal ==> Equal as morphism_remove.
Proof. fsetdec A. Qed.

Add Parametric Morphism `(FSet A) : (@union A _ _ _ _)
  with signature Equal ==> Equal ==> Equal as morphism_union.
Proof. fsetdec A. Qed.

Add Parametric Morphism `(FSet A) : (@inter A _ _ _ _)
  with signature Equal ==> Equal ==> Equal as morphism_inter.
Proof. fsetdec A. Qed.

Add Parametric Morphism `(FSet A) : (@diff A _ _ _ _)
  with signature Equal ==> Equal ==> Equal as morphism_diff.
Proof. fsetdec A. Qed.

Add Parametric Morphism `(FSet A) : (@singleton A _ _ _ _)
  with signature equiv ==> Equal as morphism_singleton.
Proof. fsetdec A. Qed. 

Add Parametric Morphism `(FSet A) : (@Empty A _ _ _ _)
  with signature Equal ==> iff as morphism_Empty.
Proof. split; fsetdec A. Qed.

Add Parametric Morphism `(FSet A) : (@Equal A _ _ _ _)
  with signature Equal ==> Equal ==> iff as morphism_Equal.
Proof. split; fsetdec A. Qed.

Add Parametric Morphism `(FSet A) : (@Subset A _ _ _ _)
  with signature Equal ==> Equal ==> iff as morphism_Subset.
Proof. split; fsetdec A. Qed.


(* *********************************************************************** *)
(** * Facts about [remove] *)

Lemma remove_empty : forall `{FSet A} (x : A),
  remove x empty [=] empty.
Proof. fsetdec A. Qed.

Lemma remove_singleton_same : forall `{FSet A} (x : A),
  remove x (singleton x) [=] empty.
Proof. fsetdec A. Qed.

Lemma remove_singleton_diff : forall `{FSet A} (x y : A),
  y =/= x ->
  remove y (singleton x) [=] singleton x.
Proof. fsetdec A. Qed.

Lemma remove_union_distrib : forall `{FSet A} (s1 s2 : fset A) (x : A),
  remove x (s1 `union` s2) [=] remove x s1 `union` remove x s2.
Proof. fsetdec A. Qed.


(* *********************************************************************** *)
(** * Facts about [Equal] *)

Lemma Equal_union_compat : forall `{FSet A} (s1 s2 s3 s4 : fset A),
  s1 [=] s3 ->
  s2 [=] s4 ->
  s1 `union` s2 [=] s3 `union` s4.
Proof. fsetdec A. Qed.


(* *********************************************************************** *)
(** * Facts about [Subset] *)

Lemma Subset_empty_any : forall `{FSet A} (s : fset A),
  empty [<=] s.
Proof. fsetdec A. Qed.

Lemma Subset_union_compat : forall `{FSet A} (s1 s2 s3 s4 : fset A),
  s1 [<=] s3 ->
  s2 [<=] s4 ->
  s1 `union` s2 [<=] s3 `union` s4.
Proof. fsetdec A. Qed.

Lemma Subset_union_left : forall `{FSet A} (s1 s2 s3 : fset A),
  s1 [<=] s2 ->
  s1 [<=] s2 `union` s3.
Proof. fsetdec A. Qed.

Lemma Subset_union_right : forall `{FSet A} (s1 s2 s3 : fset A),
  s1 [<=] s3 ->
  s1 [<=] s2 `union` s3.
Proof. fsetdec A. Qed.


(* *********************************************************************** *)
(** * Hints *)

Hint Resolve
  @remove_empty
  @remove_singleton_same
  @remove_singleton_diff
  @remove_union_distrib
  @Equal_union_compat
  @Subset_empty_any
  @Subset_union_compat
  @Subset_union_left
  @Subset_union_right
  : set.
