(* $Id: FSetsClassImpl.v 2731 2009-02-24 17:29:24Z baydemir $

   Includes code written by:
     Brian Aydemir *)

(** A type class based implementation of finite sets.  The design and
    implementation is inspired by (and uses code from) Coq's FSets
    library. *)


(***********************************************************************)
(*  v      *   The Coq Proof Assistant  /  The Coq Development Team    *)
(* <O___,, *        INRIA-Rocquencourt  &  LRI-CNRS-Orsay              *)
(*   \VV/  *************************************************************)
(*    //   *      This file is distributed under the terms of the      *)
(*         *       GNU Lesser General Public License Version 2.1       *)
(***********************************************************************)

(* Coq SVN: FSetWeakList.v 10631 2008-03-06 18:17:24Z msozeau *)


(* Implementation notes:

   1) Coq.Classes.Equivalence: We want the definition [equiv] from here.

   2) EqDecInstances: The hints it declares are useful here. *)

Require Import Coq.Lists.List.
Require Import Coq.Lists.SetoidList.
Require Import Coq.Classes.Equivalence.
Require Import Coq.Classes.EquivDec.

Require Import DefaultSolver.
Require Import EqDecInstances.
Require Import FSetsClass.


(* *********************************************************************** *)
(** * Functions over lists *)

(** We provide sets as lists which are (morally) without redundancy.
    The specs are proved under the additional condition of no
    redundancy, and the functions returning sets are proved to
    preserve this invariant.

    Since users are not expected to use directly the contents of this
    module, we hide its contents for documentation purposes. *)

Module Raw.
Section Raw_params.

  Variable A : Type.
  Variable R : relation A.
  Variable Equiv : @Equivalence A R.
  Variable EqDecA : @EqDec A R Equiv.

  (* begin hide *)

  Notation Local t := (list A).

  Fixpoint mem (x : A) (s : t) {struct s} : bool :=
    match s with
      | nil => false
      | cons y l => if x == y then true else mem x l
    end.

  Definition empty : t := nil.

  Fixpoint add (x : A) (s : t) {struct s} : t :=
    match s with
      | nil => cons x nil
      | cons y l => if x == y then s else y :: add x l
    end.

  Fixpoint remove (x : A) (s : t) {struct s} : t :=
    match s with
      | nil => nil
      | y :: l => if x == y then l else y :: remove x l
    end.

  Fixpoint fold (B : Type) (f : A -> B -> B) (s : t) {struct s} :  B -> B :=
    fun i => match s with
               | nil => i
               | cons x l => fold _ f l (f x i)
             end.

  Implicit Arguments fold [B].

  Definition union (s : t) : t -> t := fold add s.

  Definition inter (s s' : t) : t :=
    fold (fun x s => if mem x s' then add x s else s) s nil.

  Definition diff (s s' : t) : t := fold remove s' s.

  Definition elements (s : t) : list A := s.

  Section ForNotations.

  Notation In := (InA equiv).
  Notation NoDup := (NoDupA equiv).

  Lemma In_1 : forall (x y : A) (s : t), x === y -> In x s -> In y s.
  Proof.
    intros.
    eapply SetoidList.InA_eqA; eauto.
    intros.
    eapply transitivity; eauto.
  Qed.

  Lemma mem_1 : forall (x : A) (s : t), In x s -> mem x s = true.
  Proof with default_simp. induction s... inversion H... Qed.

  Lemma mem_2 : forall (x : A) (s : t), mem x s = true -> In x s.
  Proof with default_simp. induction s... Qed.

  Lemma empty_intro : forall x, ~ In x empty.
  Proof. intros x J; inversion J. Qed.

  Lemma empty_unique : NoDup empty.
  Proof. unfold empty; constructor. Qed.

  Lemma add_1 : forall (y x : A) (s : t) (Hs : NoDup s),
    x === y -> In y (add x s).
  Proof with default_simp.
    induction s...
    constructor. symmetry in H. eapply transitivity; eauto.
  Qed.

  Lemma add_2 : forall (y x : A) (s : t) (Hs : NoDup s),
    In y s -> In y (add x s).
  Proof with default_simp. induction s... inversion H... Qed.

  Lemma add_3 : forall (x y : A) (s : t) (Hs : NoDup s),
    x =/= y -> In y (add x s) -> In y s.
  Proof with default_simp.
    induction s...
    inversion H0... assert False...
    inversion H0...
  Qed.

  Lemma add_unique : forall (x:A) (s : t) (Hs : NoDup s), NoDup (add x s).
  Proof with default_simp.
    induction s...
    constructor... intros J...
    constructor...
      intros J.
      assert (In a s) by eauto using add_3.
      contradiction.
  Qed.

  Hint Local Resolve add_unique.

  Lemma remove_1 : forall y x s (Hs : NoDup s),
    x === y -> ~ In y (remove x s).
  Proof with default_simp.
    induction s...
    intros J...
    intros J. assert (In a s) by eauto using In_1. contradiction.
    intros J; inversion J...
      assert (x === a) by (eapply transitivity; eauto). contradiction.
      contradict H1...
  Qed.

  Lemma remove_2 : forall y x s (Hs : NoDup s),
    x =/= y -> In y s -> In y (remove x s).
  Proof with default_simp.
    induction s; intros ? ? H...
    inversion H...
      assert (J : x === y) by (symmetry in H2; eapply transitivity; eauto).
      contradiction.
    inversion H...
  Qed.

  Lemma remove_3 : forall x y s (Hs : NoDup s),
    In y (remove x s) -> In y s.
  Proof with default_simp. induction s; intros ? H... inversion H... Qed.

  Lemma remove_unique : forall (s : t) (Hs : NoDup s) (x : A),
    NoDup (remove x s).
  Proof with default_simp.
    induction s...
    constructor...
    intros J. assert (In a s) by eauto using remove_3. contradiction.
  Qed.

  Hint Local Resolve remove_unique.

  Lemma union_1 : forall (x : A) (s s' : t) (Hs : NoDup s) (Hs' : NoDup s'),
    In x (union s s') -> In x s \/ In x s'.
  Proof with default_simp.
    unfold union; induction s...
    destruct (x == a)...
    destruct (IHs (add a s'))...
    eauto using add_3.
  Qed.

  Lemma union_0 : forall (x : A) (s s' : t) (Hs : NoDup s) (Hs' : NoDup s'),
    In x s \/ In x s' -> In x (union s s').
  Proof with default_simp.
    unfold union; induction s...
    apply IHs... inversion H... auto using add_1.
    apply IHs... auto using add_2.
  Qed.

  Lemma union_2 : forall (x : A) (s s' : t) (Hs : NoDup s) (Hs' : NoDup s'),
    In x s -> In x (union s s').
  Proof. auto using union_0. Qed.

  Lemma union_3 : forall (x : A) (s s' : t) (Hs : NoDup s) (Hs' : NoDup s'),
    In x s' -> In x (union s s').
  Proof. auto using union_0. Qed.

  Lemma union_unique : forall (s s' : t) (Hs : NoDup s) (Hs' : NoDup s'),
    NoDup (union s s').
  Proof with default_simp. unfold union; induction s... Qed.

  Lemma inter_0 : forall (s s' : t) (x : A)  (Hs : NoDup s) (Hs' : NoDup s'),
    In x (inter s s') -> In x s /\ In x s'.
  Proof with default_simp.
    unfold inter; intros s s' x Hs Hs' H.
    set (acc := nil (A:=A)) in *.
    assert (Q : NoDup acc) by (unfold acc; auto).
    assert ((In x s /\ In x s') \/ In x acc).
      revert s' x acc Q Hs Hs' H.
      induction s...
      case_eq (mem a s'); intros eq; rewrite eq in *...
        destruct (IHs s' x (add a acc))... destruct (x == a).
          left; split; eauto using mem_2, In_1.
          assert (In x acc) by eauto using add_3; auto.
        destruct (IHs s' x acc)...
    default_simp.
  Qed.

  Lemma inter_1 : forall s' x s (Hs : NoDup s) (Hs' : NoDup s'),
    In x (inter s s') -> In x s.
  Proof. intros. cut (In x s /\ In x s'). intuition. auto using inter_0. Qed.

  Lemma inter_2 : forall s x s' (Hs : NoDup s) (Hs' : NoDup s'),
    In x (inter s s') -> In x s'.
  Proof. intros. cut (In x s /\ In x s'). intuition. auto using inter_0. Qed.

  Lemma inter_3 : forall x s s' (Hs : NoDup s) (Hs' : NoDup s'),
    In x s -> In x s' -> In x (inter s s').
  Proof with default_simp.
    intros x s s' Hs Hs'.
    assert (((In x s /\ In x s') \/ In x nil) -> In x (inter s s')).
      unfold inter.
      set (acc := nil (A:=A)) in *.
      assert (Q : NoDup acc) by (unfold acc; auto).
      revert x s s' acc Hs Hs' Q.
      induction s...
        case_eq (mem a s'); intros eq; apply IHs; default_simp.
          inversion H0... auto using add_1.
          inversion H0...
            assert (In a s') by eauto using In_1.
            assert (mem a s' = true) by auto using mem_1.
            congruence.
        case_eq (mem a s'); intros eq; apply IHs; default_simp.
          auto using add_2.
    default_simp.
  Qed.

  Lemma inter_unique : forall (s s' : t) (Hs : NoDup s) (Hs' : NoDup s'),
    NoDup (inter s s').
  Proof with default_simp.
    unfold inter; intros s.
    set (acc := nil (A:=A)).
    assert (NoDup acc) by (unfold acc; auto).
    generalize dependent acc.
    induction s...
    destruct (mem a s')...
  Qed.

  Lemma diff_0 : forall s' s x (Hs : NoDup s) (Hs' : NoDup s'),
    In x (diff s s') -> In x s /\ ~ In x s'.
  Proof with default_simp.
    unfold diff; induction s'...
    split... intros J...
    destruct (IHs' (remove a s) x)...
      assert (In x s) by eauto using remove_3.
      split...
      intros J; inversion J...
      assert (~ In x (remove a s)) by auto using remove_1.
      contradiction.
  Qed.

  Lemma diff_1 : forall s' x s (Hs : NoDup s) (Hs' : NoDup s'),
    In x (diff s s') -> In x s.
  Proof. intros. cut (In x s /\ ~ In x s'). intuition. auto using diff_0. Qed.

  Lemma diff_2 : forall s x s' (Hs : NoDup s) (Hs' : NoDup s'),
    In x (diff s s') -> ~ In x s'.
  Proof. intros. cut (In x s /\ ~ In x s'). intuition. auto using diff_0. Qed.

  Lemma diff_3 : forall (x : A) (s s' : t) (Hs : NoDup s) (Hs' : NoDup s'),
    In x s -> ~ In x s' -> In x (diff s s').
  Proof with default_simp.
    unfold diff; intros x s s'; revert x s; induction s'...
    apply IHs'...
    apply remove_2...
    intros J; assert (In x (a :: s')) by auto; contradiction.
  Qed.

  Lemma diff_unique : forall (s s' : t) (Hs : NoDup s) (Hs' : NoDup s'),
    NoDup (diff s s').
  Proof with default_simp.
    intros s s'; revert s; unfold diff; induction s'...
  Qed.

  Lemma elements_1 : forall (x : A) (s : t), In x s -> In x (elements s).
  Proof. unfold elements; auto. Qed.

  Lemma elements_2 : forall (x : A) (s : t), In x (elements s) -> In x s.
  Proof. unfold elements; auto. Qed.

  Lemma elements_3w : forall (s : t) (Hs : NoDup s), NoDup (elements s).
  Proof. unfold elements; auto. Qed.

  End ForNotations.

  (* end hide *)

  (** [fset_list] is the implementation type for finite sets. *)

  Record fset_list : Type :=
    fset_intro { this :> list A; unique : NoDupA equiv this }.

End Raw_params.
End Raw.


(* *********************************************************************** *)
(** * Class implementation *)

(** We declare some local hints and coercions in order to simplify the
    instance declaration below. *)

Section ForLocalDeclarations.

Hint Local Resolve Raw.unique.
Local Coercion Raw.unique : Raw.fset_list >-> NoDupA.
Local Implicit Arguments Raw.fset_list [ [R] [Equiv] ].
Local Implicit Arguments Raw.fset_intro [ [A] [R] [Equiv] this ].

(** We can construct an [FSet] dictionary given an [EqDec] dictionary. *)

Global Instance FSet_of_EqDec `(eqa : EqDec A) : FSet A eqa := {
  fset   := Raw.fset_list A;
  In     := fun x s => InA equiv x s;
  mem    := fun x s => Raw.mem _ _ _ eqa x s;
  empty  := Raw.fset_intro (Raw.empty_unique _ _ _);
  add    := fun x s => Raw.fset_intro (Raw.add_unique _ _ _ eqa x s s);
  remove := fun x s => Raw.fset_intro (Raw.remove_unique _ _ _ eqa s s x);
  union  := fun s s' => Raw.fset_intro (Raw.union_unique _ _ _ eqa s s' s s');
  inter  := fun s s' => Raw.fset_intro (Raw.inter_unique _ _ _ eqa s s' s s');
  diff   := fun s s' => Raw.fset_intro (Raw.diff_unique _ _ _ eqa s s' s s');
  elements := fun s => s
}.
Proof.
  eauto using Raw.In_1.
  eauto using Raw.mem_1.
  eauto using Raw.mem_2.
  simpl. eauto using Raw.empty_intro.
  simpl. eauto using Raw.add_1.
  simpl. eauto using Raw.add_2.
  simpl. eauto using Raw.add_3.
  simpl. eauto using Raw.remove_1.
  simpl. eauto using Raw.remove_2.
  simpl. eauto using Raw.remove_3.
  simpl. eauto using Raw.union_1.
  simpl. eauto using Raw.union_2.
  simpl. eauto using Raw.union_3.
  simpl. eauto using Raw.inter_1.
  simpl. eauto using Raw.inter_2.
  simpl. eauto using Raw.inter_3.
  simpl. eauto using Raw.diff_1.
  simpl. eauto using Raw.diff_2.
  simpl. eauto using Raw.diff_3.
  eauto using Raw.elements_1.
  eauto using Raw.elements_2.
  destruct s. auto.
Defined.

End ForLocalDeclarations.

(** In practice, we don't want operations on finite sets to be expanded
    out into their definitions. *)

Strategy opaque [ FSet_of_EqDec ].
