(* This file is distributed under the terms of the MIT License, also
   known as the X11 Licence.  A copy of this license is in the README
   file that accompanied the original distribution of this file.

   Based on code written by:
     Brian Aydemir
     Arthur Charg\'eraud

   $Id: FSetsClassNotin.v 2820 2009-03-02 20:26:16Z baydemir $ *)

(** A library that defines tactics for working with non-membership in
    finite sets. *)

Require Import Coq.Classes.EquivDec.

Require Import FSetsClass.


(* *********************************************************************** *)
(** * Facts about set non-membership *)

Lemma notin_empty : forall `{FSet A} (x : A),
  ~ In x empty.
Proof. intros. auto using empty_intro. Qed.

Lemma notin_union : forall `{FSet A} (x : A) (E F : fset A),
  ~ In x E ->
  ~ In x F ->
  ~ In x (union E F).
Proof.
  intros ? ? ? ? ? x E F H J K.
  edestruct union_1. eassumption. intuition. intuition.
Qed.

Lemma elim_notin_union : forall `{FSet A} (x : A) (E F : fset A),
  ~ In x (union E F) ->
  (~ In x E) /\ (~ In x F).
Proof.
  intros ? ? ? ? ? x E F H.
  split; intros J; contradiction H.
    auto using union_2.
    auto using union_3.
Qed.

Lemma notin_add : forall `{FSet A} (x y : A) (F : fset A),
  y =/= x ->
  ~ In x F ->
  ~ In x (add y F).
Proof.
  intros ? ? ? ? ? x y F H J K.
  contradiction J.
  eauto using add_3.
Qed.

Lemma elim_notin_add : forall `{FSet A} (x y : A) (F : fset A),
  ~ In x (add y F) ->
  y =/= x /\ ~ In x F.
Proof.
  intros ? ? ? ? ? x y F H; split; intros J; contradiction H.
    auto using add_1.
    auto using add_2.
Qed.

Lemma elim_notin_add' : forall `{FSet A} (x y : A) (F : fset A),
  ~ In x (add y F) ->
  y <> x /\ ~ In x F.
Proof.
  intros ? ? ? ? ? x y F H; split; intros J; contradiction H.
    subst. auto using add_1.
    auto using add_2.
Qed.

Lemma notin_diff_1 : forall `{FSet A} (x : A) (s s' : fset A),
  ~ In x s ->
  ~ In x (diff s s').
Proof. intros. eauto using diff_1. Qed.

Lemma notin_diff_2 : forall `{FSet A} (x : A) (s s' : fset A),
  In x s' ->
  ~ In x (diff s s').
Proof.
  intros ? ? ? ? ? x s s' H J.
  assert (~ In x s') by eauto using diff_2.
  contradiction.
Qed.

Lemma elim_notin_diff : forall `{FSet A} (x : A) (s s' : fset A),
  ~ In x (diff s s') ->
  ~ In x s \/ In x s'.
Proof.
  intros ? ? ? ? ? x s s' H.
  case (In_dec x s);  [ intros J1 | auto ].
  case (In_dec x s'); [ auto | intros J2 ].
  assert (In x (diff s s')) by auto using diff_3.
  eauto using diff_2.
Qed.

Lemma notin_singleton : forall `{FSet A} (x y : A),
  x =/= y ->
  ~ In x (singleton y).
Proof.
  intros ? ? ? ? ? x y H J.
  assert (x === y) by eauto using singleton_1.
  contradiction.
Qed.

Lemma elim_notin_singleton : forall `{FSet A} (x y : A),
  ~ In x (singleton y) ->
  x =/= y.
Proof.
  intros ? ? ? ? ? x y H J.
  contradiction H.
  auto using singleton_2.
Qed.

Lemma elim_notin_singleton' : forall `{FSet A} (x y : A),
  ~ In x (singleton y) ->
  x <> y.
Proof.
  intros ? ? ? ? ? x y H J.
  assert (x =/= y) by eauto using elim_notin_singleton.
  subst; auto.
Qed.

Lemma notin_remove_1 : forall `{FSet A} (x y : A) (s : fset A),
  ~ In x s ->
  ~ In x (remove y s).
Proof. intros. eauto using remove_3. Qed.

Lemma notin_remove_2 : forall `{FSet A} (x y : A) (s : fset A),
  x === y ->
  ~ In x (remove y s).
Proof. intros. auto using remove_1. Qed.

Lemma notin_remove_3 : forall `{FSet A} (x y : A) (s : fset A),
  x = y ->
  ~ In x (remove y s).
Proof. intros. subst. auto using remove_1. Qed.

Lemma elim_notin_remove : forall `{FSet A} (x y : A) (s : fset A),
  ~ In x (remove y s) ->
  y === x \/ ~ In x s.
Proof.
  intros ? ? ? ? ? x y s H.
  destruct (equiv_dec y x) as [J | J].
    auto.
    right; intros K.
      assert (In x (remove y s)) by auto using remove_2.
      contradiction.
Qed.


(* *********************************************************************** *)
(** * Tactics for non-membership *)

(** The tactic [destruct_notin] tries to destruct hypotheses of the
    form [(~ In x E)]. *)

Ltac destruct_notin :=
  match goal with
    | H : In ?x ?E -> False |- _ =>
      change (~ In x E) in H;
      destruct_notin
    | H : ~ In ?x (add ?y ?F) |- _ =>
      destruct (elim_notin_add x y F H);
      destruct (elim_notin_add' x y F H);
      clear H;
      destruct_notin
    | H : ~ In ?x (diff ?s ?s') |- _ =>
      let J := fresh in
      destruct (elim_notin_diff x s s') as [J | J];
      clear H;
      destruct_notin
    | H : ~ In _ empty |- _ =>
      clear H;
      destruct_notin
    | H : ~ In ?x (remove ?y ?E) |- _ =>
      destruct (elim_notin_remove x y E);
      clear H;
      destruct_notin
    | H : ~ In ?x (singleton ?y) |- _ =>
      let F1 := fresh in
      let F2 := fresh in
      assert (F1 := elim_notin_singleton x y H);
      assert (F2 := elim_notin_singleton' x y H);
      clear H;
      destruct_notin
    | H : ~ In ?x (union ?E ?F) |- _ =>
      destruct (elim_notin_union x E F H);
      clear H;
      destruct_notin
    | _ =>
      idtac
  end.

(** [solve_notin] tries to solves goals of the form [(x <> y)] and
    [(~ In x E)] using hypotheses about set non-membership.  It is
    intended to be simple and fast, not complete. *)

Ltac solve_notin :=
  destruct_notin;
  repeat (progress (  apply notin_add
                   || apply notin_diff_1
                   || apply notin_diff_2
                   || apply notin_empty
                   || apply notin_remove_1
                   || apply notin_remove_2
                   || apply notin_singleton
                   || apply notin_union));
  solve [ trivial | congruence        | symmetry; congruence
                  | intuition trivial | symmetry; intuition trivial ].


(* *********************************************************************** *)
(** * Examples and test cases *)

(** These examples and test cases are not meant to be exhaustive. *)

Module FSetNotinTestCases.

Lemma test_solve_notin_1 : forall `{FSet A} (x : A) (E F G : fset A),
  ~ In x (union E F) ->
  ~ In x G ->
  ~ In x (union E G).
Proof. intros. solve_notin. Qed.

Lemma test_solve_notin_2 : forall `{FSet A} (x y : A) (E F G : fset A),
  ~ In x (union E (union (singleton y) F)) ->
  ~ In x G ->
  ~ In x (singleton y) /\ ~ In y (singleton x).
Proof. intros. split. solve_notin. solve_notin. Qed.

Lemma test_solve_notin_3 : forall `{FSet A} (x y : A),
  x =/= y ->
  ~ In x (singleton y) /\
  ~ In y (singleton x).
Proof. intros. split. solve_notin. solve_notin. Qed.

Lemma test_solve_notin_4 : forall `{FSet A} (x y : A) (E F G : fset A),
  ~ In x (union E (union (singleton x) F)) ->
  ~ In y G.
Proof. intros. solve_notin. Qed.

Lemma test_solve_notin_5 : forall `{FSet A} (x y : A) (E F : fset A),
  ~ In x (union E (union (singleton y) F)) ->
  ~ In y E ->
  y =/= x /\ x =/= y.
Proof. intros. split. solve_notin. solve_notin. Qed.

Lemma test_solve_notin_6 : forall `{FSet A} (x y : A) (E : fset A),
  ~ In x (add y E) ->
  x =/= y /\ ~ In x E.
Proof. intros. split. solve_notin. solve_notin. Qed.

End FSetNotinTestCases.


(* *********************************************************************** *)
(** * Hints *)

Hint Resolve @notin_empty     : set.
Hint Resolve @notin_union     : set.
Hint Resolve @notin_add       : set.
Hint Resolve @notin_diff_1    : set.
Hint Resolve @notin_diff_2    : set.
Hint Resolve @notin_singleton : set.
Hint Resolve @notin_remove_1  : set.
Hint Resolve @notin_remove_2  : set.
Hint Resolve @notin_remove_3  : set.
