(* This file is distributed under the terms of the MIT License, also
   known as the X11 Licence.  A copy of this license is in the README
   file that accompanied the original distribution of this file.

   Based on code written by:
     Brian Aydemir
     Aaron Bohannon
     Arthur Charg\'eraud

   $Id: Metatheory.v 2768 2009-02-27 01:35:59Z baydemir $ *)

Require Export Coq.Arith.Arith.
Require Export Coq.Classes.Equivalence.
Require Export Coq.Classes.EquivDec.
Require Export Coq.Lists.List.

Require Export AdditionalTactics.
Require Export AssocList.
Require Export AssocListFacts.
Require Export Atom.
Require Export DefaultSolver.
Require Export EqDecInstances.
Require Export FSetsClass.
Require Export FSetsClassDecide.
Require Export FSetsClassFacts.
Require Export FSetsClassNotin.
Require Export SwapClass.


(* ********************************************************************** *)
(** * Assorted functions not in the standard library *)

(** [lt_ge_dec] is a useful comparison operation when defining the
    "close" operation on locally nameless terms. *)

Definition lt_ge_dec (n m : nat) : {n < m} + {n >= m} :=
  match Compare_dec.le_gt_dec m n with
    | left pf => right pf
    | right pf => left pf
  end.


(* ********************************************************************** *)
(** * Tactic for working with cofinite quantification *)

(** Consider a rule [H] (equivalently, constructor, lemma, etc.) whose
    type begins with [forall L, ...] and contains hypotheses of the
    form [(forall y, y `notin` L -> ...)].

    The tactic [(pick fresh x excluding F and apply H)] applies [H] to
    the current goal, instantiating [H]'s first argument ([L]) with
    the finite set of atoms [F].  In each new subgoal of the form
    [(forall y, y `notin` F -> ...)], the atom [y] is introduced as
    [x], and [(y `notin` F)] is introduced using a generated name.

    If we view [H] as a rule that uses cofinite quantification, the
    tactic can be read as picking a sufficiently fresh atom to open a
    term with. *)

Tactic Notation
  "pick" "fresh" ident(atom_name)
  "excluding" constr(L)
  "and" "apply" constr(H)
  :=
    first [apply (@H L) | eapply (@H L)];
      match goal with
        | |- forall _, _ `notin` _ -> _ =>
          let Fr := fresh "Fr" in intros atom_name Fr
        | |- forall _, _ `notin` _ -> _ =>
          fail 1 "because" atom_name "is already defined"
        | _ =>
          idtac
      end.


(* ********************************************************************** *)
(** * Automation *)

(** We redefine [default_auto] so that it takes advantage of facts
    about finite sets. *)

Ltac default_auto ::= auto with arith brute_force set.

(** We provide [simpl_env] and [rewrite_env] tactics that merely
    rename similar tactics from the AssocList library. *)

Ltac simpl_env :=
  autorewrite with rewr_list rewr_list_in rewr_map rewr_dom.
Tactic Notation "simpl_env" "in" hyp(H) :=
  autorewrite with rewr_list rewr_list_in rewr_map rewr_dom in H.
Tactic Notation "simpl_env" "in" "*" :=
  autorewrite with rewr_list rewr_list_in rewr_map rewr_dom in *.

Tactic Notation "rewrite_env" constr(E) :=
  match goal with
    | |- context[?x] =>
      change x with E
    | |- context[?x] =>
      replace x with E;
        [ | try reflexivity; simpl_env; reflexivity ]
  end.

Tactic Notation "rewrite_env" constr(E) "in" hyp(H) :=
  match type of H with
    | context[?x] =>
      change x with E in H
    | context[?x] =>
      replace x with E in H;
        [ | try reflexivity; simpl_env; reflexivity ]
  end.

(** The next block of hints should discharge many of the freshness and
    inequality goals that arise in programming language metatheory
    proofs, in particular those arising from cofinite quantification.

    Implementation note: The invocation of [simpl_env] is risky since
    [autorewrite] does not interact nicely with the existential
    variables introduced by [eauto].  In practice, this does not seem
    to be a problem.  However, practice also shows that a similar
    problem does occurs if we use [fsetdec] instead of [solve_notin].

    XXX BEA: In the three hints below, the [eassumption] step is
    intended to reduce the chance that [simpl_env] will trigger a
    rewriting bug in Coq.  So much for the previous note...

    XXX BEA: I need to completely review these hints. *)

Hint Extern 4 (_ `notin` _) =>
  try eassumption; simpl_env in *; solve_notin.
Hint Extern 4 (_ <> _) =>
  try eassumption; simpl_env in *; solve_notin.
Hint Extern 4 (_ =/= _) =>
  try eassumption; simpl_env in *; solve_notin.
